/**
 * @file  ChannelBase.cpp
 * @brief Implementierung der Klasse CChannelBase.
 */
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "ChannelBase.h"
#include <time.h>
#include <stdio.h>

// ##################################################################
/**
 * @brief Initialisieren der Membervariablen.
 *
 * @param id Nummer der Kanalklasse fuer z.B. Trace.
 */
void CChannelBase::Init(short id) 
{
	// Kanalnummer zuweisen
	m_id = id;

	// Membervars initialisieren
	m_type   = Undefined;
	m_comm   = INVALID_HANDLE_VALUE;
	m_socket = INVALID_SOCKET;
	m_value  = 0;
	m_to     = 0;
	memset(m_commString, 0, sizeof(m_commString));
	// Statistik
	m_transmittedSize = 0;
	m_receivedSize = 0;
#ifndef _WIN32
	memset(&m_flock, 0, sizeof(flock));
	memset(&m_termio, 0, sizeof(termio));
#endif
	m_trace  = FALSE;
	sprintf(m_traceFileName, "DFComDLL_%i.log", m_id);

#ifdef _DEBUG
	// Nur im Debugmodus standardmaessig aktiv.
	SetTrace();
#endif
}

// ##################################################################
/**
 * @brief Destruktor schliesst die eventuell noch geoeffnete Schnittstelle. 
 *
 */
CChannelBase::~CChannelBase() 
{
	// Schliessen der Schnittstelle, wenn nicht bereits.
	Close();
}


//////////////////////////////////////////////////////////////////////
// Members
//////////////////////////////////////////////////////////////////////

// ##################################################################
/**
 * @brief Verwirft anstehende Zeichen der Schnittstelle.
 */
void CChannelBase::Flush()
{
	Trace("%s(%i):INFO(3): CChannelBase::Flush\n", __FILE__, __LINE__);

	unsigned char buffer[0x100];
	int receivedCount = 0xFF;

	// eventuell anstehende Datenpakete verwerfen
	int i;
	for (i=0; (i < 5) && receivedCount; i++)
	{
		if (ReceiveData(buffer, 0xFF, receivedCount) != IV_ERROR_SUCCESS)
		{	
			Trace("%s(%i):INFO(2): CChannelBase::Flush [ Abbruch nach %i durchlufen. ]\n", __FILE__, __LINE__, i);
#ifdef _DEBUG
			int z; 
			for (z=0; z < receivedCount; z++) 
			{
				Trace("%02X ", buffer[z]);
			}
			Trace("\n");
#endif
			break;
		}
		Trace("%s(%i):INFO(2): CChannelBase::Flush [ Verworfene Anzahl Zeichen (%i Byte) ]\n", __FILE__, __LINE__, receivedCount);
#ifdef _DEBUG
		int z; 
		for (z=0; z < receivedCount; z++) 
		{
			Trace("%02X ", buffer[z]);
		}
		Trace("\n");
#endif
	}
}

// ##################################################################
/**
 * @brief Hauptlogfunktion der Kanalklassen.
 */
void CChannelBase::Trace(const char *format, ...)
{
	va_list ap;
	char szTime[14];
#ifdef _WIN32
	char szID[20];
#endif
	char buf[0x800];
	char *pos;
	FILE *fp;
	unsigned short msec = 0;

	if ( IsTrace() )
	{
		if (/*strstr(format, "INFO(1)") != NULL ||
			strstr(format, "INFO(2)") != NULL ||*/
			strstr(format, "INFO(3)") != NULL || /*
		  strstr(format, "WARNING(1)") != NULL ||
		  strstr(format, "WARNING(2)") != NULL ||*/
		  strstr(format, "WARNING(3)") != NULL /*||
		  strstr(format, "ERROR(1)") != NULL ||
		  strstr(format, "ERROR(2)") != NULL ||
		  strstr(format, "ERROR(3)") != NULL*/ )
		{
			  return;
		}

		va_start(ap, format);
		vsprintf(buf, format, ap);
		va_end(ap);
		// Uhrzeit des Logeintrages
		time_t t = time(NULL);
		tm *ltm = localtime(&t);
#ifdef _WIN32
		SYSTEMTIME st;
		GetLocalTime(&st);
		msec = st.wMilliseconds;
#endif
		sprintf(szTime, ":%02i:%02i:%02i.%03i", ltm->tm_hour, ltm->tm_min, ltm->tm_sec, msec);
		if ((pos = (char*)strstr(buf, ":INFO(")) != NULL || 
			(pos = (char*)strstr(buf, ":WARNING(")) != NULL || 
			(pos = (char*)strstr(buf, ":ERROR(")) != NULL)
		{
				// Uhrzeit einfuegen
				memmove(&buf[(pos - buf) + 13], pos, strlen(pos) + 1);
				strncpy(pos, szTime, 13);
#ifdef _WIN32
				// ThreadID einfuegen
				memset(szID, 0, sizeof(szID));
				sprintf(szID, ":(%i, %i)", GetID(), GetCurrentThreadId());
				memmove(&buf[(pos - buf) + strlen(szID)], pos, strlen(pos) + 1);
				strncpy(pos, szID, strlen(szID));
#endif
		}

#ifndef _WIN32
		if (buf[strlen(buf) - 1] == '\n')
		{
			strcpy(&buf[strlen(buf) - 1], "\r\n");
		}
#endif

		if ((fp = fopen(m_traceFileName, "a")) == (FILE*)0)
		{
			return;
		}
		fprintf(fp, buf);

		fclose(fp);

#ifdef _DEBUG
#ifdef _WIN32
		TRACE(buf);
#endif
#endif
	}
}

// ##################################################################
/**
 * @brief Oeffnen der Kommunikationsschnittstelle.
 */
int CChannelBase::Open(enumType type, char *string, int val, int to)
{
	Trace("%s(%i):INFO(3): CChannelBase::Open [ %i, 0x%08X, %i, %i ]\n", __FILE__, __LINE__, type, string, val, to);

	int res = IV_ERROR_FAIL;

	switch (type)
	{
	case Serial:
		// Mindesttimeout sicherstellen.
		if (to < 500)
		{		
			Trace("%s(%i):WARNING(1): CChannelBase::Open [ Zu kleiner Timeout (%i), wurde auf 500 gesetzt. ]\n", __FILE__, __LINE__, to);
			to = 500;
		}

		if (OpenSerial(string, val, to) == IV_ERROR_SUCCESS)
		{
			m_type = Serial;
			m_value = val;
			m_to = to;
			strcpy(m_commString, string);

			res = IV_ERROR_SUCCESS;
		}
		break;

	case Socket:
		// Mindesttimeout sicherstellen.
		if (to < 2000)
		{
			Trace("%s(%i):WARNING(1): CChannelBase::Open [ Zu kleiner Timeout (%i), wurde auf 2000 gesetzt. ]\n", __FILE__, __LINE__, to);
			to = 2000;
		}

		if (OpenSocket(string, val, to) == IV_ERROR_SUCCESS)
		{
			m_type = Socket;
			m_value = val;
			m_to = to;
			strcpy(m_commString, string);

			res = IV_ERROR_SUCCESS;
		}
		break;

	default:
		Trace("%s(%i):ERROR(1): CChannelBase::Open [ Unbekannter Type (%i) ]\n", __FILE__, __LINE__, type);
		return res;
	}

	return res;
}

// ##################################################################
/**
 * @brief Ermitteln ob die Kommunikationsschnittstelle geoeffnet ist.
 */
int CChannelBase::IsOpen()
{
	Trace("%s(%i):INFO(3): CChannelBase::IsOpen\n", __FILE__, __LINE__);

	return (GetType() != Undefined);
}

// ##################################################################
/**
 * @brief Schliessen der Kommunikationsschnittstelle.
 */
int CChannelBase::Close(void)
{
	Trace("%s(%i):INFO(3): CChannelBase::Close\n", __FILE__, __LINE__);

	switch (GetType())
	{
	case Undefined:
		return IV_ERROR_SUCCESS;

	case Socket:
		CloseSocket();
		break;

	case Serial:
		CloseSerial();
		break;

	default:
		Trace("%s(%i):ERROR(1): CChannelBase::Close [ Unbekannter Type (%i) ]\n", __FILE__, __LINE__, GetType());
		return IV_ERROR_FAIL;
	}

	m_type = Undefined;
	m_value = 0;
	m_to   = 0;
	memset(m_commString, 0, sizeof(m_commString));

	return IV_ERROR_SUCCESS;
}

// ##################################################################
/**
 * @brief Zuweisen einer vorhandenen Kommunikationsschnittstelle.
 * 
 */
int CChannelBase::AttachChannel(enumType type, HANDLE val, int timeOut)
{
	Trace("%s(%i):INFO(3): CChannelBase::AttachChannel [ %i, %i, %i ]\n", __FILE__, __LINE__, type, val, timeOut);

	switch (type)
	{
	case Serial:
		if (timeOut < 500)
		{		
			Trace("%s(%i):WARNING(1): CChannelBase::AttachChannel [ Zu kleiner Timeout (%i), wurde auf 500 gesetzt. ]\n", __FILE__, __LINE__, timeOut);
			timeOut = 500;
		}
		m_type = Serial;
		m_comm = (HANDLE)val;
		m_to   = timeOut;
		break;

	case Socket:
		if (timeOut < 2000)
		{		
			Trace("%s(%i):WARNING(1): CChannelBase::AttachChannel [ Zu kleiner Timeout (%i), wurde auf 2000 gesetzt. ]\n", __FILE__, __LINE__, timeOut);
			timeOut = 2000;
		}
		m_type   = Socket;
		m_socket = (SOCKET)val;
		m_to     = timeOut;
		break;

	default:
		Trace("%s(%i):ERROR(1): CChannelBase::AttachChannel [ Unbekannter Type (%i) ]\n", __FILE__, __LINE__, type);
		return IV_ERROR_FAIL;
	}

	// Kanal Zugewiesen
	strcpy(m_commString, "ATTACH");

	return IV_ERROR_SUCCESS;
}

// ##################################################################
/**
 * @brief Freigeben der aktuellen Kommunikationsschnittstelle.
 */
int CChannelBase::ReleaseChannel(void)
{
	Trace("%s(%i):INFO(3): CChannelBase::ReleaseChannel\n", __FILE__, __LINE__);

	if (GetType() != Undefined)
	{
		m_type = Undefined;
		m_value = 0;
		m_to   = 0;
		m_comm = INVALID_HANDLE_VALUE;
		m_socket = INVALID_SOCKET;

		// Kommunikationsstring loeschen
		memset(m_commString, 0, sizeof(m_commString));

		return IV_ERROR_SUCCESS;
	}

	Trace("%s(%i):ERROR(1): CChannelBase::ReleaseChannel [ Freigabe des Kanals nicht moeglich, da undefiniert. ]\n", __FILE__, __LINE__);

	return IV_ERROR_FAIL;
}

// ##################################################################
/**
 * @brief Oeffnen einer seriellen Kommunikationsschnittstelle.
 */
int CChannelBase::OpenSerial(char *comm, int baud, int to)
{
	Trace("%s(%i):INFO(3): CChannelBase::OpenSerial [ 0x%08X, %i, %i ]\n", __FILE__, __LINE__, comm, baud, to);

	int rc;

	if (GetType() != Undefined)
	{
		// Kanal bereits geoeffnet
		Trace("%s(%i):ERROR(1): CChannelBase::OpenSerial [ Kanal bereits geoeffnet. ]\n", __FILE__, __LINE__);
		return IV_ERROR_FAIL;
	}

#ifdef _WIN32

	if ((m_comm = CreateFileA(
				comm,						// pointer to name of the file 
				(GENERIC_READ | GENERIC_WRITE),	// access (read-write) mode 
				0,								// share mode 
				NULL,							// pointer to security descriptor 
				OPEN_EXISTING,					// how to create 
				0,								// file attributes 
				0 								// handle to file with attributes to copy  
				)) == INVALID_HANDLE_VALUE)
	{ 
		Trace("%s(%i):ERROR(1): CChannelBase::OpenSerial [ CreateFile fail (%i) ]\n", __FILE__, __LINE__, GetLastError());
		return IV_ERROR_FAIL;
	}
#else
	errno = 0;

	if ((rc = open(comm, O_RDWR|O_NDELAY)) < 0)
	{
		Trace("%s(%i):ERROR(1): CChannelBase::OpenSerial [ open fail ]\n", __FILE__, __LINE__);
		return IV_ERROR_FAIL;
	}

	m_comm = rc;

	if ((rc = fcntl(m_comm, F_SETFL, O_RDWR)) < 0)
	{
		Trace("%s(%i):ERROR(1): CChannelBase::OpenSerial [ fcntl fail ]\n", __FILE__, __LINE__);
		return IV_ERROR_FAIL;
	}

	m_flock.l_type    = F_WRLCK;  // lock stream for read and write
	m_flock.l_whence  = 0;        // lock whole file -> 0,0,0
	m_flock.l_start   = 0;
	m_flock.l_len     = 0;

	if ((rc = fcntl(m_comm, F_SETLK, &m_flock)) < 0)
	{
		Trace("%s(%i):WARINING(1): CChannelBase::OpenSerial [ fcntl (file locking) fail ]\n", __FILE__, __LINE__);
		//return IV_ERROR_FAIL;
	}
#endif

	if ((rc = ChangeStateSerial(m_comm, baud, to)) != IV_ERROR_SUCCESS)
	{
		return IV_ERROR_FAIL;
	}

	return IV_ERROR_SUCCESS;
}

// ##################################################################
/**
 * @brief Aendern der aktuellen Parameter zur seriellen Kommunikationsschnittstelle.
 */
int CChannelBase::ChangeStateSerial(HANDLE handle, int baud, int to)
{
	Trace("%s(%i):INFO(3): CChannelBase::ChangeStateSerial [ %i, %i, %i ]\n", __FILE__, __LINE__, handle, baud, to);

#ifdef _WIN32
	DCB          dcb;
	COMMTIMEOUTS ctout;
	DWORD        cerr = 0;

	memset(&dcb, 0, sizeof(dcb));
	dcb.DCBlength = sizeof(dcb);
	dcb.fBinary   = 1;

	dcb.BaudRate  = baud;
	dcb.ByteSize  = 8;
	dcb.Parity    = NOPARITY;
	dcb.StopBits  = ONESTOPBIT;

	if (!SetCommState(handle, &dcb))
	{
		Trace("%s(%i):ERROR(1): CChannelBase::ChangeStateSerial [ SetCommState fail ]\n", __FILE__, __LINE__);
		CloseHandle(handle);
		return IV_ERROR_FAIL;
	}

	// Handelt es sich um die eigene Schnittstelle?
	if (handle == m_comm)
	{
		// Akutuelle Baudrate anpassen.
		m_value = baud;
	}

	ctout.ReadIntervalTimeout         = 0;
	ctout.ReadTotalTimeoutMultiplier  = 0;
	ctout.ReadTotalTimeoutConstant    = (DWORD)to;
	ctout.WriteTotalTimeoutConstant   = 0;
	ctout.WriteTotalTimeoutMultiplier = 0;

	if (!SetCommTimeouts(handle, &ctout))
	{
		Trace("%s(%i):ERROR(1): CChannelBase::ChangeStateSerial [ SetCommTimeouts fail ]\n", __FILE__, __LINE__);
		CloseHandle(handle);
		return IV_ERROR_FAIL;
	}

	// Handelt es sich um die eigene Schnittstelle?
	if (handle == m_comm)
	{
		// Aktuellen Timeout anpassen
		m_to = to;
	}

	// unsolved: fuer modems + RS485 notwendig !?!
	if (!EscapeCommFunction(handle, SETRTS))
	{
		Trace("%s(%i):ERROR(1): CChannelBase::ChangeStateSerial [ EscapeCommFunction fail ]\n", __FILE__, __LINE__);
		CloseHandle(handle);
		return IV_ERROR_FAIL;
	}
	if (!EscapeCommFunction(handle, SETDTR))
	{
		Trace("%s(%i):ERROR(1): CChannelBase::ChangeStateSerial [ EscapeCommFunction fail ]\n", __FILE__, __LINE__);
		CloseHandle(handle);
		return IV_ERROR_FAIL;
	}

	if (!ClearCommError(handle, &cerr, NULL))
	{
		Trace("%s(%i):ERROR(1): CChannelBase::ChangeStateSerial [ ClearCommError fail ]\n", __FILE__, __LINE__);
		CloseHandle(handle);
		return IV_ERROR_FAIL;
	}

#else
	int rc;
	int baudmode, charmode, paricheck, stopmode;
	int termio_cm;

	switch(baud) {
	case 9600:  baudmode = B9600;  break;
	case 19200:	baudmode = B19200; break;
	case 38400:	baudmode = B38400; break;
	default: return IV_ERROR_FAIL;
	}

	charmode = CS8;
	paricheck = 0;
	stopmode = 0;

	m_termio.c_iflag = 0;
	m_termio.c_oflag = 0;
	m_termio.c_cflag = 0;
	m_termio.c_lflag = 0;
	m_termio.c_line  = 0;
	(void) memset((void *)m_termio.c_cc, 0, NCCS);
   
	m_termio.c_cflag |= HUPCL;        // hangup line when close/exit(DTR of)
	m_termio.c_cflag |= CLOCAL;       // local/direct connection
   
	m_termio.c_cflag |= baudmode;     // set baudrate
	m_termio.c_cflag |= charmode;     // set databits
	m_termio.c_cflag |= paricheck;    // set parity
	m_termio.c_cflag |= stopmode;     // set stopbits

	m_termio.c_cflag |= CREAD;        // enable input queue

	m_termio.c_cc[VMIN]  = 1;         // read waiting for 1 char
   
	if ((rc = ioctl(handle, TCSETA, &m_termio)) < 0)
	{
		Trace("%s(%i):ERROR(1): CChannelBase::ChangeStateSerial [ ioctl fail ]\n", __FILE__, __LINE__);
		return IV_ERROR_FAIL;
	}

	if ((rc = ioctl(handle, TCFLSH, TCIOFLUSH)) < 0)
	{
		Trace("%s(%i):ERROR(1): CChannelBase::ChangeStateSerial [ ioctl fail ]\n", __FILE__, __LINE__);
		return IV_ERROR_FAIL;
	}

	if ((rc = ioctl(handle, TIOCMGET, &termio_cm)) < 0)
	{
		Trace("%s(%i):ERROR(1): CChannelBase::ChangeStateSerial [ ioctl fail ]\n", __FILE__, __LINE__);
		return IV_ERROR_FAIL;
	}

	if (! (termio_cm & TIOCM_RTS)) termio_cm |= TIOCM_RTS;
	if (! (termio_cm & TIOCM_DTR)) termio_cm |= TIOCM_DTR;

	if ((rc = ioctl(handle, TIOCMSET, &termio_cm)) < 0)
	{
		Trace("%s(%i):ERROR(1): CChannelBase::ChangeStateSerial [ ioctl fail ]\n", __FILE__, __LINE__);
		return IV_ERROR_FAIL;
	}

	// Handelt es sich um die eigene Schnittstelle?
	if (handle == m_comm)
	{
		// Akutuelle Baudrate und Timeout anpassen.
		m_value = baud;
		m_to = to;
	}
#endif

	return IV_ERROR_SUCCESS;
}

// ##################################################################
/**
 * @brief Schliessen der seriellen Kommunikationsschnittstelle.
 */
int CChannelBase::CloseSerial(void)
{
	Trace("%s(%i):INFO(3): CChannelBase::CloseSerial\n", __FILE__, __LINE__);

#ifdef _WIN32
	if (!EscapeCommFunction(m_comm, CLRRTS))
	{
		Trace("%s(%i):ERROR(1): CChannelBase::CloseSerial [ EscapeCommFunction fail ]\n", __FILE__, __LINE__);
		CloseHandle(m_comm);
		return IV_ERROR_FAIL;
	}
	if (!EscapeCommFunction(m_comm, CLRDTR))
	{
		Trace("%s(%i):ERROR(1): CChannelBase::CloseSerial [ EscapeCommFunction fail ]\n", __FILE__, __LINE__);
		CloseHandle(m_comm);
		return IV_ERROR_FAIL;
	}

	CloseHandle(m_comm);
#else
	int rc;
	int termio_cm;

	if ((rc = ioctl(m_comm, TIOCMGET, &termio_cm)) < 0)
	{
		Trace("%s(%i):ERROR(1): CChannelBase::CloseSerial [ ioctl fail ]\n", __FILE__, __LINE__);
		return IV_ERROR_FAIL;
	}

	if ((termio_cm & TIOCM_RTS)) termio_cm &= ~TIOCM_RTS;
	if ((termio_cm & TIOCM_DTR)) termio_cm &= ~TIOCM_DTR;

	if ((rc = ioctl(m_comm, TIOCMSET, &termio_cm)) < 0)
	{
		Trace("%s(%i):ERROR(1): CChannelBase::CloseSerial [ ioctl fail ]\n", __FILE__, __LINE__);
		return IV_ERROR_FAIL;
	}

	close(m_comm);
#endif

	m_comm = INVALID_HANDLE_VALUE;

	return IV_ERROR_SUCCESS;
}

// ##################################################################
/**
 * @brief Lesen aus der seriellen Kommunikationsschnittstelle.
 */
int CChannelBase::ReceiveSerial(void *buffer, int receiveCount, int &receivedCount, int receiveTrials)
{
	Trace("%s(%i):INFO(3): CChannelBase::ReceiveSerial [ 0x%08X, %i, %i, %i, %i ]\n", __FILE__, __LINE__, buffer, receiveCount, receivedCount, receiveTrials);

	unsigned int res;
	unsigned char trials = 0;
	receivedCount = 0;
#ifndef _WIN32
	struct timeval seltime;
	fd_set readfds;
	errno = 0;
#endif

	// Daten emfpangen
	for (; (receiveCount != receivedCount) && (trials < receiveTrials); trials++)
	{
		res = 0;
#ifdef _WIN32
		if (!ReadFile(m_comm, (unsigned char*)buffer + receivedCount, receiveCount - receivedCount, (unsigned long*)&res, NULL))
		{
			Trace("%s(%i):ERROR(1): CChannelBase::ReceiveSerial [ trial: %i, receiveCount: %i, receivedCount: %i ]\n", __FILE__, __LINE__, trials, receiveCount, receivedCount);
			return -IV_ERROR_RECEIVE;
		}
#else
		FD_ZERO((fd_set FAR *) &readfds);
		FD_SET(m_comm, (fd_set FAR *) &readfds);
		seltime.tv_sec  = m_to / 1000;
		seltime.tv_usec = (m_to % 1000) * 1000;

		if ((res = select(FD_SETSIZE,
			(fd_set FAR *)&readfds, (fd_set FAR *) 0, (fd_set FAR *) 0,
			(struct timeval *)&seltime)) < 0)
		{
			Trace("%s(%i):ERROR(1): CChannelBase::ReceiveSerial [ trial: %i, receiveCount: %i, receivedCount: %i ]\n", __FILE__, __LINE__, trials, receiveCount, receivedCount);
			return -IV_ERROR_RECEIVE;
		}

		// Liegen Daten an?
		if (res)
		{ 
			if ((res = read(m_comm, (unsigned char*)buffer + receivedCount, receiveCount - receivedCount)) < 0)
			{
				Trace("%s(%i):ERROR(1): CChannelBase::ReceiveSerial [ trial: %i, receiveCount: %i, receivedCount: %i ]\n", __FILE__, __LINE__, trials, receiveCount, receivedCount);
				return -IV_ERROR_RECEIVE;
			}
		}
#endif
		if (res == 0)
		{
			Trace("%s(%i):WARNING(1): CChannelBase::ReceiveSerial [ Leseversuch ohne Daten. ]\n", __FILE__, __LINE__);
			break;
		}
		else if (res < (unsigned int)(receiveCount - receivedCount))
		{
			Trace("%s(%i):WARNING(1): CChannelBase::ReceiveSerial [ Nachlesen erforderlich. trial: %i, waren %i, gefordert %i, bereits %i ]\n", __FILE__, __LINE__, trials, res, receiveCount - receivedCount, receivedCount);
		}

		receivedCount += res;
		// Gesamtanzahl gelesener Bytes fuer Statistik halten
		m_receivedSize += res;
	}
	
	// Wurde gewuenschte Anzahl nicht empfangen?
	if (receiveCount != receivedCount)
	{
		Trace("%s(%i):ERROR(1): CChannelBase::ReceiveSerial [ trial: %i, receiveCount: %i, receivedCount: %i ]\n", __FILE__, __LINE__, trials, receiveCount, receivedCount);
		return -IV_ERROR_RECEIVE;
	}

	return IV_ERROR_SUCCESS;
}

// ##################################################################
/**
 * @brief Schreiben in die serielle Kommunikationsschnittstelle.
 */
int CChannelBase::TransmitSerial(const void *buffer, int transmitCount, int &transmittedCount, int transmitTrials)
{
	Trace("%s(%i):INFO(3): CChannelBase::TransmitSerial [ 0x%08X, %i, %i, %i ]\n", __FILE__, __LINE__, buffer, transmitCount, transmittedCount, transmitTrials);

#ifndef _WIN32
	errno = 0;
#endif

	int res = IV_ERROR_SUCCESS;
	unsigned char trials = 0;
	transmittedCount = 0;

	for (; (transmitCount != transmittedCount) && (trials < transmitTrials); trials++)
	{
#ifdef _WIN32
		if (!WriteFile(m_comm, (unsigned char*)buffer + transmittedCount, transmitCount - transmittedCount, (unsigned long*)&res, NULL))
		{
#else
		if ((res = write(m_comm, (unsigned char*)buffer + transmittedCount, transmitCount - transmittedCount)) < 0)
		{
#endif
			Trace("%s(%i):ERROR(1): CChannelBase::TransmitSerial [ trial: %i, transmitCount: %i, transmittedCount: %i ]\n", __FILE__, __LINE__, trials, transmitCount, transmittedCount);
			return -IV_ERROR_TRANSMIT;
		}

		// Wurden alle Zeichen versendet?
		if (res == 0)
		{
			Trace("%s(%i):WARNING(1): CChannelBase::TransmitSerial [ Es wurden keine Daten gesendet. ]\n", __FILE__, __LINE__);
			break;

		}
		else if (res < (transmitCount - transmittedCount))
		{		
			Trace("%s(%i):WARNING(1): CChannelBase::TransmitSerial [ Nachsenden erforderlich. ] (trial: %i, waren %i, gefordert %i, bereits %i)\n", __FILE__, __LINE__, trials, res, transmitCount - transmittedCount, transmittedCount);
		}

		transmittedCount += res;
		// Gesamtanzahl geschriebener Bytes fuer Statistik halten
		m_transmittedSize += res;
	}

	// Wurden nicht alle Zeichen versendet?
	if (transmitCount != transmittedCount)
	{
		Trace("%s(%i):ERROR(1): CChannelBase::TransmitSerial [ trial: %i, transmitCount: %i, transmittedCount: %i ]\n", __FILE__, __LINE__, trials, transmitCount, transmittedCount);
		return -IV_ERROR_TRANSMIT;
	}

	return IV_ERROR_SUCCESS;
}

// ##################################################################
/**
 * @brief Oeffnen einer TCP/IP Kommunikationsschnittstelle.
 */
int CChannelBase::OpenSocket(char *remote, int port, int to)
{
	Trace("%s(%i):INFO(3): CChannelBase::OpenSocket [ 0x%08X, %i, %i ]\n", __FILE__, __LINE__, remote, port, to);

	SOCKADDR_IN sin;
	memset(&sin, 0, sizeof(sin));
   	sin.sin_family = AF_INET;
	sin.sin_port = htons(port);
	sin.sin_addr.s_addr = inet_addr(remote);

	if (GetType() != Undefined)
	{
		// Kanal bereits belegt
		return IV_ERROR_FAIL;
	}

	if (sin.sin_addr.s_addr == INADDR_NONE)
	{ // Adresse ueber DNS aufloesen
		HOSTENT *he;
		he = gethostbyname(remote);
		if (!he)
		{
			Trace("%s(%i):ERROR(1): CChannelBase::OpenSocket [ gethostbyname fail (%i) ]\n", __FILE__, __LINE__, WSAGetLastError());
			return IV_ERROR_FAIL;
		}

		sin.sin_addr.s_addr = *((unsigned int*)he->h_addr);
	}

	if ((m_socket = socket(AF_INET, SOCK_STREAM, 0)) == INVALID_SOCKET)
	{
		Trace("%s(%i):ERROR(1): CChannelBase::OpenSocket [ socket fail (%i) ]\n", __FILE__, __LINE__, WSAGetLastError());
		return IV_ERROR_FAIL;
	}
      
	if (connect(m_socket,(SOCKADDR*)&sin,sizeof(sin)) == SOCKET_ERROR)
	{
		Trace("%s(%i):ERROR(1): CChannelBase::OpenSocket [ connect fail (%i) ]\n", __FILE__, __LINE__, WSAGetLastError());
		CloseSocket();
		return IV_ERROR_FAIL;
	}

	return IV_ERROR_SUCCESS;
}

// ##################################################################
/**
 * @brief Schliessen der TCP/IP Kommunikationsschnittstelle.
 */
int CChannelBase::CloseSocket(void)
{
	Trace("%s(%i):INFO(3): CChannelBase::CloseSocket\n", __FILE__, __LINE__);

#ifdef _WIN32
	if (closesocket(m_socket) == SOCKET_ERROR)
	{
		Trace("%s(%i):ERROR(1): CChannelBase::CloseSocket [ closesocket fail (%i) ]\n", __FILE__, __LINE__, WSAGetLastError());
		return IV_ERROR_FAIL;
	}
#else
	if (shutdown(m_socket, 2) == SOCKET_ERROR)
	{
		Trace("%s(%i):ERROR(1): CChannelBase::CloseSocket [ shutdown fail (%i) ]\n", __FILE__, __LINE__, WSAGetLastError());
		return IV_ERROR_FAIL;
	}
	if (close(m_socket) == SOCKET_ERROR)
	{
		Trace("%s(%i):ERROR(1): CChannelBase::CloseSocket [ close fail (%i) ]\n", __FILE__, __LINE__, WSAGetLastError());
		return IV_ERROR_FAIL;
	}
#endif

	m_socket = INVALID_SOCKET;

	return IV_ERROR_SUCCESS;
}

// ##################################################################
/**
 * @brief Lesen aus der TCP/IP Kommunikationsschnittstelle.
 */
int CChannelBase::ReceiveSocket(void *buffer, int receiveCount, int &receivedCount, int receiveTrials)
{
	Trace("%s(%i):INFO(3): CChannelBase::ReceiveSocket [ 0x%08X, %i, %i, %i ]\n", __FILE__, __LINE__, buffer, receiveCount, receivedCount, receiveTrials);

#ifdef _WIN32
#define CONST const
#else
#define CONST
#endif
	struct timeval seltime;
	fd_set readfds;

	int res;
	unsigned char trials = 0;
	unsigned char lTrial = receiveCount;

	receivedCount = 0;

	// Daten emfpangen
	for (; (receiveCount != receivedCount) && (trials < receiveTrials); trials++)
	{
		FD_ZERO((fd_set FAR *) &readfds);
		FD_SET(m_socket, (fd_set FAR *) &readfds);
		seltime.tv_sec  = m_to / 1000;
		seltime.tv_usec = (m_to % 1000) * 1000;

		if ((res = select(FD_SETSIZE,
			(fd_set FAR *)&readfds, (fd_set FAR *) 0, (fd_set FAR *) 0,
			(CONST struct timeval FAR *)&seltime)) < 0)
		{
			Trace("%s(%i):ERROR(1): CChannelBase::ReceiveSocket(%i) [ trial: %i, receiveCount: %i, receivedCount: %i, error: %i ]\n", __FILE__, __LINE__, GetID(), trials, receiveCount, receivedCount, WSAGetLastError());
			return -IV_ERROR_RECEIVE;
		}

		// Liegen Daten an?
		if (res)
		{			
			if ((res = recv(m_socket, (char*)buffer + receivedCount, receiveCount - receivedCount, 0)) == SOCKET_ERROR)
			{
#ifdef _WIN32
				if (WSAECONNRESET == WSAGetLastError() || WSAECONNABORTED == WSAGetLastError())
				{
					Trace("%s(%i):ERROR(1): CChannelBase::ReceiveSocket(%i) [ recv error: %i ]\n", __FILE__, __LINE__, GetID(), WSAGetLastError());
					return -IV_ERROR_CHANNEL_OPEN_REMOTE;
				}
#endif
				Trace("%s(%i):ERROR(1): CChannelBase::ReceiveSocket(%i) [ trial: %i, receiveCount: %i, receivedCount: %i, error: %i ]\n", __FILE__, __LINE__, GetID(), trials, receiveCount, receivedCount, WSAGetLastError());
				return -IV_ERROR_RECEIVE;
			}
			else if (res == 0)
			{
				// Wenn Daten gemeldet werden,
				// die Anzahl aber 0 ist, so wurde die Verbindung beendet.

				// Verbindung wurde von Gegenseite geschlossen.
				Trace("%s(%i):ERROR(1): CChannelBase::ReceiveSocket(%i) [ trial: %i, receiveCount: %i, receivedCount: %i ]\n", __FILE__, __LINE__, GetID(), trials, receiveCount, receivedCount);	
				return -IV_ERROR_CHANNEL_OPEN_REMOTE;
			}
			else if (res < (receiveCount - receivedCount))
			{
				Trace("%s(%i):WARNING(1): CChannelBase::ReceiveSocket(%i) [ Nachlesen erforderlich. trial: %i, waren %i, gefordert %i, bereits %i ]\n", __FILE__, __LINE__, GetID(), trials, res, receiveCount - receivedCount, receivedCount);
				// Fuer Kommunikation ueber Com-Server mit Mehrfachdocking viele Teilpackete
				if (lTrial)
				{
					trials--;
				}
				else
				{
					lTrial = (receiveCount - receivedCount);
				}
				lTrial--;
			}
		}
		else
		{
			Trace("%s(%i):WARNING(1): CChannelBase::ReceiveSocket(%i) [ Es wurden keine Daten gelesen. ]\n", __FILE__, __LINE__, GetID());
			break;
		}

		receivedCount += res;
		// Gesamtanzahl gelesener Bytes fuer Statistik halten
		m_receivedSize += res;
	}

	// Wurde gewuenschte Anzahl nicht empfangen?
	if (receiveCount != receivedCount)
	{
		Trace("%s(%i):ERROR(1): CChannelBase::ReceiveSocket(%i) [ trial: %i, receiveCount: %i, receivedCount: %i ]\n", __FILE__, __LINE__, GetID(), trials, receiveCount, receivedCount);	
		return -IV_ERROR_RECEIVE;
	}

	return IV_ERROR_SUCCESS;
}

// ##################################################################
/**
 * @brief Schreiben in die TCP/IP Kommunikationsschnittstelle.
 */
int CChannelBase::TransmitSocket(const void *buffer, int transmitCount, int &transmittedCount, int transmitTrials)
{
	Trace("%s(%i):INFO(3): CChannelBase::TransmitSocket [ 0x%08X, %i, %i, %i ]\n", __FILE__, __LINE__, buffer, transmitCount, transmittedCount, transmitTrials);

	int res;
	unsigned char trials = 0;

	transmittedCount = 0;

	for (; (transmitCount != transmittedCount) && (trials < transmitTrials); trials++)
	{
		// Daten senden
		if ((res = send(m_socket, (char*)buffer + transmittedCount, transmitCount - transmittedCount, 0)) == SOCKET_ERROR)
		{
#ifdef _WIN32
			if (WSAECONNRESET == WSAGetLastError() || WSAECONNABORTED == WSAGetLastError())
			{
				Trace("%s(%i):ERROR(1): CChannelBase::TransmitSocket(%i) [ send error: %i ]\n", __FILE__, __LINE__, GetID(), WSAGetLastError());
				return -IV_ERROR_CHANNEL_OPEN_REMOTE;
			}
#endif
			Trace("%s(%i):ERROR(1): CChannelBase::TransmitSocket(%i) [ trial: %i, transmitCount %i, transmittedCount: %i, Error: %i ]\n", __FILE__, __LINE__, GetID(), trials, transmitCount, transmittedCount, WSAGetLastError());
			return -IV_ERROR_TRANSMIT;
		}
		
		// Wurden alle Daten versendet?
		if (res == 0)
		{
			Trace("%s(%i):WARNING(1): CChannelBase::TransmitSocket(%i) [ Es wurden keine Daten gesendet. ]\n", __FILE__, __LINE__, GetID());
			break;
		}
		else if (res < (transmitCount - transmittedCount))
		{		
			Trace("%s(%i):WARNING(1): CChannelBase::TransmitSocket(%i) [ Nachsenden erforderlich. trial: %i, waren %i, gefordert %i, bereits %i ]\n", __FILE__, __LINE__, GetID(), trials, res, transmitCount - transmittedCount, transmittedCount);
		}

		transmittedCount += res;
		// Gesamtanzahl geschriebener Bytes fuer Statistik halten
		m_transmittedSize += res;
	}

	// Wurden nicht alle Daten im vorgegebenen Rahmen versendet?
	if (transmitCount != transmittedCount)
	{
		Trace("%s(%i):ERROR(1): CChannelBase::TransmitSocket(%i) [ trial: %i, transmitCount %i, transmittedCount: %i ]\n", __FILE__, __LINE__, GetID(), trials, transmitCount, transmittedCount);
		return -IV_ERROR_TRANSMIT;
	}

	// Ausfuehrung ist erfolgt.
	return IV_ERROR_SUCCESS;
}

// ##################################################################
/**
 * @brief Lesen aus der Kommunikationsschnittstelle.
 */
int CChannelBase::ReceiveData(void *buffer, int receiveCount, int &receivedCount, int receiveTrials)
{
	Trace("%s(%i):INFO(3): CChannelBase::ReceiveData [ 0x%08X, %i, %i, %i ]\n", __FILE__, __LINE__, buffer, receiveCount, receivedCount, receiveTrials);

	int res;
	switch (GetType())
	{
	case Serial:
		res = ReceiveSerial(buffer, receiveCount, receivedCount, receiveTrials);
		break;

	case Socket:
		res = ReceiveSocket(buffer, receiveCount, receivedCount, receiveTrials);
		break;

	default:
		Trace("%s(%i):ERROR(1): CChannelBase::ReceiveData [ Unbekannter Type (%i) ]\n", __FILE__, __LINE__, GetType());
		return -IV_ERROR_TRANSMIT;
	}

	return res;
}

// ##################################################################
/**
 * @brief Schreiben in die Kommunikationsschnittstelle.
 */
int CChannelBase::TransmitData(const void *buffer, int transmitCount, int &transmittedCount, int transmitTrials)
{
	Trace("%s(%i):INFO(3): CChannelBase::TransmitData [ 0x%08X, %i, %i, %i ]\n", __FILE__, __LINE__, buffer, transmitCount, transmittedCount, transmitTrials);

	int res;
	switch (GetType())
	{
	case Serial:
		res = TransmitSerial(buffer, transmitCount, transmittedCount, transmitTrials);
		break;

	case Socket:
		res = TransmitSocket(buffer, transmitCount, transmittedCount, transmitTrials);
		break;

	default:
		Trace("%s(%i):ERROR(1): CChannelBase::TransmitData [ Unbekannter Type (%i) ]\n", __FILE__, __LINE__, GetType());
		return -IV_ERROR_TRANSMIT;
	}

	return res;
}

