/**
 * @file  AEChannel.cpp
 * @brief Implementierung der Klasse CAEChannel.
 */
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "AEChannel.h"
//#include "ChannelTBProtocol.h"

#include "../Common/eeprom.h"
#include "../Common/crc.h"
#include <stdlib.h>
#include <time.h>
#ifdef _WIN32
#include <io.h>
#else
#include <sys/uio.h>
#endif
#include "../File/DFFileStudioXML.h"
#include "../File/DFFileTouchKeys.h"
#include "../File/DFFileDisplayColors.h"
// Versionspruefung
#include "../Common/DFComUtils.h"

#ifdef _WIN32

#define S_ISREG(m) ((m & _S_IFMT) == _S_IFREG)
#define S_ISDIR(m) ((m & _S_IFMT) == _S_IFDIR)

#endif

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Konstruktion/Destruktion
//////////////////////////////////////////////////////////////////////

CAEChannel::CAEChannel(short channelID) : CAEBuffer(channelID), CDFLog(channelID)
{
	// Kanalnummer setzen
	SetChannelID(channelID);

	// Datensatzpuffer loeschen
	for (int i=0; i < MAX_MB_PAKET; i++) 
	{
		m_pucRecordBuf[i] = 0;
	}

	m_iRecordLength = 0;
	m_iRecordError  = 0;

	// Wiederholungsversuche bei CheckDevice
	m_iDevicePollRetry = 0;

	// Protokollzustaende der angesprochenen Geraete
	memset(m_protocolVersion, 0, sizeof(m_protocolVersion));

	// Bevorzugt AEIII+ pruefen
	m_preferAEIII = 0;

	// Textdaten als Default uebertragen
	m_loadAsDefault = 0;
	// Trace aus Log aktivieren
	SetTraceFileName(m_szFileNameTrace);

	// Info zum binden des Kanals an ein bestimmtes Geraet.
	memset(&m_bindInfo, 0, sizeof(m_bindInfo));
	m_bindInfo.typeDevice = -1; // 0 wuerde PZE-MasterIV sein. -1 ist kein Geraet.

	m_channel4.SetTrace(m_bTrace);
}

CAEChannel::~CAEChannel()
{

}

void CAEChannel::Reset()
{
	// Puffer reset
	CAEBuffer::Reset();
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 */
int CAEChannel::OpenAEIII(int iType, char *szComString, int iComValue, int iComTimeOut)
{
	// Bevorzugt AEIII+ pruefen
	m_preferAEIII = 1;

	return Open(IV_DEFAULT_DEVICEID, iType, szComString, iComValue, iComTimeOut);
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 */
int CAEChannel::Open(int deviceID, int iType, char *szComString, int iComValue, int iComTimeOut)
{
	int res = IV_ERROR_FAIL;

	// Kanal darf nicht gebunden sein.
	if (IsBind()) 
	{
		return IV_ERROR_FAIL;
	}

	// Kanal 4 ist Massgebend
	if (m_channel4.GetType() == CChannelBase::Undefined) 
	{
		if (strcmp(m_channel4.m_commString, szComString)) 
		{
			// Bestimmt die Gueltigkeit der Protokollversion

			// Alle Teilnehmer erhalten unbekanntes Protokoll
			memset(m_protocolVersion, 0, sizeof(m_protocolVersion));
      // Ggf. unbekannte Dockingstation
      m_channel4.ResetTimeboyStation();
		}

		// Je nach Typ der Verbindung die Klasse initialisieren.
		switch (iType) 
		{
		case 1: // RS232 direkt
			// Schnittstelle oeffnen
			res = m_channel4.Open(CChannelBase::Serial, szComString, iComValue, iComTimeOut);
			break;

		case 3: // TCP/IP
			// Schnittstelle oeffnen
			res = m_channel4.Open(CChannelBase::Socket, szComString, iComValue, iComTimeOut);
			break;

		default: // Unbekannter Schnittstellentyp
			res = IV_ERROR_FAIL;
		}

		// Nach Erfolg die Grundeinstellung wiederherstellen
		if (res == IV_ERROR_SUCCESS) 
		{
			SetDeviceID(deviceID);

			if (strcmp(m_channel4.m_commString, szComString)) 
			{
				// Bestimmt die Gueltigkeit der Protokollversion

				// Alle Teilnehmer erhalten unbekanntes Protokoll
				memset(m_protocolVersion, 0, sizeof(m_protocolVersion));
        // Ggf. unbekannte Dockingstation
        m_channel4.ResetTimeboyStation();
			}
		}
	} 

	return res;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 */
void CAEChannel::Close(void)
{
	// Ist zusaetzlich in Verwendung als Aktiver Kanal?
	if (m_channel4.GetActiveState() > is_free)
	{
		DirectClose();
	}

	// Kanal 4 ist Massgebend
	m_channel4.Close();

	// Bevorzugt AEIII+ pruefen
	m_preferAEIII = 0;
}

// ##################################################################
/**
 * @brief Ermitteln ob die Schnittstelle geoeffnet ist.
 */
int CAEChannel::IsOpen()
{
	// Channel4 ist Massgebend!
	return m_channel4.IsOpen();
}

// ##################################################################
/**
 * @brief Schnittstelle nach Fehler bereinigen.
 */
void CAEChannel::ResetWithError(int err) 
{
	// Gehaltene Protokollversion auswerten
	switch(m_protocolVersion[(m_channel4.GetDeviceIDRemote() == 0xFE) ? 0 : m_channel4.GetDeviceIDRemote()]) 
	{
	case 0:
	case 1:
	case 2:
	case 3: // Anstehende Daten verwerfen
		if (err == IV_ERROR_CRC_PACKET || err == IV_ERROR_CHANNEL_COMMUNICATION || err == IV_ERROR_RECEIVE) 
		{
			m_channel4.Flush();
		}
	case 4: 
		// Protocolversion zuruecksetzen
		if (err == IV_ERROR_CHANNEL_COMMUNICATION || err == IV_ERROR_RECEIVE) 
		{
			SetProtocolVersion(GetDeviceIDRemote(), 0);
      // Ggf. unbekannte Dockingstation
      m_channel4.ResetTimeboyStation();
		}
	}
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 */
short CAEChannel::SetDeviceIDRemote(short deviceIDRemote)
{
	short deviceIDPrev = m_channel4.GetDeviceIDRemote();

	m_channel1_2.SetDeviceIDRemote((unsigned char)deviceIDRemote);
	m_channel4.SetDeviceIDRemote(deviceIDRemote);

	return deviceIDPrev;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 */
short CAEChannel::GetDeviceIDRemote()
{
	return m_channel4.GetDeviceIDRemote();
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 */
short CAEChannel::SetDeviceID(short deviceID)
{
	short deviceIDPrev = m_channel4.GetDeviceID();

	m_channel1_2.SetDeviceID((unsigned char)deviceID);
	m_channel4.SetDeviceID(deviceID);

	return deviceIDPrev;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 */
short CAEChannel::GetDeviceID()
{
	return m_channel4.GetDeviceID();
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 */
int CAEChannel::ComReleaseHandle()
{
	if (m_channel4.GetType() != CChannel4::Serial)
	{
		return IV_ERROR_FAIL;
	}

	// CChannel4
	m_channel4.ReleaseChannel();

	return IV_ERROR_SUCCESS;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 */
int CAEChannel::ComUseHandle(HANDLE hCom)
{
	if (m_channel4.GetType() != CChannel4::Undefined)
	{
		return IV_ERROR_FAIL;
	}

	// CChannel4
	m_channel4.AttachChannel(CChannelBase::Serial, hCom, 800);

	return IV_ERROR_SUCCESS;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 */
HANDLE CAEChannel::GetComPort()
{
	HANDLE hCom = INVALID_HANDLE_VALUE;

	if (m_channel4.GetType() == CChannelBase::Serial)
	{
		hCom = m_channel4.m_comm;
	}
	else if (m_channel4.GetType() == CChannelBase::Socket)
	{
		hCom = (HANDLE)m_channel4.m_socket;
	}

	return hCom;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 */
int CAEChannel::SetComPort(HANDLE hCom, int iType, int iComValue, int iTimeOut)
{
	switch (iType)
	{
	case 0:
	case 2:
		// Muss undefiniert sein oder mit gewuenschtem uebereinstimmen
		if ((m_channel4.GetType() == CChannelBase::Undefined) || 
			((m_channel4.GetType() == CChannelBase::Serial) && (m_channel4.m_comm == hCom)) )
		{
			if (m_channel4.ChangeStateSerial(hCom, iComValue, iTimeOut) != IV_ERROR_SUCCESS)
			{
				return IV_ERROR_FAIL;
			}

			// CChannel4
			m_channel4.AttachChannel(CChannelBase::Serial, hCom, iTimeOut);
		}
		else
		{
			return IV_ERROR_FAIL;
		}
		break;

	case 1:
	case 3:
		// Muss undefiniert sein oder mit gewuenschtem uebereinstimmen
		if ((m_channel4.GetType() == CChannelBase::Undefined) || 
			((m_channel4.GetType() == CChannelBase::Socket) && (m_channel4.m_socket == (SOCKET)hCom)) )
		{
			// CChannel4
			m_channel4.AttachChannel(CChannelBase::Socket, hCom, iTimeOut);
		}
		else
		{
			return IV_ERROR_FAIL;
		}
		break;
	}

	return IV_ERROR_SUCCESS;
}

void CAEChannel::SetCBFunction(int (*pFunction)(void))
{
	m_channel1_2.SetCBFunction(pFunction);
	m_channel4.SetCBFunction(pFunction);
}

// Setzen des Logdateinamen.
void CAEChannel::SetLogFileName(const char *szLogFileName)
{
	m_channel4.SetTraceFileName(szLogFileName);
}	


//#########################################################
//####################### HELPER ##########################
//#########################################################

int CAEChannel::Receive(void *buffer, int receiveCount, int *receivedCount)
{
	return m_channel4.ReceiveData(buffer, receiveCount, (int&)*receivedCount);
}

int CAEChannel::Transmit(void *buffer, int transmitCount, int *transmittedCount)
{
	return m_channel4.TransmitData(buffer, transmitCount, (int&)*transmittedCount);
}

// Aufbereiten und senden eines Protokollpakets
int CAEChannel::TransmitPacket(short deviceIDRemote, unsigned char type, short module, short func, unsigned char *buf, int count, int transmitTrials)
{
	return m_channel4.TransmitPacket(deviceIDRemote, type, module, func, buf, count, transmitTrials);
}

// Empfang und aufbereiten eines Protokollpakets
int CAEChannel::ReceivePacket(short deviceID, unsigned char &type, short &module, short &func, unsigned char *buf, int &count, int receiveTrials)
{
	return m_channel4.ReceivePacket(deviceID, type, module, func, buf, count, receiveTrials);
}


// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 */
unsigned char CAEChannel::GetProtocolVersion(short deviceIDRemote)
{
	switch(m_protocolVersion[(deviceIDRemote == 0xFE) ? 0 : deviceIDRemote]) 
	{
	case 1: // Version 1 des Protokolls
	case 2: // Version 2 des MM-Protokolls
	case 3: // Mehrfachdockingstation
	case 4: // Version 4 des MM-Protokolls
		return m_protocolVersion[(deviceIDRemote == 0xFE) ? 0 : deviceIDRemote];

	default: // Ermitteln der Protokollversion
		
		// Bevorzugt auf AEIII+ pruefen?
		if (m_preferAEIII && (m_channel1_2.Request((unsigned char)deviceIDRemote) == TRUE)) 
		{ 
			// Pruefen ob Protokollversion 1 vorhanden
			return m_protocolVersion[(deviceIDRemote == 0xFE) ? 0 : deviceIDRemote] = 1;	
		} 
		else if (m_channel4.IsTimeboySlot(deviceIDRemote)) 
		{ 
			// Pruefen ob Protokollversion 4 vorhanden
			return m_protocolVersion[(deviceIDRemote == 0xFE) ? 0 : deviceIDRemote] = 4;
		} 
		else if (m_channel4.Ping(deviceIDRemote, 0, 0) >= IV_ERROR_SUCCESS) 
		{ 
			// Pruefen ob Protokollversion 4 vorhanden
			return m_protocolVersion[(deviceIDRemote == 0xFE) ? 0 : deviceIDRemote] = 4;
		} 
		else 
		{	
			// Hoher Modbustimeout aelterer Firmwarestaende uebergehen.
			Sleep(100);
			if (m_channel1_2.Request2((unsigned char)deviceIDRemote) == TRUE) 
			{ 
				// Pruefen ob Protokollversion 2 vorhanden
				return m_protocolVersion[(deviceIDRemote == 0xFE) ? 0 : deviceIDRemote] = 2;
			} 
			else 
			{
				// Hoher Modbustimeout aelterer Firmwarestaende uebergehen.
				Sleep(100);
				if (m_channel1_2.Request((unsigned char)deviceIDRemote) == TRUE) 
				{ 
					// Pruefen ob Protokollversion 1 vorhanden
					return m_protocolVersion[(deviceIDRemote == 0xFE) ? 0 : deviceIDRemote] = 1;
				}
			}
		}
	}

	return 0;
}

// ##################################################################
/**
 * @brief Beschreibung fehlt noch.
 */
unsigned char CAEChannel::SetProtocolVersion(short deviceIDRemote, unsigned char version)
{
	unsigned char old = m_protocolVersion[(deviceIDRemote == 0xFE) ? 0 : deviceIDRemote];

	m_protocolVersion[(deviceIDRemote == 0xFE) ? 0 : deviceIDRemote] = version;

	return old;
}

//////////////////////////////////////////////////////////////////////
// Members
//////////////////////////////////////////////////////////////////////

int CAEChannel::CollectData(short deviceIDRemote, int *piError)
{
	int iRes = -1;
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	// Loeschen des Datenpuffer
	if (m_pucRecordBuf[0])
	{
		int i;
		for (i=0; i < MAX_MB_PAKET; i++)
		{		
			m_pucRecordBuf[i] = 0;
		}

		m_iRecordLength = 0;
		m_iRecordError = 0;
	}

	switch ( GetProtocolVersion(GetDeviceIDRemote()) )
	{
	case 1:
	case 2:
		// Datensatz ermitteln
		iRes = m_channel1_2.ReadRecord((unsigned char)GetDeviceIDRemote(), m_pucRecordBuf, &m_iRecordLength, &m_iRecordError);
		break;

	case 4: {
		unsigned char status = 0;
		m_iRecordLength = 0;
		m_pucRecordBuf[0] = (unsigned char)GetDeviceIDRemote();

		// Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
		if ((iRes = m_channel4.SetTimeboySlotState(GetDeviceIDRemote(), 1)) == IV_ERROR_SUCCESS)
		{
			// Befehlsausfuehrung starten.
			iRes = m_channel4.GetRecord(GetDeviceIDRemote(), &m_pucRecordBuf[3], m_iRecordLength, status);

			// Versuchen die rote LED abzuschalten.
			m_channel4.SetTimeboySlotState(GetDeviceIDRemote(), 0, iRes);
		}

		iRes = ResultToError(iRes, &m_iRecordError);
		if (iRes)
		{
			// OK
			iRes = status;

			// Groesze anpassen
			m_iRecordLength += 5;
		}
		else
		{
		  iRes = -1;
		}
		break;}

	default:
		*piError = IV_ERROR_UNREACHABLE;
		return -1;
	}

	switch ( iRes )
	{
	case 0: // Keine Datensaetze vorhanden		
		memset(m_pucRecordBuf, 0, MAX_MB_PAKET);
		m_iRecordLength = 0;
		m_iRecordError = 0;
		return 0;

	case 1: // Online Datensatz
	case 2: // Wiederholter Online Datensatz
		return 2;

	case 3: // Offline Datensatz
	case 4: // Wiederholter Offline Datensatz
		return 1;

	default:
		// Fehler bei Ermittlung
		*piError = m_iRecordError;
	}

	return -1;
}

int CAEChannel::GetDatensatz(unsigned char *pucBuf, int *piLength, int *piError)
{
	int iRes = 0;

	// Wenn BusNum des letzten Datensatzpakets 0 ist dann sind 
	// keine gueltigen Daten vorhanden.
	if (m_pucRecordBuf[0] == 0)
	{
		return 0;
	}

	// Daten in Puffer uebergeben
	unsigned char ucTmp[MAX_MB_PAKET_DATA * 3 + 1];
	int i, iFail;
	// Zu grosze Datensaetze abfangen!
	if ((m_iRecordLength - 6) > MAX_MB_PAKET_DATA)
	{
		m_iRecordLength -= (m_iRecordLength - 6) - MAX_MB_PAKET_DATA;
	}
	for (i=0; i < (m_iRecordLength - 6); i++)
	{
		pucBuf[i] = m_pucRecordBuf[i + 4];
		//sprintf((char*)&ucTmp[i * 3], "%02X ", (unsigned char)m_pucRecordBuf[i + 4]);
	}
	ucTmp[i * 3] = 0;

	*piLength = m_iRecordLength - 6;

	//theApp.m_log.Append("\t%s\t%i", ucTmp, *piLength);

	switch(GetProtocolVersion(GetDeviceIDRemote()))
	{
	case 1:
	case 2:
		// Datensatz quittieren (3x -> Versuchen)
		for (i=0; (i < 3) && (iRes = m_channel1_2.QuitRecord((unsigned char)GetDeviceIDRemote(), m_pucRecordBuf,
			&m_iRecordLength, &m_iRecordError)) == -1; i++);

		// Quittierung erfolgreich?
		if (iRes)
		{	
			// Naechsten Datensatz mit Duplikattest anfordern
			for (i=0, iFail=0; (i < 3) && (iRes = m_channel1_2.ReadRecord((unsigned char)GetDeviceIDRemote(), m_pucRecordBuf, 
				&m_iRecordLength, &m_iRecordError)); i++);
		} 
		break;

	case 4: {
		unsigned char status = 0;

		// Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
		if ((iRes = m_channel4.SetTimeboySlotState(GetDeviceIDRemote(), 1)) == IV_ERROR_SUCCESS)
		{
			// Befehlsausfuehrung starten.
			// Datensatz quittieren
			iRes = m_channel4.SetRecord(GetDeviceIDRemote(), status);

			// Versuchen die rote LED abzuschalten.
			m_channel4.SetTimeboySlotState(GetDeviceIDRemote(), 0, iRes);
		}

		iRes = ResultToError(iRes, &m_iRecordError);
		if (iRes)
		{
			// OK
			// naechsten Datensatz lesen.
			unsigned char status = 0;
			m_iRecordLength = 0;
			m_pucRecordBuf[0] = (unsigned char)GetDeviceIDRemote();

			// Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
			if ((iRes = m_channel4.SetTimeboySlotState(GetDeviceIDRemote(), 1)) == IV_ERROR_SUCCESS)
			{
				// Befehlsausfuehrung starten.
				iRes = m_channel4.GetRecord(GetDeviceIDRemote(), &m_pucRecordBuf[3], m_iRecordLength, status);

				// Versuchen die rote LED abzuschalten.
				m_channel4.SetTimeboySlotState(GetDeviceIDRemote(), 0, iRes);
			}

			iRes = ResultToError(iRes, &m_iRecordError);
			if (iRes)
			{
				// OK
				iRes = status;

				// Groesze anpassen (wegen veralgemeinerter Ausfuehrung am Funktionsanfang)
				m_iRecordLength += 5;
			}
			else
			{
				iRes = -1;
			}
		}
		else
		{
			iRes = -1;
		}
		break;}

	default:
		*piError = IV_ERROR_UNREACHABLE;
		return -1;
	}
		
	switch ( iRes )
	{
	case 0: // Keine Datensaetzte vorhanden
			// Aktuell uebergebener ist letzter moeglicher, bei Quittierungsmisserfolg.
		memset(m_pucRecordBuf, 0, MAX_MB_PAKET);
		m_iRecordLength = 0;
		m_iRecordError = 0;
		return 2;

		// Online Datensaetze
	case 1: // Online Datensatz
	case 2: // Wiederholter Online Datensatz
		return 1;

		// Offline Datensaetze
	case 3: // Offline Datensatz
	case 4: // Wiederholter Offline Datensatz
		return 1;

	default:
		*piError = m_iRecordError;
	}

	return -1;
}

int CAEChannel::SystemReadRecord(int deviceIDRemote, char *buf, int *length, int *error)
{
	int iRes = -1;
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch(GetProtocolVersion(GetDeviceIDRemote()))
	{
	case 1:
	case 2: 
		*error = IV_ERROR_VERSION_FIRMWARE;
		return -1;

	case 4: {
		// OK
		// naechsten Datensatz lesen.
		unsigned char status = 0;
		m_iRecordLength = 0;
		m_pucRecordBuf[0] = (unsigned char)GetDeviceIDRemote();

		// Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
		if ((iRes = m_channel4.SetTimeboySlotState(GetDeviceIDRemote(), 1)) == IV_ERROR_SUCCESS)
		{
			// Befehlsausfuehrung starten.
			iRes = m_channel4.SystemReadRecord(GetDeviceIDRemote(), &m_pucRecordBuf[3], m_iRecordLength, status);

			// Versuchen die rote LED abzuschalten.
			m_channel4.SetTimeboySlotState(GetDeviceIDRemote(), 0, iRes);
		}

		iRes = ResultToError(iRes, &m_iRecordError);
		if (iRes)
		{
			// OK
			iRes = status;

			// Groesse anpassen
			m_iRecordLength += 5;
		}
		else
		{
			iRes = -1;
		}
		break;}

	default:
		*error = IV_ERROR_UNREACHABLE;
		return -1;
	}

	//	TRACE("( %i )\n", iRes);
	switch ( iRes )
	{
	case 0: // Keine Datensaetze vorhanden
		memset(m_pucRecordBuf, 0, MAX_MB_PAKET_SYSTEM_DATA);
		m_iRecordLength = 0;
		m_iRecordError = 0;
		return 0;

	case 1: // System Datensatz
		// Daten in Puffer uebergeben, zu grosse Datensaetze sind abfangen!
		for (int i=0; i < (m_iRecordLength - 6); i++)
		{
			buf[i] = m_pucRecordBuf[i + 4];
		}
		*length = m_iRecordLength - 6;
		return iRes;

	default:
		// Fehler bei Ermittlung
		*error = m_iRecordError;
		ResetWithError(*error);
		break;
	}

	return -1;
}

int CAEChannel::SystemQuitRecord(int deviceIDRemote, int *error)
{
	int iRes = -1;
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(GetDeviceIDRemote()))
	{
	case 1:
	case 2: 
		*error = IV_ERROR_VERSION_FIRMWARE;
		return -1;

	case 4: {
		// OK
		// naechsten Datensatz lesen.
		unsigned char status = 0;

		// Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
		if ((iRes = m_channel4.SetTimeboySlotState(GetDeviceIDRemote(), 1)) == IV_ERROR_SUCCESS)
		{
			// Befehlsausfuehrung starten.
			iRes = m_channel4.SystemQuitRecord(GetDeviceIDRemote(), status);

			// Versuchen die rote LED abzuschalten.
			m_channel4.SetTimeboySlotState(GetDeviceIDRemote(), 0, iRes);
		}

		iRes = ResultToError(iRes, &m_iRecordError);
		if (iRes)
		{
			// OK
			iRes = status;
		}
		else
		{
			iRes = -1;
		}
		break;}

	default:
		*error = IV_ERROR_UNREACHABLE;
		return -1;
	}

	//	TRACE("( %i )\n", iRes);
	switch ( iRes )
	{
	case 0: // Keine Datensaetze vorhanden
		return 0;

	case 1: // Datensatz quitiert
		return 1;

	default:
		// Fehlercode uebernehmen
		*error = m_iRecordError;
	}

	return -1;
}

int CAEChannel::SystemRestoreRecords(short deviceIDRemote, int *piError)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(GetDeviceIDRemote()))
	{
	case 1:
	case 2: 
		*piError = IV_ERROR_VERSION_FIRMWARE;
		return 0;

	case 4: { // PZEMasterIV mit Protokollversion 4
		int iRes;
		
		// Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
		if ((iRes = m_channel4.SetTimeboySlotState(GetDeviceIDRemote(), 1)) == IV_ERROR_SUCCESS)
		{
			// Befehlsausfuehrung starten.
			iRes = m_channel4.SystemRestoreRecords(GetDeviceIDRemote());

			// Versuchen die rote LED abzuschalten.
			m_channel4.SetTimeboySlotState(GetDeviceIDRemote(), 0, iRes);
		}

		return ResultToError(iRes, piError);}

	default:
		*piError = IV_ERROR_UNREACHABLE;
	}

	return 0;
}

int CAEChannel::DFReadRecord(int deviceIDRemote, char *buf, int *length, int *error)
{
	int iRes = -1;
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);
	
	switch (GetProtocolVersion(GetDeviceIDRemote()))
	{
	case 1:
	case 2:

		// Datensatz ermitteln
		iRes = m_channel1_2.ReadRecord((unsigned char)GetDeviceIDRemote(), m_pucRecordBuf, &m_iRecordLength, &m_iRecordError);
		break;

	case 4: {
		// OK
		// naechsten Datensatz lesen.
		unsigned char status = 0;
		m_iRecordLength = 0;
		m_pucRecordBuf[0] = (unsigned char)GetDeviceIDRemote();

		// Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
		if ((iRes = m_channel4.SetTimeboySlotState(GetDeviceIDRemote(), 1)) == IV_ERROR_SUCCESS)
		{
			// Befehlsausfuehrung starten.
			iRes = m_channel4.GetRecord(GetDeviceIDRemote(), &m_pucRecordBuf[3], m_iRecordLength, status);

			// Versuchen die rote LED abzuschalten.
			m_channel4.SetTimeboySlotState(GetDeviceIDRemote(), 0, iRes);
		}

		iRes = ResultToError(iRes, &m_iRecordError);
		if (iRes)
		{
			// OK
			iRes = status;

			// Groesse anpassen
			m_iRecordLength += 5;
		}
		else
		{
			iRes = -1;
		}
		break;}

	default:
		*error = IV_ERROR_UNREACHABLE;
		return -1;
	}

//	TRACE("( %i )\n", iRes);
	switch ( iRes )
	{
	case 0: // Keine Datensaetze vorhanden
		memset(m_pucRecordBuf, 0, MAX_MB_PAKET);
		m_iRecordLength = 0;
		m_iRecordError = 0;
		return 0;

	case 1: // Online Datensatz
	case 2: // Wiederholter Online Datensatz
	case 3: // Offline Datensatz
	case 4: // Wiederholter Offline Datensatz
		
		// Daten in Puffer uebergeben
		int i;
		// Zu grosse Datensaetze abfangen!
		if ((m_iRecordLength - 6) > MAX_MB_PAKET_DATA)
		{
			m_iRecordLength -= (m_iRecordLength - 6) - MAX_MB_PAKET_DATA;
		}
		for (i=0; i < (m_iRecordLength - 6); i++)
		{
			buf[i] = m_pucRecordBuf[i + 4];
		}

		*length = m_iRecordLength - 6;
		return iRes;

	default:
		// Fehler bei Ermittlung
		*error = m_iRecordError;
		ResetWithError(*error);
		break;
	}

	return -1;
}

int CAEChannel::DFQuitRecord(int deviceIDRemote, int *error)
{
	int iRes = -1;
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);
	
	switch (GetProtocolVersion(GetDeviceIDRemote()))
	{
	case 1:
	case 2:

		// Datensatz ermitteln
		iRes = m_channel1_2.QuitRecord((unsigned char)GetDeviceIDRemote(), m_pucRecordBuf, &m_iRecordLength, &m_iRecordError);
		break;

	case 4: {
		// OK
		// naechsten Datensatz lesen.
		unsigned char status = 0;

		// Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
		if ((iRes = m_channel4.SetTimeboySlotState(GetDeviceIDRemote(), 1)) == IV_ERROR_SUCCESS)
		{
			// Befehlsausfuehrung starten.
			iRes = m_channel4.SetRecord(GetDeviceIDRemote(), status);

			// Versuchen die rote LED abzuschalten.
			m_channel4.SetTimeboySlotState(GetDeviceIDRemote(), 0, iRes);
		}

		iRes = ResultToError(iRes, &m_iRecordError);
		if (iRes)
		{
			// OK
			iRes = status;
		}
		else
		{
			iRes = -1;
		}
		break;}

	default:
		*error = IV_ERROR_UNREACHABLE;
		return -1;
	}

//	TRACE("( %i )\n", iRes);
	switch ( iRes ) 
	{
	case 0: // Keine Quittung noetig	
		return 0;

	case 1: // Online Datensatzquit.
	case 2: // Wiederholte Online Datensatzquit.
	case 3: // Offline Datensatzquit.
	case 4: // Wiederholte Offline Datensatzquit.
		return iRes;

	default:
		// Fehler bei Ermittlung
		*error = m_iRecordError;
		ResetWithError(*error);
		break;
	}

	return -1;
}

// Loescht alle Datensaetze
int CAEChannel::ClearData(short deviceIDRemote, int *piError)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(GetDeviceIDRemote())) 
	{
	case 1:
	case 2:
		return m_channel1_2.ClearData((unsigned char)GetDeviceIDRemote(), piError);

	case 4: { // PZEMasterIV mit Protokollversion 4
		int iRes;
		
		// Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
		if ((iRes = m_channel4.SetTimeboySlotState(GetDeviceIDRemote(), 1)) == IV_ERROR_SUCCESS)
		{
			// Befehlsausfuehrung starten.
			iRes = m_channel4.ClearRecords(GetDeviceIDRemote());

			// Versuchen die rote LED abzuschalten.
			m_channel4.SetTimeboySlotState(GetDeviceIDRemote(), 0, iRes);
		}

		return ResultToError(iRes, piError);}

	default:
		*piError = IV_ERROR_UNREACHABLE;
	}

	return 0;
}

// Stellt den Datensatzzeiger an Ringbeginn. Alle Datensaetze wieder abrufbar.
int CAEChannel::RestoreRecords(short deviceIDRemote, int *piError)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(GetDeviceIDRemote())) {
	case 1:
	case 2: 
		*piError = IV_ERROR_VERSION_FIRMWARE;
		return 0;

	case 4: { // PZEMasterIV mit Protokollversion 4
		int iRes;

		// Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
		if ((iRes = m_channel4.SetTimeboySlotState(GetDeviceIDRemote(), 1)) == IV_ERROR_SUCCESS)
		{
			// Befehlsausfuehrung starten.
			iRes = m_channel4.RestoreRecords(GetDeviceIDRemote());

			// Versuchen die rote LED abzuschalten.
			m_channel4.SetTimeboySlotState(GetDeviceIDRemote(), 0, iRes);
		}

		return ResultToError(iRes, piError);}

	default:
		*piError = IV_ERROR_UNREACHABLE;
	}

	return 0;
}

int CAEChannel::GetDateTime(short deviceIDRemote, unsigned char *pucDateTime, int *piError)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(GetDeviceIDRemote()))
	{
	case 1:
	case 2:

		if(!m_channel1_2.ReadMem(MB_GET_TIME, 7, pucDateTime, piError)) 
		{		
			return 0;
		}
		return 1;

	case 4: {
		int iRes = IV_ERROR_SUCCESS;
		char szTime[20];

		// Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
		if ((iRes = m_channel4.SetTimeboySlotState(GetDeviceIDRemote(), 1)) == IV_ERROR_SUCCESS)
		{
			// Befehlsausfuehrung starten.
			iRes = m_channel4.GetRTC(GetDeviceIDRemote(), szTime);

			// Versuchen die rote LED abzuschalten.
			m_channel4.SetTimeboySlotState(GetDeviceIDRemote(), 0, iRes);
		}

		iRes = ResultToError(iRes, piError);
		if (iRes)
		{
			// OK
			pucDateTime[0] = (szTime[0] - 0x30) * 10 + szTime[1] - 0x30;
			pucDateTime[1] = (szTime[2] - 0x30) * 10 + szTime[3] - 0x30;

			pucDateTime[2] = (szTime[5] - 0x30) * 10 + szTime[6] - 0x30;

			pucDateTime[3] = (szTime[8] - 0x30) * 10 + szTime[9] - 0x30;

			pucDateTime[4] = (szTime[11] - 0x30) * 10 + szTime[12] - 0x30;
			pucDateTime[5] = (szTime[14] - 0x30) * 10 + szTime[15] - 0x30;
			pucDateTime[6] = (szTime[17] - 0x30) * 10 + szTime[18] - 0x30;
		}

		return iRes;}

	default:
		*piError = IV_ERROR_UNREACHABLE;
	}

	return 0;
}

int CAEChannel::SetDateTime(short deviceIDRemote, unsigned char *pucDateTime, int *piError)
{
	unsigned char ucDateTime[8];

	if (pucDateTime == NULL ||
		(pucDateTime[0] != 20) ||
		(pucDateTime[1] > 99) ||
		(pucDateTime[2] < 1) || (pucDateTime[2] > 12) || 
		(pucDateTime[3] < 1) || (pucDateTime[3] > 31) ||
		(pucDateTime[4] > 23) || (pucDateTime[5] > 59) || (pucDateTime[6] > 59))
	{	
		time_t tm_time;
		struct tm *ptm;
		tm_time = time(NULL);
		ptm = localtime(&tm_time);
		
		ucDateTime[0] = (ptm->tm_year + 1900) / 100;
		ucDateTime[1] = (ptm->tm_year + 1900) % 100;
		ucDateTime[2] = ptm->tm_mon + 1;
		ucDateTime[3] = ptm->tm_mday;
		ucDateTime[4] = ptm->tm_hour;
		ucDateTime[5] = ptm->tm_min;
		ucDateTime[6] = ptm->tm_sec;
	}
	else
	{
		memcpy(ucDateTime, pucDateTime, 7);
	}

	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	int iRes = IV_ERROR_SUCCESS;
	switch (GetProtocolVersion(GetDeviceIDRemote()))
	{
	case 1:
	case 2:

		if(!m_channel1_2.WriteMem(MB_SET_TIME, 7, ucDateTime, piError))
		{	
			return 0;
		}
		return 1;

	case 4: {
		char szTime[20];

		sprintf(szTime, "%02i%02i-%02i-%02i %02i:%02i:%02i", 
			ucDateTime[0], ucDateTime[1], ucDateTime[2], ucDateTime[3], 
			ucDateTime[4], ucDateTime[5], ucDateTime[6]);

		// Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
		if ((iRes = m_channel4.SetTimeboySlotState(GetDeviceIDRemote(), 1)) == IV_ERROR_SUCCESS)
		{
			// Befehlsausfuehrung starten.
			iRes = m_channel4.SetRTC(GetDeviceIDRemote(), szTime);

			// Versuchen die rote LED abzuschalten.
			m_channel4.SetTimeboySlotState(GetDeviceIDRemote(), 0, iRes);
		}

		return ResultToError(iRes, piError);}

	default:
		*piError = IV_ERROR_UNREACHABLE;
	}

	return 0;
}

int CAEChannel::SendInfotext(short deviceIDRemote, char *szMessage, int iLength, int *piError)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(GetDeviceIDRemote()))
	{
	case 1: {
		unsigned char Buffer[245];
		int counter;

		Buffer[0] = 0xFF;
		Buffer[1] = 0;
		Buffer[2] = 0;

		for(counter=0; counter < iLength; counter++)
		{
			Buffer[counter + 3] = szMessage[counter];
		}

		if(!m_channel1_2.WriteMem(INFOTEXT_MEMORY, iLength + 3, &Buffer[0], piError))
		{			
			return 0;
		}

		return 1;}

	case 2:
		*piError = IV_ERROR_FUNCTION_UNAVAILABLE;
		return 0;

	case 4: {
		int iRes = IV_ERROR_SUCCESS;

		// Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
		if ((iRes = m_channel4.SetTimeboySlotState(GetDeviceIDRemote(), 1)) == IV_ERROR_SUCCESS)
		{
			// Befehlsausfuehrung starten.
			iRes = m_channel4.ShowInfo(GetDeviceIDRemote(), 0, 0, 0, 0, 0, szMessage);

			// Versuchen die rote LED abzuschalten.
			m_channel4.SetTimeboySlotState(GetDeviceIDRemote(), 0, iRes);
		}

		return ResultToError(iRes, piError);}

	default:
		*piError = IV_ERROR_UNREACHABLE;
	}

	return 0;
}

int CAEChannel::SendMessage(short deviceIDRemote, unsigned char ucDelayTime, 
							 unsigned char ucAction,  unsigned char ucAudio, 
							 char *szMessage, int iLength, int *piError)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(GetDeviceIDRemote()))
	{
	case 1:
	case 2: {
		unsigned char Buffer[245];
		int counter;

		Buffer[0] = ucDelayTime;
		Buffer[1] = ucAction;
		Buffer[2] = ucAudio;

		for(counter=0; (counter < iLength) && (counter < (int)(sizeof(Buffer) - 3)); counter++)
		{			
			Buffer[3 + counter] = szMessage[counter];
		}

		if(!m_channel1_2.WriteMem(DISPLAY_MEMORY, counter + 3, &Buffer[0], piError))
		{			
			return 0;
		}
		return 1;}

	case 4: {
		int iRes = IV_ERROR_SUCCESS;

		// Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
		if ((iRes = m_channel4.SetTimeboySlotState(GetDeviceIDRemote(), 1)) == IV_ERROR_SUCCESS)
		{
			// Befehlsausfuehrung starten.
			iRes = m_channel4.ShowMessage(GetDeviceIDRemote(), ucDelayTime, ucAction, ucAudio, szMessage);

			// Versuchen die rote LED abzuschalten.
			m_channel4.SetTimeboySlotState(GetDeviceIDRemote(), 0, iRes);
		}

		return ResultToError(iRes, piError);}

	default:
		*piError = IV_ERROR_UNREACHABLE;
	}

	return 0;
}

// Datensatzbeschreibung oder Listenbeschreibung aus dem Geraet lesen.
int CAEChannel::GetTableDescription(short deviceIDRemote, unsigned char ucList, int *piError)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(GetDeviceIDRemote()))
	{
	case 1:  // AEIII+
	case 2: {// PZE - Master IV
		ULONG pDat, pFeld, pList[MAX_DATENSATZ_BESCHREIBUNG], pListEnd;
		int i, j;

		// EEPROM lesen 
 		if (!m_channel1_2.ReadMem(MB_START_EEP, EEP_SIZE_SU, m_file.m_eep, piError))
		{
			return 0;
		}

		// Start und Ende der Datensatzbeschreibungen lesen
		ChangeOrderLong(&m_file.m_eep[P_DAT_STRUCT]);
		StartDatensatzB = *(ULONG*)(&m_file.m_eep[P_DAT_STRUCT]);

		ChangeOrderLong(&m_file.m_eep[P_LIST_STRUCT]);
		EndDatensatzB = *(ULONG*)(&m_file.m_eep[P_LIST_STRUCT]);

		if (StartDatensatzB >=  MAX_FLASH_MEM || EndDatensatzB >= MAX_FLASH_MEM)
		{			
			*piError = IV_ERROR_CHANNEL_COMMUNICATION;
			return 0;
		}

		if (ucList)
		{
			//Pointer der Listenbeschreibungen lesen (Ende Datensatzbeschreibung ist anfang Listenb.
			if (!m_channel1_2.ReadMem(EndDatensatzB, MAX_DATENSATZ_BESCHREIBUNG * sizeof(ULONG), &m_file.m_flash[EndDatensatzB - FLASH_KONFIG_START], piError))
			{	
				return 0;
			}

			// Pointer extrahieren und Anzahl Listenb. ermitteln
			for (i=0; i < MAX_DATENSATZ_BESCHREIBUNG; i++)
			{
				ChangeOrderLong(&m_file.m_flash[EndDatensatzB - FLASH_KONFIG_START + i * sizeof(ULONG)]);
				pList[i] = *(ULONG*)(&m_file.m_flash[EndDatensatzB - FLASH_KONFIG_START + i *sizeof(ULONG)]);

				if (pList[i] == 0ul)
				{					
					MaxListB = i; 
					i = MAX_DATENSATZ_BESCHREIBUNG; 
					continue;
				}
			}

			// Ende Listenbereich angenaehert berechnen (mit Reserve)
			pListEnd = pList[MaxListB - 1] + sizeof(DatensatzBeschreibung[0]) + 
						(sizeof(DatensatzFeld[0][0]) + 
						sizeof(ULONG)) * (MAX_DATENSATZ_BESCHREIBUNG + 1);

			// Rest der Listenbeschreibung aus AEIII+ lesen
			pDat = EndDatensatzB + MAX_DATENSATZ_BESCHREIBUNG * sizeof(ULONG);

			if (!m_channel1_2.ReadMem(pDat, pListEnd - pDat, &m_file.m_flash[pDat - FLASH_KONFIG_START], piError))
			{				
				return 0;
			}

			// Listenbeschreibungen auswerten
			for (i=0; i < MaxListB; i++)
			{
				CpyFromKonfigBuf((unsigned char*)&ListenBeschreibung[i], m_file.m_flash, pList[i], sizeof(ListenBeschreibung[i]));
				for (j=0; j < ListenBeschreibung[i].Felder; j++)
				{ 
					// Feldbeschreibung einlesen
					ChangeOrderLong  (&m_file.m_flash[ChangeOrderULong(ListenBeschreibung[i].P) - FLASH_KONFIG_START + j * sizeof(ULONG)]);
					pFeld = *(ULONG*)(&m_file.m_flash[ChangeOrderULong(ListenBeschreibung[i].P) - FLASH_KONFIG_START + j * sizeof(ULONG)]);

					ASSERT (pFeld + sizeof(DatensatzFeld[i][j]) < pListEnd);

					CpyFromKonfigBuf((unsigned char*)&ListenFeld[i][j], m_file.m_flash, pFeld, sizeof(ListenFeld[i][j]));
				}
			}
		}
		else
		{
			// Datensatzbeschreibungen lesen
			if (!m_channel1_2.ReadMem(StartDatensatzB, EndDatensatzB - StartDatensatzB, &m_file.m_flash[StartDatensatzB - FLASH_KONFIG_START], piError))
			{				
				return 0;
			}

			// Datensatzbeschreibung auswerten
			for (i=0; i < MAX_DATENSATZ_BESCHREIBUNG; i++)
			{
				ChangeOrderLong(&m_file.m_flash[StartDatensatzB - FLASH_KONFIG_START + i * sizeof(ULONG)]);
				pDat = *(ULONG*)(&m_file.m_flash[StartDatensatzB - FLASH_KONFIG_START + i * sizeof(ULONG)]);

				if (pDat == 0ul)
				{					
					MaxDatB = i; 
					i = MAX_DATENSATZ_BESCHREIBUNG; 
					continue;
				}

				CpyFromKonfigBuf((unsigned char*)&DatensatzBeschreibung[i], m_file.m_flash, pDat, sizeof(DatensatzBeschreibung[i]));
				
				// Feldbeschreibung einlesen
				for (j=0; j < DatensatzBeschreibung[i].Felder; j++)
				{ 				
					ChangeOrderLong(&m_file.m_flash[ChangeOrderULong(DatensatzBeschreibung[i].P) - FLASH_KONFIG_START + j * sizeof(ULONG)]);
					pFeld = *(ULONG*)(&m_file.m_flash[ChangeOrderULong(DatensatzBeschreibung[i].P) - FLASH_KONFIG_START + j * sizeof(ULONG)]);

					CpyFromKonfigBuf((unsigned char*)&DatensatzFeld[i][j], m_file.m_flash, pFeld, sizeof(DatensatzFeld[i][j]));
				}
			}
		}		
		break;}

	case 4: {

		int iRes;

		// Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
		if ((iRes = m_channel4.SetTimeboySlotState(GetDeviceIDRemote(), 1)) == IV_ERROR_SUCCESS)
		{
			// Befehlsausfuehrung starten.
			iRes = GetTables4(GetDeviceIDRemote(), (ucList) ? (char*)"LIST" : (char*)"RECORD");

			// Versuchen die rote LED abzuschalten.
			m_channel4.SetTimeboySlotState(GetDeviceIDRemote(), 0, iRes);
		}

		return ResultToError(iRes, piError);}

	default:
		*piError = IV_ERROR_UNREACHABLE;
		return 0;
	}

	return 1;
}

// Datensatzbeschreibungen aus Geraet lesen
int CAEChannel::GetTables4(short deviceIDRemote, char *typeName)
{
	char name[64];
	unsigned char tableCount, fieldCount, i;
	int err = IV_ERROR_SUCCESS;

	STable2 table;
	STable2Field field;
	TDatensatzBeschreibung *pTable = NULL;
	TDatensatzFeld *pField = NULL;

	// Struktur uebernehmen
	if (strcmp(typeName, "LIST") == 0)
	{
		MaxListB = 0;
	}
	else if (strcmp(typeName, "RECORD") == 0)
	{
		MaxDatB = 0;
	}
	else
	{
		return -IV_ERROR_LIST_PROCESSING;
	}

	// Alle Tabellenbeschreibungen lesen
	for (tableCount=0; tableCount < MAX_DATENSATZ_BESCHREIBUNG; tableCount++)
	{
		sprintf(name, "%s.%i", typeName, tableCount);

		if ((err = m_channel4.GetTable(deviceIDRemote, name, table)) != IV_ERROR_SUCCESS)
		{
			// Nur wenn der Minorname (Index) nicht bekannt ist, 
			// liegt wirklich ein Fehler vor.
			if (err < 0)
			{
				// Kommunikationsfehler
				return err;
			}
			else if (err == 2)
			{ 
				// Tabellenname unbekannt
				// Fehlercode zuruecksetzen
				err = IV_ERROR_SUCCESS;
			}
			else
			{
				// Funktionsfehler
				return err;
			}

			break;
		}

		// Struktur uebernehmen
		if (strcmp(typeName, "LIST") == 0)
		{
			pTable = &ListenBeschreibung[tableCount];
		}
		else if (strcmp(typeName, "RECORD") == 0)
		{
			pTable = &DatensatzBeschreibung[tableCount];
		}

		strcpy((char*)pTable->Name, table.name);
		pTable->Felder = table.count;
		pTable->P = 0;
	}

	// Feldbeschreibung lesen
	for (i=0; i < tableCount; i++)
	{
		// Struktur uebernehmen
		if (strcmp(typeName, "LIST") == 0)
		{
			pTable = &ListenBeschreibung[i];
		}
		else if (strcmp(typeName, "RECORD") == 0)
		{
			pTable = &DatensatzBeschreibung[i];
		}

		// Alle Felder lesen
		for (fieldCount=0; fieldCount < pTable->Felder; fieldCount++)
		{
			sprintf(name, "%s.%i.%i", typeName, i, fieldCount);

			if ((err = m_channel4.GetField(deviceIDRemote, name, field)) != IV_ERROR_SUCCESS)
			{
				if (err < 0)
				{
					// Kommunikationsfehler
					return err;
				}
				else
				{
					// Funktionsfehler
					return err;
				}
			}

			// Struktur uebernehmen
			if (strcmp(typeName, "LIST") == 0)
			{
				pField = &ListenFeld[i][fieldCount];
			}
			else if (strcmp(typeName, "RECORD") == 0)
			{
				pField = &DatensatzFeld[i][fieldCount];
			}

			pField->Typ = field.type;
			pField->Cnt = field.size;
			strcpy((char*)pField->Name, field.name);
		}
	}

	// Information uebernehmen
	if (strcmp(typeName, "LIST") == 0)
	{
		MaxListB = tableCount;
	}
	else if (strcmp(typeName, "RECORD") == 0)
	{
		MaxDatB = tableCount;
	}

	return err;
}

// Listenbeschreibung aus dem Geraet lesen.
int CAEChannel::GetListName(short deviceIDRemote, unsigned char ucList, char *name, int *piError)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	ULONG add;
	TDatensatzBeschreibung dsb;

	switch (GetProtocolVersion(GetDeviceIDRemote()))
	{
	case 1:
	case 2:
		break;

	case 4:
		*piError = IV_ERROR_FUNCTION_UNAVAILABLE;
		return 0;

	default:
		*piError = IV_ERROR_UNREACHABLE;
		return 0;
	}

	if (!m_channel1_2.SetDeviceIDRemote((unsigned char)deviceIDRemote))
	{
		*piError = IV_ERROR_DEVICE_ID_REMOTE;
		return 0;
	}

	// Die Addresse auf die Listenbeschreibungen aus dem EEPROM-Bereich lesen 
 	if (!m_channel1_2.ReadMem(MB_START_EEP + P_LIST_STRUCT, sizeof(ULONG), (unsigned char*)&add, piError))
	{
		return 0;
	}
	ChangeOrderLong((unsigned char*)&add);

	// Die Addresse auf die Listenbeschreibung lesen.
	if (!m_channel1_2.ReadMem(add + ucList * sizeof(ULONG), sizeof(ULONG), (unsigned char*)&add, piError))
	{	
		return 0;
	}
	ChangeOrderLong((unsigned char*)&add);


	// Wenn die Rueckgabe FALSE und Error = IV_ERROR_SUCCESS, dann war die Liste nicht vorhanden.
	*piError = IV_ERROR_SUCCESS;

	if (add)
	{
		// Die Listenbeschreibung lesen.
		if (!m_channel1_2.ReadMem(add, sizeof(TDatensatzBeschreibung), (unsigned char*)&dsb, piError))
		{		
			return 0;
		}

		strcpy(name, (char*)dsb.Name);
		return 1;
	}

	return 0;
}

int CAEChannel::UploadStudioFile(short deviceIDRemote, char *szFileName, int *piError)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	int i, fileError;
  char ext[5];
  memset(ext, 0, sizeof(ext));
  strcpy(ext, &szFileName[strlen(szFileName) - 4]);
  for (i=0; ext[i] && i < 4; i++) {
    ext[i] = tolower(ext[i]);
  }
  if (strcmp(ext, ".aes") != 0 && strcmp(ext, ".ael") != 0) {
    // Die Anwendung kann ueber den gewaehlten Weg, nur Dateien vom Typ (*.aes; *.ael) oeffnen.
    *piError = IV_ERROR_FILE_EXTENSION;
    return 0;
  }

  // Reset der StudioFile-Klasse
  SetDocumentDefault(m_file);

  // Laden der Dateidaten. Keine Beachtung der Studioversion.
 	if (!m_file.Load(szFileName, &fileError, AESHEADER_IGNORE_MIN_STUDIO_VERSION | AESHEADER_IGNORE_SHOULD_STUDIO_VERSION)) {
    switch (fileError) {
      case -3:
        *piError = IV_ERROR_FILE_READ;
        break;
      case -7: // Mindestversion DLL wird nicht eingehalten.
        *piError = IV_ERROR_SETUPFILE_MIN_LIBRARY_VERSION;
        break;
      case -8: // Sollteversion DLL
        *piError = IV_ERROR_SETUPFILE_SHOULD_LIBRARY_VERSION;
        break;
      case -9: // Format der Setupdatei
        *piError = IV_ERROR_FILE_FORMAT;
        break;
      case -4: // CRC in Dateidaten
      case -10: case -11: case -12: case -13: case -14: case -15: case -16: case -17: case -18:
        *piError = IV_ERROR_FILE_CRC;
        break;
      default:
		  *piError = IV_ERROR_FILE_OPEN;
    }
		return 0;
	}

	// Pruefen ob Setupdatei fuer diesen Geraetetyp ist.
	switch (GetProtocolVersion(GetDeviceIDRemote())) {
	case 1: // AEIII+
    if (m_file.IsHeader()) {
      // Die Firmware wird durch diese Setupdatei nicht unterstuetzt.
      *piError = IV_ERROR_VERSION_SETUPFILE;
      return 0;
    }
    if (strcmp(ext, ".ael") == 0)
    {
      // Nicht gueltige Endung der Setupdatei.
      *piError = IV_ERROR_FILE_EXTENSION;
      return 0;
    }

		// Ist eine AEIII+ Datei?
		if ((m_file.m_eep[P_SETUP_VERSION] != 0xFF) && (m_file.m_eep[P_SETUP_VERSION] != 3)) {		
			*piError = IV_ERROR_FILE_FORMAT;
			return 0;
		}
		if (m_file.m_eep[P_DEVICE_TYPE] != 0) {
			*piError = IV_ERROR_DEVICE_WRONG;
			return 0;
		}
		break;

  case 2: // PZE Master IV
    if (m_file.IsHeader()) {
      // Die Firmware wird durch diese Setupdatei nicht unterstuetzt.
      *piError = IV_ERROR_VERSION_SETUPFILE;
      return 0;
    }
    if (strcmp(ext, ".ael") == 0)
    {
      // Nicht gueltige Endung der Setupdatei.
      *piError = IV_ERROR_FILE_EXTENSION;
      return 0;
    }

		if (m_file.m_eep[P_DEVICE_TYPE] != 1) {
			*piError = IV_ERROR_DEVICE_WRONG;
			return 0;
		}
		break;

	case 4: { // Master IV
		int iRes;
		unsigned char type = 0;
		iRes = ResultToError(m_channel4.GetTypeDevice(GetDeviceIDRemote(), type), piError);
		if (iRes == 0) {
			if (*piError == IV_ERROR_FUNCTION_UNAVAILABLE) {
				// Geraet als PZE-MasterIV annehmen
				type = 0;
				*piError = IV_ERROR_SUCCESS;
			} else {
				return 0;
			}
		}

    char version[32];
    memset(version, 0, sizeof(version));
    iRes = ResultToError(m_channel4.GetVersion(GetDeviceIDRemote(), version), piError);
    if (iRes == 0) {
     return 0;
    }
    // Pruefen ob die Firmwareversion unterstuetzt wird.
    if (!m_file.IsSupportedFirmwareVersion(version)) {
      *piError = IV_ERROR_VERSION_FIRMWARE;
      return 0;
    }

    if ((m_file.m_eep[P_DEVICE_TYPE] - 1) != type) {
      *piError = IV_ERROR_DEVICE_WRONG;
      return 0;
    }

    // Ist ein Dateiheader verfuegbar, ist die Firmwareversion zu pruefen.
    if (m_file.IsHeader()) {

      CDFSoftwarebuildVersion fwVersion(version);
      switch (m_file.ValidFirmwareVersion(fwVersion)) {
      case 0: // Kein Dateiheader, hier unmoeglich.
        *piError = IV_ERROR_FILE_FORMAT;
        return 0;
      case -1: // Unsupported 04.01.xx
        *piError = IV_ERROR_SETUPFILE_UNSUPPORTED_FIRMWARE1_VERSION;
        return 0;
      case -2: // Mindestversion 04.01.xx
        *piError = IV_ERROR_SETUPFILE_MIN_FIRMWARE1_VERSION;
        return 0;
      case -3: // Sollteversion 04.01.xx
        *piError = IV_ERROR_SETUPFILE_SHOULD_FIRMWARE1_VERSION;
        return 0;
      case -4: // Unsupported 04.02.xx
        *piError = IV_ERROR_SETUPFILE_UNSUPPORTED_FIRMWARE2_VERSION;
        return 0;
      case -5: // Mindestversion 04.02.xx
        *piError = IV_ERROR_SETUPFILE_MIN_FIRMWARE2_VERSION;
        return 0;
      case -6: // Sollteversion 04.02.xx
        *piError = IV_ERROR_SETUPFILE_SHOULD_FIRMWARE2_VERSION;
        return 0;
      case -7: // Firmware wird nicht unterstuetz.
        *piError = IV_ERROR_VERSION_FIRMWARE;
        return 0;
      }
    } else {

      // Wenn kein Dateiheader vorliegt, wird sichergestellt, dass eine Setupdatei mit der hinterlegten
      // DatafoxStudioIV Version 04.02.xx nur auf eine Hardware >= V3 aufgespielt werden kann.

		  // Boardtyp ermitteln und Pruefen ob die Setupversion eingespielt werden darf.
		  iRes = ResultToError(GetTypeBoard(GetDeviceIDRemote(), type), piError);
		  if (iRes == 0) {
			  if (*piError == IV_ERROR_FUNCTION_UNAVAILABLE) {
				  // Baugruppe als 1.4 annehmen
				  type = 0x14;
				  *piError = IV_ERROR_SUCCESS;
			  } else {
				  return 0;
			  }
		  }

		  if (type < IV_VALUE_TYPE_BOARD_3_0) {
			  // Neuere Setupdateien nicht in aeltere Hardwarestaende einspielen.
			  //############################
			  // 2010-03-04
			  if ((m_file.m_eep[P_SETUP_VERSION] > 4) || ((m_file.m_eep[P_SETUP_VERSION] == 4) && (m_file.m_eep[P_SETUP_VERSION + 1] > 1))) {
				  *piError = IV_ERROR_VERSION_SETUPFILE;
				  return 0;
			  }
		  }
    }
		break;}

	default:
		*piError = IV_ERROR_UNREACHABLE;
		return 0;
	}

	switch (GetProtocolVersion(GetDeviceIDRemote())) {
	case 1: 
	case 2: {
    // Kommunikationsversionsnummer
    int commVersion = 0;

    // Bei AEIII+ zusaetzliche Pruefung wegen Befehlsunterstuetzung notwendig
    commVersion = GetProtocolVersion2(GetDeviceIDRemote(), TRUE, piError);
    switch (commVersion)
    {
    case 0x0400:
    case 0x0401:
    case 0x0402:
    case 0x0501:
      break;

    default:
      *piError = IV_ERROR_VERSION_FIRMWARE;
      return 0;
    }

    if (commVersion != 0x0400 && commVersion != 0x0401)	{
      char version[32];
      memset(version, 0, sizeof(version));
      // Firmwareversion lesen
      if (!GetVersionFirmware(GetDeviceIDRemote(), version, piError))
      {
        return 0;
      }

      CDFComUtils util;
      // Firmwareversion pruefen
      if (!util.CheckVersionFirmware(version))
      {
        *piError = IV_ERROR_VERSION_FIRMWARE;
        return 0;
      }
    }

		return UploadSetupData(piError);}

	case 4: { // PZEMasterIV mit Protokollversion 4
		int iRes = IV_ERROR_SUCCESS;

		if ((iRes = CheckData(GetDeviceIDRemote(), piError)) == -1) {
			return 0;
		} else if (iRes) {
			*piError = IV_ERROR_DATA_AVAILABLE;
			return 0;
		}

		// Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
		if ((iRes = m_channel4.SetTimeboySlotState(GetDeviceIDRemote(), 1)) == IV_ERROR_SUCCESS) {

			// Befehlsausfuehrung starten.
			iRes = UploadSetupData4(GetDeviceIDRemote());
      if (iRes == IV_ERROR_SUCCESS && !m_channel4.IsPasswordProtection()) {
        iRes = m_channel4.ClearRecords(GetDeviceIDRemote());
      }
			// Versuchen die rote LED abzuschalten.
			m_channel4.SetTimeboySlotState(GetDeviceIDRemote(), 0, iRes);
		}

    return ResultToError(iRes, piError);}

	default:
		*piError = IV_ERROR_UNREACHABLE;
	}

	return 0;
}

int CAEChannel::UploadSetupData(int *piError)
{
  unsigned char ucIsData = 0, temp = 1; // fuer Bedienstopp

  // Abfrage ob Daten vorhanden. 
  // (wegen alter FW, nicht mit CheckData ausgefuehrt.)
  if (!m_channel1_2.ReadMem(MB_RECORD_AVAILABLE, 1, &ucIsData, piError)) {			
    return 0;
  }
  // Abbruch wenn Daten im Geraet vorliegen
  if (ucIsData) {
    *piError = IV_ERROR_DATA_AVAILABLE;
    return 0;
  }

  // Setupgroesse ermitteln
  ULONG ulSetupSize = GetFlashSetupSize();
  // Vorhandene zur moeglichen Groesse pruefen
  if (ulSetupSize > FLASH_SIZE_SU) {
    *piError = IV_ERROR_FILE_OPEN;
    return 0;
  }
  // Bedienstopp setzen
  if (!m_channel1_2.WriteMem(MB_CONTROL_STOP, 1, &temp, piError)) {
    return 0;
  }

  // Groesse der Flashdaten in EepMem eintragen
  *(ULONG*)&m_file.m_eep[P_FLASH_SIZE_SU] = ulSetupSize;
  ChangeOrderLong(&m_file.m_eep[P_FLASH_SIZE_SU]);
  // Groesse der MDE - Eingabeketten in EepMem loeschen.
  *(ULONG*)&m_file.m_eep[P_MDE_EK_SIZE] = 0xFFFFFFFF;
  ChangeOrderLong(&m_file.m_eep[P_MDE_EK_SIZE]);
  // Groesse der Zutrittskontroll - Eingabeketten in EepMem loeschen.
  *(ULONG*)&m_file.m_eep[P_ENTRANCE_EK_SIZE] = 0xFFFFFFFF;
  ChangeOrderLong(&m_file.m_eep[P_ENTRANCE_EK_SIZE]);
  // Groesse der MDE - Eingabeketten in EepMem eintragen.
  *(ULONG*)&m_file.m_eep[P_MDE_EK_SIZE] = GetMDEEKFlashSize();
  ChangeOrderLong(&m_file.m_eep[P_MDE_EK_SIZE]);

  ULONG commVersion = GetProtocolVersion2(GetDeviceIDRemote(), TRUE, piError);
  if (commVersion == 0) return FALSE;
  if (commVersion != 0x0401) {
    // Groesse der Zutrittskontroll - Eingabeketten in EepMem eintragen.
    *(ULONG*)&m_file.m_eep[P_ENTRANCE_EK_SIZE] = GetEntranceEKFlashSize();
    ChangeOrderLong(&m_file.m_eep[P_ENTRANCE_EK_SIZE]);
  }

#define MB_DATABLOCKSIZE 32

  // EEPROM - schreiben 
  if (!m_channel1_2.WriteMem(MB_START_EEP, EEP_SIZE_SU, m_file.m_eep, piError, MB_DATABLOCKSIZE)) {
    return 0;
  }
  // Flash - schreiben
  if (!m_channel1_2.WriteMem(FLASH_KONFIG_START, ulSetupSize, m_file.m_flash, piError, MB_DATABLOCKSIZE)) {
    return 0;
  }
  // Konfigurationsdaten - schreiben
  if (!m_channel1_2.WriteMem(MB_START_CONFIG + 2, CONFIG_SIZE_SU - 2, &m_file.m_configMem[2], piError, MB_DATABLOCKSIZE)) {
    return 0;
  }
  // IOModule und MDE ab Version 3.0
  if (commVersion != 0x0401) {
    int i;
    TExtendedFlash extBegin = *(TExtendedFlash*)m_file.m_flashExtended;
    // IOModul - Konfiguration - schreiben
    if (!m_channel1_2.WriteMem(MB_START_IOMODULE, sizeof(TIOModule), &m_file.m_flashExtended[extBegin.pulIOModule], piError, MB_DATABLOCKSIZE)) {
      return 0;
    }
    // IOModul - DigitalMDE - schreiben
    for (i=0; i < 8; i++) {
      if (!m_channel1_2.WriteMem(MB_DIGITAL_IO + i * sizeof(TDigitalInput), sizeof(TDigitalInput), 
        &m_file.m_flashExtended[extBegin.pulIODigital + i * sizeof(TDigitalInput)], piError, MB_DATABLOCKSIZE))	{
          return 0;
      }
    }

    // Interne MDE - DigitalMDE - schreiben
    for (i=0; i < 4; i++) {
      if (!m_channel1_2.WriteMem(MB_DIGITAL_MDE + i * sizeof(TDigitalInput), sizeof(TDigitalInput), 
        &m_file.m_flashExtended[extBegin.pulMDEDigital + i * sizeof(TDigitalInput)], piError, MB_DATABLOCKSIZE)) {
          return 0;
      }
    }

    // MDE - Eingabeketten - schreiben
    // Bei 4.0 kommt erst 1Byte Anzahl im Array
    // Ab 4.1 kommen erst 4Byte Anzahl im Array
    ULONG ulSize = 0;
    if (commVersion == 0x0400) {
      ulSize = GetMDEEKFlashSize_0400();
    } else {				
      ulSize = GetMDEEKFlashSize();
    }

    if (!m_channel1_2.WriteMem(MB_MDE_EK, ulSize, &m_file.m_flashExtended[extBegin.pulEKArray], piError, MB_DATABLOCKSIZE)) {
      return 0;
    }

    // Zutrittskontrolle
    if (commVersion != 0x0400) {
      // Zutrittskontroll - Eingabeketten - schreiben
      if (!m_channel1_2.WriteMem(MB_ENTRANCE_EK, GetEntranceEKFlashSize(), 
        &m_file.m_flashExtended[extBegin.pulEKArrayEntrance], piError, MB_DATABLOCKSIZE)) {
          return 0;
      }
    }
  }


  temp = 0; // Bedienstopp ruecksetzen
  if (!m_channel1_2.WriteMem(MB_CONTROL_STOP, 1, &temp, piError)) {
    return 0;
  }

  // Aktuelle Setup - Daten wurden erst durch zuruecksetzen von Bedienstopp uebernommen!

  // Standardwerte fuer Globale Variable
  if (commVersion != 0x0400) {
    int i;
    TExtendedFlash extBegin = *(TExtendedFlash*)m_file.m_flashExtended;

    // GV setzen?
    if (extBegin.pulValueGVx) {			
      unsigned char ucNum;
      // Struktur lesen
      TValueGVx ValueGVx = *(TValueGVx*)&m_file.m_flashExtended[extBegin.pulValueGVx];

      // Zeit fuer Umschaltung aus Bedienstopp
      Sleep(200);

      for (i=0; i < 8; i++) {
        if (ValueGVx.ucFlags & (1 << i)) {					
          ucNum = 1 + i;
          if (!SetGlobVar(GetDeviceIDRemote(), &ucNum, 0, ValueGVx.szValue[i], piError)) {						
            return 0;
          }
        }
      }
    }	
  }

  // Standardwerte fuer Systemvariable
  // Ab 04.00.23
  if (commVersion == 0x0502) {
    //############################
    // 2005-09-19: Erweiterte Struktur fuer Parameter
    if (((m_file.m_eep[P_SETUP_VERSION    ] == 3) &&
      (m_file.m_eep[P_SETUP_VERSION + 1] == 1) &&
      ((m_file.m_eep[P_SETUP_VERSION + 2] >= 34) &&
      (m_file.m_eep[P_SETUP_VERSION + 3] >  5)) ) ||

      ((m_file.m_eep[P_SETUP_VERSION   ] == 4) &&
      (m_file.m_eep[P_SETUP_VERSION + 1] == 1) &&
      (m_file.m_eep[P_SETUP_VERSION + 3] > 2)) )
    {
      TExtendedFlash extBegin = *(TExtendedFlash*)m_file.m_flashExtended;

      // Liegt die Struktur vor?
      if (extBegin.pulExtended) {
        TExtended extParam = *(TExtended*)&m_file.m_flashExtended[extBegin.pulExtended];

        unsigned char val[32], name[64];
        // Buchungswiederholsperrdauer
        sprintf((char*)name, "_SERVICE.DBLINPUT");
        sprintf((char*)val, "%i", extParam.doubleInputTime);
        // 
        if (!SetGlobVar(GetDeviceIDRemote(), name, 2, (char*)val, piError))	{					
          if (// Fehler liegt bei Wert
            *piError != IV_ERROR_VARIABLE_VALUE &&
            // Variable nicht bekannt
            *piError != IV_ERROR_VARIABLE_UNDEFINED)
            return 0;
        }
        // Abblendedauer fuer Bestaetigungsfeld
        sprintf((char*)name, "_SERVICE.DEMANDTIMER");
        sprintf((char*)val, "%i", extParam.confirmationTime);
        // 
        if (!SetGlobVar(GetDeviceIDRemote(), name, 2, (char*)val, piError)) {					
          if (// Fehler liegt bei Wert
            *piError != IV_ERROR_VARIABLE_VALUE &&
            // Variable nicht bekannt
            *piError != IV_ERROR_VARIABLE_UNDEFINED)
            return 0;
        }
      }
    }
  }

  return 1;
}

int CAEChannel::UploadSetupData4(short deviceIDRemote)
{
	int i, res = IV_ERROR_SUCCESS;
	unsigned char dummyConfig[0x1C]; // Groesse darf nicht veraendert werden!!!
	                                 // Berechnung 0x80 - sizeof(TConfig)

	//@@: Uebergangsweise zur 04.01.01 release
	unsigned char *buf = NULL;
	unsigned char countTable, countField;
	unsigned int sizeAccess;

	//////////////////////////////
	// Vorbereitung (Berechnung von Groessen und CRC)
	unsigned int sizeSetup = GetFlashSetupSize();

	// Vorhandene zur moeglichen Groesse pruefen
// 	if (sizeSetup > (EEP_SIZE_SU + sizeof(TConfig) + sizeof(dummyConfig) + FLASH_SIZE_SU)) {
// 
// 		return -IV_ERROR_FILE_FORMAT;
// 	}

	unsigned int sizeEK = GetEntrance2EKFlashSize();
	// Speicher reservieren
	buf = new unsigned char[sizeEK];
	if (buf == NULL) {
		return -IV_ERROR_ALLOC_MEM;
	}
	memset(buf, 0, sizeEK);
	

	//@@: Uebergangsweise zur 04.01.01 release
	// Zeiger auf Eingabeketten ohne FLASH_KONFIG_START
	// und die Zeiger der Eingabeketten mit einem Offset versehen.
	PrepareEntrance2EK(sizeSetup, buf);

	// Groesse der Flashdaten in EepMem eintragen
	*(unsigned int*)&m_file.m_eep[P_FLASH_SIZE_SU] = sizeSetup;
	ChangeOrderLong(&m_file.m_eep[P_FLASH_SIZE_SU]);

	// Anfangsadresse der Zutrittseingabeketten eintragen
	*(unsigned int*)&m_file.m_eep[P_IC_ACCESS] = sizeSetup;
	ChangeOrderLong(&m_file.m_eep[P_IC_ACCESS]);

	// Anfangsadresse der Zutrittskontrolllisten eintragen
	*(unsigned int*)&m_file.m_eep[P_LIST_ACCESS] = sizeSetup + sizeEK;
	ChangeOrderLong(&m_file.m_eep[P_LIST_ACCESS]);

	// Groesse der MDE - Eingabeketten in EepMem eintragen.
	*(unsigned int*)&m_file.m_eep[P_MDE_EK_SIZE] = GetMDEEKFlashSize();
	//ChangeOrderLong(&m_file.m_eep[P_MDE_EK_SIZE]);
	
	// Groesse der Zutrittskontroll - Eingabeketten in EepMem eintragen.
	*(unsigned int*)&m_file.m_eep[P_ENTRANCE_EK_SIZE] = sizeEK;
	ChangeOrderLong(&m_file.m_eep[P_ENTRANCE_EK_SIZE]);


	//
	unsigned int address, size, crc;
	address = size = crc = 0;

	// Segmentbegin: Bedienung stoppen
	if ((res = m_channel4.BeginSetSegment(deviceIDRemote, IV_VALUE_SEGMENT_UPDATE, address, size, crc)) != IV_ERROR_SUCCESS)
	{
		// Puffer Freigeben
		delete[] buf;
		return res;
	}
	m_channel4.Trace("%s(%i):INFO(1): CAEChannel::UploadSetupData4 [ BeginSetSegment(%i): %i, %i, 0x%08X ]\n", __FILE__, __LINE__, IV_VALUE_SEGMENT_UPDATE, address, size, crc);

	///////////////
	// Grunddaten
	memset(dummyConfig, 0, sizeof(dummyConfig));

	// CRC berechnen
	crc = 0xFFFFFFFF;
	crc = m_channel4.GetCRC32(m_file.m_eep, EEP_SIZE_SU, crc, FALSE);
	crc = m_channel4.GetCRC32(m_file.m_configMem, sizeof(TConfig), crc, FALSE);
	crc = m_channel4.GetCRC32(dummyConfig, sizeof(dummyConfig), crc, FALSE);
	crc = m_channel4.GetCRC32(m_file.m_flash, sizeSetup, crc, FALSE);
	//@@: Uebergangsweise zur 04.01.01 release
	crc = m_channel4.GetCRC32(buf, sizeEK, crc, FALSE);
//	crc = m_channel4.GetCRC32(buf, sizeEK, crc);


	///////////////
	// ZK II

	// Anzahl der Zutrittslisten ermitteln
	for (i=0, countTable=0; countTable < (sizeof(m_entrance2Pointer) / sizeof(unsigned int)); countTable++)
	{
		if (m_entrance2Pointer[countTable] == 0)
		{
			break;
		}

		// Gesamte Feldanzahl fuer Groeszenberechnung sammeln.
		i += m_entrance2Table[countTable].count;
	}
	sizeAccess = sizeof(m_entrance2Pointer);
	sizeAccess += countTable * sizeof(STable2);
	sizeAccess += i * sizeof(STable2Field);

	// CRC der Zutrittslisten ermitteln
	crc = m_channel4.GetCRC32((unsigned char*)m_entrance2Pointer, sizeof(m_entrance2Pointer), crc, FALSE);

	for (i=0; i < countTable; i++)
	{
		crc = m_channel4.GetCRC32((unsigned char*)&m_entrance2Table[i], sizeof(STable2), crc, FALSE);

		for (countField=0; countField < m_entrance2Table[i].count; countField++)
		{
			// Beim letzten Feld der letzten Tabelle musz der CRC abgeschlossen werden.
			crc = m_channel4.GetCRC32((unsigned char*)&m_entrance2TableField[i][countField], sizeof(STable2Field), crc, ((i == (countTable - 1)) && (countField == (m_entrance2Table[i].count - 1))));
		}
	}

	//////////////////////////////
	// Begin
//	double now = (double)clock() / (double)CLOCKS_PER_SEC;
//	int nr = 0;
//	m_channel4.Trace("Elapse (%i): %f\n", nr++, ((double)clock() / (double)CLOCKS_PER_SEC) - now);

	// Gesamtgroesze berechnen
	size = EEP_SIZE_SU + sizeof(TConfig) + sizeof(dummyConfig) + sizeSetup + sizeEK + sizeAccess;

	// Segmentbeginn: Setup uebertragen
	if ((res = m_channel4.BeginSetSegment(deviceIDRemote, IV_VALUE_SEGMENT_SETUP, address, size, crc)) != IV_ERROR_SUCCESS)
	{
		// Puffer Freigeben
		delete[] buf;
		return res;
	}
	m_channel4.Trace("%s(%i):INFO(1): CAEChannel::UploadSetupData4 [ BeginSetSegment(%i): %i, %i, 0x%08X ]\n", __FILE__, __LINE__, IV_VALUE_SEGMENT_SETUP, address, size, crc);

	//////////////////////////////
	// Grunddaten

	// EEPROM uebertragen
	if ((res = m_channel4.SetFlash(deviceIDRemote, address, m_file.m_eep, EEP_SIZE_SU)) != IV_ERROR_SUCCESS)
	{
		// Puffer Freigeben
		delete[] buf;
		return res;
	}

	// CONFIG uebertragen (100 Byte)
	address += EEP_SIZE_SU;
	if ((res = m_channel4.SetFlash(deviceIDRemote, address, m_file.m_configMem, sizeof(TConfig))) != IV_ERROR_SUCCESS)
	{
		// Puffer Freigeben
		delete[] buf;
		return res;
	}

	// CONFIG dummyConfig (28 Byte)
	address += sizeof(TConfig);
	if ((res = m_channel4.SetFlash(deviceIDRemote, address, dummyConfig, sizeof(dummyConfig))) != IV_ERROR_SUCCESS)
	{
		// Puffer Freigeben
		delete[] buf;
		return res;
	}

	// FLASH uebertragen
	address += sizeof(dummyConfig);
	if ((res = m_channel4.SetFlash(deviceIDRemote, address, m_file.m_flash, sizeSetup)) != IV_ERROR_SUCCESS)
	{
		// Puffer Freigeben
		delete[] buf;
		return res;
	}

	//@@: Uebergangsweise zur 04.01.01 release
	//////////////////////////////
	// Pointerarray + Strukturen der Zutrittslisten uebertragen
	
	// Eingabekette der Zutrittskontrolle
	address += sizeSetup;
	if ((res = m_channel4.SetFlash(deviceIDRemote, address, buf, sizeEK)) != IV_ERROR_SUCCESS)
	{
		// Puffer Freigeben
		delete[] buf;
		return res;
	}

	// Puffer nach Uebertragung freigeben
	delete[] buf;
	buf = NULL;

	address += sizeEK;
	// Pointerarray uebertragen
	if ((res = m_channel4.SetFlash(deviceIDRemote, address, (unsigned char*)m_entrance2Pointer, sizeof(m_entrance2Pointer))) != IV_ERROR_SUCCESS)
	{
		return res;
	}
	address += sizeof(m_entrance2Pointer);

	// Strukturen uebertragen
	for (i=0; i < countTable; i++)
	{
		// Tabellenstruktur uebertragen
		if ((res = m_channel4.SetFlash(deviceIDRemote, address, (unsigned char*)&m_entrance2Table[i], sizeof(STable2))) != IV_ERROR_SUCCESS)
		{
			return res;
		}
		address += sizeof(STable2);

		// Feldstrukturen uebertragen
		if ((res = m_channel4.SetFlash(deviceIDRemote, address, (unsigned char*)m_entrance2TableField[i], m_entrance2Table[i].count * sizeof(STable2Field))) != IV_ERROR_SUCCESS)
		{
			return res;
		}
		address += m_entrance2Table[i].count * sizeof(STable2Field);
	}

	//////////////////////////////
	// Ende
	// Segmentend: Setup uebertragen
	if ((res = m_channel4.EndSetSegment(deviceIDRemote, IV_VALUE_SEGMENT_SETUP, crc)) != IV_ERROR_SUCCESS)
	{
		return res;
	}
	m_channel4.Trace("%s(%i):INFO(1): CAEChannel::UploadSetupData4 [ EndSetSegment(%i): 0x%08X ]\n", __FILE__, __LINE__, IV_VALUE_SEGMENT_SETUP, crc);

	// Extendedstruktur lesen
	TExtendedFlash extBegin = *(TExtendedFlash*)m_file.m_flashExtended;

	///////////////
	// Extended - Speicher Vorbereitung

	// Zeigeroffsets anpassen (substrahieren von Offset)
	AdjustExtendedMemPointer();
	
	// Groesse MDE und dessen Eingabeketten
	size = extBegin.pulFreeScope;

	// CRC berechnen
	crc = 0xFFFFFFFF;
	// IOModul - Konfiguration
	crc = m_channel4.GetCRC32(m_file.m_flashExtended, size, crc, TRUE);

	//////////////////////////////
	// Begin

	// Segmentbegin: Setup uebertragen
	if ((res = m_channel4.BeginSetSegment(deviceIDRemote, IV_VALUE_SEGMENT_SETUP_EXTENDED, address, size, crc)) != IV_ERROR_SUCCESS)
	{
		if ((res & 0xFF) == 1)
		{
			// Funktionsfehler Typ IV_VALUE_SEGMENT_SETUP_EXTENDED unbekannt.
		}
		else
		{		
			return res;
		}
	}
	m_channel4.Trace("%s(%i):INFO(1): CAEChannel::UploadSetupData4 [ (%i)BeginSetSegment(%i): %i, %i, 0x%08X ]\n", __FILE__, __LINE__, res, IV_VALUE_SEGMENT_SETUP_EXTENDED, address, size, crc);

	if (res == IV_ERROR_SUCCESS)
	{
		//////////////////////////////
		// Extended - Speicher Uebertragen

		if ((res = m_channel4.SetFlash(deviceIDRemote, address, m_file.m_flashExtended, size)) != IV_ERROR_SUCCESS)
		{
			return res;
		}
		address += size;

		//////////////////////////////
		// Ende

		// Segmentend: Setup uebertragen
		if ((res = m_channel4.EndSetSegment(deviceIDRemote, IV_VALUE_SEGMENT_SETUP_EXTENDED, crc)) != IV_ERROR_SUCCESS)
		{
			return res;
		}
		m_channel4.Trace("%s(%i):INFO(1): CAEChannel::UploadSetupData4 [ EndSetSegment(%i): 0x%08X ]\n", __FILE__, __LINE__, IV_VALUE_SEGMENT_SETUP_EXTENDED, crc);
	}

	// Zeigeroffsets anpassen (addieren von Offset)
	AdjustExtendedMemPointer();

	// Pause wegen Verarbeitungsporblemen bis 4.1.5
	Sleep(1000);

	crc = 0;
	// Segmentend: Bedienung stoppen
	if ((res = m_channel4.EndSetSegment(deviceIDRemote, IV_VALUE_SEGMENT_UPDATE, crc)) != IV_ERROR_SUCCESS)
	{
		return res;
	}
	m_channel4.Trace("%s(%i):INFO(1): CAEChannel::UploadSetupData4 [ EndSetSegment(%i): 0x%08X ]\n", __FILE__, __LINE__, IV_VALUE_SEGMENT_UPDATE, crc);


  // Pruefen und setzen des Status von Passwortschutz.
  PasswordCheck(deviceIDRemote, "ISPW");
	////////////////////////////////////////////////////////////////////
	// Der folgende Teil ist nur bei Versionen unter 04.01.07 noetig.
	// Ab dieser Version ist auch der Passwortschutz moeglich.
	// Es ist hier nicht moeglich, ein Setup mit einem gesetzten Passwort zu uebertragen 
	// und darauffolgend gleich weitere Funktionen aufrufen, weil ansonsten der
	// Schluessel fuer die Passwortverschluesselung offen gelegt werden muesste.
	if (!m_channel4.IsPasswordProtection())
	{
		// GV setzen?
		if (extBegin.pulValueGVx)
		{		
			SGlobVar *globVar = (SGlobVar*)m_file.m_flash;
			char name[64];
			memset(name, 0, sizeof(name));

			// Struktur lesen
			TValueGVx ValueGVx = *(TValueGVx*)&m_file.m_flashExtended[extBegin.pulValueGVx];

			for (i=0; i < 8; i++)
			{
				if (ValueGVx.ucFlags & (1 << i))
				{				
					sprintf(name, "SETUP.%s", (char*)&globVar->SFeld[i]);
					if ((res = m_channel4.SetVariable(deviceIDRemote, name, (char*)&ValueGVx.szValue[i])) != IV_ERROR_SUCCESS)
					{					
						// Das Befehlsbyte wird bei unbekannter Variabler nicht veraendert
						// und wenn der Wert nicht gesetzt werden konnte wird das hoechste bit gesetzt.
						if ((res & 0xFF) == 3)
						{
							// Fehler liegt bei Wert
							//res = -IV_ERROR_VARIABLE_VALUE;
						}
						else if ((res & 0xFF) == 2)
						{
							// Variable nicht bekannt
							//res = -IV_ERROR_VARIABLE_UNDEFINED;
						}

						return res;
					}
				}
			}
		}	

		// Standardwerte fuer Systemvariable
		//############################
		// 2005-09-19: Erweiterte Struktur fuer Parameter
		if (((m_file.m_eep[P_SETUP_VERSION    ] == 3) &&
			 (m_file.m_eep[P_SETUP_VERSION + 1] == 1) &&
			 ((m_file.m_eep[P_SETUP_VERSION + 2] >= 34) &&
			  (m_file.m_eep[P_SETUP_VERSION + 3] >  5)) ) ||

			 ((m_file.m_eep[P_SETUP_VERSION   ] == 4) &&
			  (m_file.m_eep[P_SETUP_VERSION + 1] == 1) &&
			  (m_file.m_eep[P_SETUP_VERSION + 3] > 2)) )
		{
			TExtendedFlash extBegin = *(TExtendedFlash*)m_file.m_flashExtended;

			// Liegt die Struktur vor?
			if (extBegin.pulExtended)
			{
				TExtended extParam = *(TExtended*)&m_file.m_flashExtended[extBegin.pulExtended];

				char val[32], name[64];
				// Buchungswiederholsperrdauer
				sprintf(name, "_SERVICE.DBLINPUT");
				if (extParam.doubleInputTime > 65499)
				{
					extParam.doubleInputTime = 65499;
				}
				sprintf(val, "%i", extParam.doubleInputTime);
				// 
				if ((res = m_channel4.SetSystemVariable(deviceIDRemote, name, val)) != IV_ERROR_SUCCESS)
				{				
					// Das Befehlsbyte wird bei unbekannter Variabler nicht veraendert
					// und wenn der Wert nicht gesetzt werden konnte wird das hoechste bit gesetzt.
					if ((res & 0xFF) == 3)
					{
						// Fehler liegt bei Wert
						//res = -IV_ERROR_VARIABLE_VALUE;
						res = IV_ERROR_SUCCESS;
					}
					else if ((res & 0xFF) == 2)
					{
						// Variable nicht bekannt
						//res = -IV_ERROR_VARIABLE_UNDEFINED;
						res = IV_ERROR_SUCCESS;
					}
				}
				// Abblendedauer fuer Bestaetigungsfeld
				sprintf(name, "_SERVICE.DEMANDTIMER");
				if (extParam.confirmationTime > 255)
				{
					extParam.confirmationTime = 255;
				}
				sprintf(val, "%i", extParam.confirmationTime);
				// 
				if ((res = m_channel4.SetSystemVariable(deviceIDRemote, name, val)) != IV_ERROR_SUCCESS)
				{				
					// Das Befehlsbyte wird bei unbekannter Variabler nicht veraendert
					// und wenn der Wert nicht gesetzt werden konnte wird das hoechste bit gesetzt.
					if ((res & 0xFF) == 3)
					{
						// Fehler liegt bei Wert
						//res = -IV_ERROR_VARIABLE_VALUE;
						res = IV_ERROR_SUCCESS;
					}
					else if ((res & 0xFF) == 2)
					{
						// Variable nicht bekannt
						//res = -IV_ERROR_VARIABLE_UNDEFINED;
						res = IV_ERROR_SUCCESS;
					}
				}

				// Tiefentladungsschutz
				sprintf(name, "POWER.LOW");
				sprintf(val, "%i", extParam.energieOption.powerLow);
				// 
				if ((res = m_channel4.SetSystemVariable(deviceIDRemote, name, val)) != IV_ERROR_SUCCESS)
				{			
					// Das Befehlsbyte wird bei unbekannter Variabler nicht veraendert
					// und wenn der Wert nicht gesetzt werden konnte wird das hoechste bit gesetzt.
					if ((res & 0xFF) == 3)
					{
						// Fehler liegt bei Wert
						//res = -IV_ERROR_VARIABLE_VALUE;
						res = IV_ERROR_SUCCESS;
					}
					else if ((res & 0xFF) == 2)
					{
						// Variablenname nicht bekannt
						//res = -IV_ERROR_VARIABLE_UNDEFINED;
						res = IV_ERROR_SUCCESS;
					}
					else if ((res & 0xFF) == 1)
					{
						// Majorbestandteil unbekannt
						//res = -IV_ERROR_VARIABLE_UNDEFINED;
						res = IV_ERROR_SUCCESS;
					}
				}

			}
		}
	}

	return res;
}

int CAEChannel::DownloadStudioFile(short deviceIDRemote, char *szFileName, int *piError)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	int i;

	char ext[_MAX_EXT];          // 256
#ifdef _WIN32
	char drive[_MAX_DRIVE];      // 3
	char dir[_MAX_DIR];          // 256
	char fname[_MAX_FNAME];      // 256

	// Dateiendung pruefen
	_splitpath(szFileName, drive, dir, fname, ext);
#else
	strcpy(ext, &szFileName[strlen(szFileName) - 4]);
#endif
	for (i=0; ext[i] && i < 4; i++)
	{
		ext[i] = tolower(ext[i]);
	}
	if (strcmp(ext, ".aes") != 0 && strcmp(ext, ".ael") != 0)
	{
		// Die Anwendung kann ueber den gewaehlten Weg, nur Dateien vom Typ (*.aes; *.ael) oeffnen.
		*piError = IV_ERROR_FILE_EXTENSION;
		return FALSE;
	}

	int res = 0;

	// Speicher initialisieren
	SetDocumentDefault(m_file);

	switch (GetProtocolVersion(GetDeviceIDRemote())) {
	case 1: 
	case 2: {
    if (strcmp(ext, ".ael") == 0)
    {
      // Nicht gueltige Endung der Setupdatei.
      *piError = IV_ERROR_FILE_EXTENSION;
      return 0;
    }

		res = DownloadSetupData(szFileName, piError);
		break;}

  case 4: { // PZEMasterIV mit Protokollversion 4

    char version[32];
		// Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
		if ((res = m_channel4.SetTimeboySlotState(GetDeviceIDRemote(), 1)) == IV_ERROR_SUCCESS)
		{
      memset(version, 0, sizeof(version));
      res = ResultToError(m_channel4.GetVersion(GetDeviceIDRemote(), version), piError);
      if (res == 0) {
        return 0;
      }
      // Pruefen ob die Firmwareversion unterstuetzt wird.
      if (!m_file.IsSupportedFirmwareVersion(version)) {
        *piError = IV_ERROR_VERSION_FIRMWARE;
        return 0;
      }

			// Befehlsausfuehrung starten.
			res = DownloadSetupData4(GetDeviceIDRemote());

			// Versuchen die rote LED abzuschalten.
			m_channel4.SetTimeboySlotState(GetDeviceIDRemote(), 0, res);
		}

		res = ResultToError(res, piError);

    // Setupdatei nach erfolgreichem Auslesen, abspeichern.
    if (res) {
      // Diese Version der Firmware der Setupdatei als Mindestversion vorgeben.
      CDFSoftwarebuildVersion fwVersion(version);
      m_file.NeedFirmwareVersion(fwVersion);

      res = m_file.Save(szFileName, piError);
    }
    break;}

	default:
		*piError = IV_ERROR_UNREACHABLE;
		res = 0;
	}

	return res;
}

int CAEChannel::DownloadSetupData(char *szFileName, int *piError)
{
  unsigned char temp = 1; // fuer Bedienstopp
  unsigned int ulTmp, ulSize = FLASH_SIZE_SU;

  // Bedienstopp setzen.
  if (!m_channel1_2.WriteMem(MB_CONTROL_STOP, 1, &temp, piError))
  {
    return 0;
  }

  // EEPROM - lesen 
  if (!m_channel1_2.ReadMem(MB_START_EEP, EEP_SIZE_SU, m_file.m_eep, piError))
  {
    return 0;
  }

  // Groesse der Setupdaten bereits im Eep?
  ulTmp = *(ULONG*)&m_file.m_eep[P_FLASH_SIZE_SU];
  ChangeOrderLong((unsigned char*)&ulTmp);
  // Liegt im moeglichen Bereich?
  if (!(ulTmp > FLASH_SIZE_SU) && !(ulTmp == 0))
  {
    ulSize = ulTmp;
  }

  if (ulSize > FLASH_SIZE_SU)
  {
    *piError = IV_ERROR_FILE_OPEN;
    return 0;
  }

  // Flashbereich - lesen
  if (!m_channel1_2.ReadMem(FLASH_KONFIG_START, ulSize, m_file.m_flash, piError))
  {
    return 0;
  }

  // Konfiguration - lesen
  if (!m_channel1_2.ReadMem(MB_START_CONFIG, CONFIG_SIZE_SU, m_file.m_configMem, piError))
  {
    return 0;
  }

  int i;
  TExtendedFlash extBegin;
  // gesetze Struktur uebernehmen
  extBegin = *(TExtendedFlash*)m_file.m_flashExtended;

  // IOModul - lesen
  extBegin.pulIOModule = extBegin.pulFreeScope;

  // IOModul - Konfiguration - lesen
  if (!m_channel1_2.ReadMem(MB_START_IOMODULE, sizeof(TIOModule), &m_file.m_flashExtended[extBegin.pulIOModule], piError))
  {
    return 0;
  }

  // Auf naechsten freien Speicher
  extBegin.pulFreeScope += sizeof(TIOModule);

  // Strukturen fuer Digitale Eingaenge des IOModules eintragen.
  extBegin.pulIODigital = extBegin.pulFreeScope;

  // IOModul - DigitalMDE - lesen
  for (i=0; i < 8; i++)
  {
    if (!m_channel1_2.ReadMem(MB_DIGITAL_IO + i * sizeof(TDigitalInput), sizeof(TDigitalInput), 
      &m_file.m_flashExtended[extBegin.pulIODigital + i * sizeof(TDigitalInput)], piError))
    {
      return 0;
    }
  }

  // Auf naechsten freien Speicher
  extBegin.pulFreeScope += 8 * sizeof(TDigitalInput);

  // Strukturen fuer Digitale Eingaenge der MDE eintragen.
  extBegin.pulMDEDigital = extBegin.pulFreeScope;

  // Interne MDE - DigitalMDE - lesen
  for (i=0; i < 4; i++)
  {
    if (!m_channel1_2.ReadMem(MB_DIGITAL_MDE + i * sizeof(TDigitalInput), sizeof(TDigitalInput), 
      &m_file.m_flashExtended[extBegin.pulMDEDigital + i * sizeof(TDigitalInput)], piError))
    {
      return 0;
    }
  }

  // Auf naechsten freien Speicher
  extBegin.pulFreeScope += 4 * sizeof(TDigitalInput);


  // Strukturen der Eingabeketten von MDE eintragen
  extBegin.pulEKArray = extBegin.pulFreeScope;

  // Groesse der MDE - Eingabeketten bereits im Eep?
  ulTmp = *(ULONG*)&m_file.m_eep[P_MDE_EK_SIZE];
  ChangeOrderLong((unsigned char*)&ulTmp);
  // Liegt im moeglichen Bereich?
  if (!(ulTmp > EXTENDED_SIZE_SU) && !(ulTmp == 0))
  {
    ulSize = ulTmp;
  }
  else
  {		
    ulSize = EXTENDED_SIZE_SU - extBegin.pulEKArray;
  }


  // MDE - Eingabeketten - lesen
  // Bei 4.0 kommt erst 1Byte Anzahl im Array
  // Ab 4.1 kommen erst 4Byte Anzahl im Array
  if (!m_channel1_2.ReadMem(MB_MDE_EK, ulSize, &m_file.m_flashExtended[extBegin.pulEKArray], piError))
  {
    return 0;
  }


  if (GetProtocolVersion2(GetDeviceIDRemote(), TRUE, piError) != 0x0400)
  {
    // Auf naechsten freien Speicher
    extBegin.pulFreeScope += GetMDEEKFlashSize();

    // Dieses sollte nicht sein!, fehlende MDE - Groesse.
    if (ulSize == (EXTENDED_SIZE_SU - extBegin.pulEKArray))
    {
      *piError = IV_ERROR_FILE_FORMAT;
      return 0;
    }

    // Da die Tabellendefinitionen fuer die Zutrittskontrolle eine
    // feste Struktur haben, wird aufgrund des 0 - Zeigers im Strukt
    // und des Flags fuer Auslesen, im Setup das Einfuegen vorgenommen.
    extBegin.pulEKArrayEntrance = extBegin.pulFreeScope;

    // Groesse der Zutrittskontrolle - Eingabeketten bereits im Eep?
    ulTmp = *(ULONG*)&m_file.m_eep[P_ENTRANCE_EK_SIZE];
    ChangeOrderLong((unsigned char*)&ulTmp);
    // Liegt im moeglichen Bereich?
    if (!(ulTmp > EXTENDED_SIZE_SU) && !(ulTmp == 0))
    {
      ulSize = ulTmp;
    }
    else
    {			
      *piError = IV_ERROR_FILE_FORMAT;
      return 0;
    }

    // Zutritts - Eingabeketten - lesen
    if (!m_channel1_2.ReadMem(MB_ENTRANCE_EK, ulSize, &m_file.m_flashExtended[extBegin.pulEKArrayEntrance], piError))
    {
      return 0;
    }

    // Auf naechsten freien Speicher
    extBegin.pulFreeScope += GetEntranceEKFlashSize();

    // Der Zeiger auf den naechsten freien Speicherplatz in der
    // erweiterten Struktur muss aufgrund der Abwaertskompatibilitaet
    // auf 0 gesetzt bleiben.
    *(TExtendedFlash*)m_file.m_flashExtended = extBegin;
  }
  else
  {
    // Auf naechsten freien Speicher
    extBegin.pulFreeScope += GetMDEEKFlashSize_0400();
  }

  temp = 0; // Bedienstopp ruecksetzen
  if (!m_channel1_2.WriteMem(MB_CONTROL_STOP, 1, &temp, piError))
  {
    return 0;
  }

  int res = 1;
  unsigned int sizeSetup, flashSizeSU2;
  // Setup-Datei oeffnen
  FILE *pFile;
  if ((pFile = fopen(szFileName, "wb")) == NULL)
  {
    *piError = IV_ERROR_FILE_OPEN;
    return 0;
  }

  // EEPROM - schreiben
  if (fwrite(m_file.m_eep, sizeof(unsigned char), EEP_SIZE_SU, pFile) != EEP_SIZE_SU)
  {
    // Fehler beim schreiben in die Datei.
    *piError = IV_ERROR_FILE_WRITE;
    //m_channel4.Trace("EEP konnte nicht geschrieben werden. ( %i )", ferror(pFile));
    res = 0;
    goto error_label;
  }

  // Dateikopf schreiben
  unsigned char mem[EEP_SIZE - EEP_SIZE_SU];
  memset(mem, 0, sizeof(mem));
  if (fwrite(mem, sizeof(unsigned char), sizeof(mem), pFile) != sizeof(mem))
  {
    // Fehler beim schreiben in die Datei.
    *piError = IV_ERROR_FILE_WRITE;
    //m_channel4.Trace("EEP konnte nicht geschrieben werden. ( %i )", ferror(pFile));
    res = 0;
    goto error_label;
  }

  // Groesse der Flashdaten fuer Setup aus EepMem entnehmen
  sizeSetup = *(unsigned int*)&m_file.m_eep[P_FLASH_SIZE_SU], flashSizeSU2 = 0;
  ChangeOrderLong((unsigned char*)&sizeSetup);
  // Groesse des zu schreibenden Flashbereichs ermitteln.
  if (sizeSetup > FLASH_SIZE_SU)
  {
    flashSizeSU2 = FLASH_SIZE_SU2;
  }

  if (res)
  {
    // Leere Filelist schreiben
    unsigned char buf[MAX_DATENSATZ_BESCHREIBUNG];
    memset(buf, 0, MAX_DATENSATZ_BESCHREIBUNG);

    // Flashbereich - schreiben
    if (fwrite(m_file.m_flash, sizeof(unsigned char), FLASH_SIZE_SU + flashSizeSU2, pFile) != (FLASH_SIZE_SU + flashSizeSU2) ||
      // Leere Filelist schreiben
      fwrite(buf, sizeof(unsigned char), MAX_DATENSATZ_BESCHREIBUNG, pFile) != MAX_DATENSATZ_BESCHREIBUNG ||
      // Konfiguration - schreiben
      fwrite(m_file.m_configMem, sizeof(unsigned char), CONFIG_SIZE_SU, pFile) != CONFIG_SIZE_SU ||
      // Erweiterter Flashbereich - schreiben
      fwrite(m_file.m_flashExtended, sizeof(unsigned char), EXTENDED_SIZE_SU, pFile) != EXTENDED_SIZE_SU ||
      // Leere Filelist fuer Zutrittskontrolle (Version 1)
      fwrite(buf, sizeof(unsigned char), 4, pFile) != 4 ||
      // Leere Filelist fuer Zutrittskontrolle (Version 2)
      fwrite(buf, sizeof(unsigned char), MAX_DATENSATZ_BESCHREIBUNG, pFile) != MAX_DATENSATZ_BESCHREIBUNG)
    {
      // Fehler beim schreiben in die Datei.
      *piError = IV_ERROR_FILE_WRITE;
      //m_channel4.Trace("Datenbereich konnte nicht geschrieben werden. ( %i )", ferror(pFile));
      res = 0;
      goto error_label;
    }
  }

error_label:
  if (pFile)
  {		
    // Datei schlieszen
    fclose(pFile);
  }

  return res;
}

int CAEChannel::DownloadSetupData4(short deviceIDRemote)
{
	int i, res = IV_ERROR_SUCCESS;
	unsigned char dummyConfig[0x1C]; // Groesse nicht veraendern !!!
	// Berechnet sich aus 0x80 - sizeof(TConfig)

	unsigned char *buf = NULL;
	unsigned int sizeSetup = 0;
	unsigned char countTable, countField;
	unsigned int sizeAccess;

	//
	unsigned int address, size, crc, crcTest = 0;
	address = size = crc = 0;

	// Segmentbegin: Bedienung stoppen
	if ((res = m_channel4.BeginGetSegment(deviceIDRemote, IV_VALUE_SEGMENT_GET_UPDATE_BG, address, size, crc)) != IV_ERROR_SUCCESS)
	{
		if ((res & 0xFF) == 1)
		{
			// Funktionsfehler Typ: IV_VALUE_SEGMENT_GET_UPDATE_BG unbekannt.
			// Wenn dass der Fall ist, wird das Auslesen noch nicht unterstuetzt.
			res = -IV_ERROR_FUNCTION_UNAVAILABLE;
		}
		else
		{
			return res;
		}
	}
	m_channel4.Trace("%s(%i):INFO(1): CAEChannel::DownloadSetupData4 [ BeginGetSegment(%i): %i, %i, 0x%08X ]\n", __FILE__, __LINE__, IV_VALUE_SEGMENT_GET_UPDATE_BG, address, size, crc);

	//////////////////////////////
	// Begin

	// Segmentbegin: Setup uebertragen
	if ((res = m_channel4.BeginGetSegment(deviceIDRemote, IV_VALUE_SEGMENT_GET_SETUP, address, size, crc)) != IV_ERROR_SUCCESS)
	{
		return res;
	}
	m_channel4.Trace("%s(%i):INFO(1): CAEChannel::DownloadSetupData4 [ BeginGetSegment(%i): %i, %i, 0x%08X ]\n", __FILE__, __LINE__, IV_VALUE_SEGMENT_GET_SETUP, address, size, crc);

	// Vorhandene zur moeglichen Groesse pruefen
	if (size < (EEP_SIZE_SU + (sizeof(TConfig) + 0x1C)))
	{
		return -IV_ERROR_PROCESSING;
	}

	//////////////////////////////
	// Grunddaten

	// EEPROM uebertragen
	if ((res = m_channel4.GetFlash(deviceIDRemote, address, m_file.m_eep, EEP_SIZE_SU)) != IV_ERROR_SUCCESS)
	{
		return res;
	}
	address += EEP_SIZE_SU;
	// CONFIG uebertragen (100 Byte)
	if ((res = m_channel4.GetFlash(deviceIDRemote, address, m_file.m_configMem, sizeof(TConfig))) != IV_ERROR_SUCCESS)
	{
		return res;
	}
	address += sizeof(TConfig);
	// CONFIG dummybytes (28 Byte)
	if ((res = m_channel4.GetFlash(deviceIDRemote, address, dummyConfig, sizeof(dummyConfig))) != IV_ERROR_SUCCESS)
	{
		return res;
	}
	address += sizeof(dummyConfig);

	// Groesse der Flashdaten fuer Setup aus EepMem entnehmen
	ChangeOrderLong(&m_file.m_eep[P_FLASH_SIZE_SU]);
	sizeSetup = *(unsigned int*)&m_file.m_eep[P_FLASH_SIZE_SU];
	ChangeOrderLong(&m_file.m_eep[P_FLASH_SIZE_SU]);

	// FLASH uebertragen
	if ((res = m_channel4.GetFlash(deviceIDRemote, address, m_file.m_flash, sizeSetup)) != IV_ERROR_SUCCESS)
	{
		return res;
	}
	address += sizeSetup;


	unsigned int sizeEK;

	// Groesse der Zutrittskontroll - Eingabeketten aus EepMem entnehmen.
	ChangeOrderLong(&m_file.m_eep[P_ENTRANCE_EK_SIZE]);
	sizeEK = *(unsigned int*)&m_file.m_eep[P_ENTRANCE_EK_SIZE];
	ChangeOrderLong(&m_file.m_eep[P_ENTRANCE_EK_SIZE]);

	// Speicher reservieren
	buf = new unsigned char[sizeEK + 1];
	// Zeigeroffsets fuer Firmware anpassen
	if (buf == NULL)
	{
		// Speicherallokation nicht moeglich
		return -IV_ERROR_ALLOC_MEM;
	}
	memset(buf, 0, sizeEK + 1);

	//////////////////////////////
	// Pointerarray + Strukturen der Zutrittslisten uebertragen

	// Eingabekette der Zutrittskontrolle
	if ((res = m_channel4.GetFlash(deviceIDRemote, address, buf, sizeEK)) != IV_ERROR_SUCCESS)
	{
		delete[] buf; 
		buf = NULL;
		return res;
	}
	address += sizeEK;

	// Pointerarray lesen
	if ((res = m_channel4.GetFlash(deviceIDRemote, address, (unsigned char*)m_entrance2Pointer, sizeof(m_entrance2Pointer))) != IV_ERROR_SUCCESS)
	{
		delete[] buf; 
		buf = NULL;
		return res;
	}
	address += sizeof(m_entrance2Pointer);

	///////////////
	// ZK II

	// Anzahl der Zutrittslisten ermitteln
	for (i=0, countTable=0; countTable < (sizeof(m_entrance2Pointer) / sizeof(unsigned int)); countTable++)
	{
		if (m_entrance2Pointer[countTable] == 0)
		{
			break;
		}

		// Gesamte Feldanzahl fuer Groeszenberechnung sammeln.
		i += m_entrance2Table[countTable].count;
	}
	sizeAccess = sizeof(m_entrance2Pointer);
	sizeAccess += countTable * sizeof(STable2);
	sizeAccess += i * sizeof(STable2Field);

	// Strukturen lesen
	for (i=0; i < countTable; i++)
	{
		// Tabellenstruktur uebertragen
		if ((res = m_channel4.GetFlash(deviceIDRemote, address, (unsigned char*)&m_entrance2Table[i], sizeof(STable2))) != IV_ERROR_SUCCESS)
		{
			delete[] buf; 
			buf = NULL;
			return res;
		}
		address += sizeof(STable2);

		// Feldstrukturen lesen
		if ((res = m_channel4.GetFlash(deviceIDRemote, address, (unsigned char*)m_entrance2TableField[i], m_entrance2Table[i].count * sizeof(STable2Field))) != IV_ERROR_SUCCESS)
		{
			delete[] buf; 
			buf = NULL;
			return res;
		}
		address += m_entrance2Table[i].count * sizeof(STable2Field);
	}


	///////////////
	// Grunddaten

	// CRC berechnen
	crcTest = 0xFFFFFFFF;
	crcTest = m_channel4.GetCRC32(m_file.m_eep, EEP_SIZE_SU, crcTest, FALSE);
	crcTest = m_channel4.GetCRC32(m_file.m_configMem, sizeof(TConfig), crcTest, FALSE);
	crcTest = m_channel4.GetCRC32(dummyConfig, sizeof(dummyConfig), crcTest, FALSE);
	crcTest = m_channel4.GetCRC32(m_file.m_flash, sizeSetup, crcTest, FALSE);
	crcTest = m_channel4.GetCRC32(buf, sizeEK, crcTest, FALSE);

	// Speicher wieder freigeben.
	delete[] buf; 
	buf = NULL;

	// CRC der Zutrittslisten ermitteln
	crcTest = m_channel4.GetCRC32((unsigned char*)m_entrance2Pointer, sizeof(m_entrance2Pointer), crcTest, FALSE);

	for (i=0; i < countTable; i++)
	{
		crcTest = m_channel4.GetCRC32((unsigned char*)&m_entrance2Table[i], sizeof(STable2), crcTest, FALSE);

		for (countField=0; countField < m_entrance2Table[i].count; countField++)
		{
			// Beim letzten Feld der letzten Tabelle musz der CRC abgeschlossen werden.
			crcTest = m_channel4.GetCRC32((unsigned char*)&m_entrance2TableField[i][countField], sizeof(STable2Field), crcTest, ((i == (countTable - 1)) && (countField == (m_entrance2Table[i].count - 1))));
		}
	}

	//////////////////////////////
	// Ende

	if (crc != crcTest)
	{
		m_channel4.Trace("%s(%i):ERROR(1): CAEChannel::DownloadSetupData4 [ CRC von Segment falsch: 0x%08X, 0x%08X ]\n", __FILE__, __LINE__, crc, crcTest);
		return -IV_ERROR_CRC_SEGMENT;
	}

	// Segmentend: Setup uebertragen
	if ((res = m_channel4.EndGetSegment(deviceIDRemote, IV_VALUE_SEGMENT_GET_SETUP)) != IV_ERROR_SUCCESS)
	{
		return res;
	}
	m_channel4.Trace("%s(%i):INFO(1): CAEChannel::DownloadSetupData4 [ EndGetSegment(%i) ]\n", __FILE__, __LINE__, IV_VALUE_SEGMENT_GET_SETUP);

	//////////////////////////////
	// Begin
	address = size = crc = 0;
	// Segmentbegin: Setup uebertragen
	if ((res = m_channel4.BeginGetSegment(deviceIDRemote, IV_VALUE_SEGMENT_GET_SETUP_EXTENDED, address, size, crc)) != IV_ERROR_SUCCESS)
	{
		return res;
	}
	m_channel4.Trace("%s(%i):INFO(1): CAEChannel::DownloadSetupData4 [ BeginGetSegment(%i): %i, %i, 0x%08X ]\n", __FILE__, __LINE__, IV_VALUE_SEGMENT_GET_SETUP_EXTENDED, address, size, crc);

	//////////////////////////////
	// Extended - Speicher Lesen

	if ((res = m_channel4.GetFlash(deviceIDRemote, address, m_file.m_flashExtended, size)) != IV_ERROR_SUCCESS)
	{
		return res;
	}
	address += size;

	///////////////
	// Extended - Speicher pruefen

	// Extended - Speicher
	crcTest = 0xFFFFFFFF;
	crcTest = m_channel4.GetCRC32(m_file.m_flashExtended, size, crcTest, TRUE);

	if (crc != crcTest)
	{
		m_channel4.Trace("%s(%i):ERROR(1): CAEChannel::DownloadSetupData4 [ CRC von Segment falsch: 0x%08X, 0x%08X ]\n", __FILE__, __LINE__, crc, crcTest);
		return -IV_ERROR_CRC_SEGMENT;
	}

	//////////////////////////////
	// Ende

	// Segmentend: Setup uebertragen
	if ((res = m_channel4.EndGetSegment(deviceIDRemote, IV_VALUE_SEGMENT_GET_SETUP_EXTENDED)) != IV_ERROR_SUCCESS)
	{
		return res;
	}
	m_channel4.Trace("%s(%i):INFO(1): CAEChannel::DownloadSetupData4 [ EndGetSegment(%i) ]\n", __FILE__, __LINE__, IV_VALUE_SEGMENT_GET_SETUP_EXTENDED);


	// Die Eingabekettenzeiger anpassen, da diese fuer die FW vor dem Einspielen angepasst wurden.
	// Zeigeroffsets anpassen (addieren von Offset)
	AdjustExtendedMemPointer();

	crc = 0;
	// Segmentend: Bedienung stoppen
	if ((res = m_channel4.EndGetSegment(deviceIDRemote, IV_VALUE_SEGMENT_GET_UPDATE_BG)) != IV_ERROR_SUCCESS)
	{
		return res;
	}
	m_channel4.Trace("%s(%i):INFO(1): CAEChannel::DownloadSetupData4 [ EndGetSegment(%i) ]\n", __FILE__, __LINE__, IV_VALUE_SEGMENT_GET_UPDATE_BG);

	return res;
}

int CAEChannel::ModifyStudioFile(char *sourceFile, char *destFile, char *changeData, int flags, int *piError)
{
	CDFFileStudioXML file;
  int fileError = 0;
	struct stat st;
	
	if (stat(sourceFile, &st) != 0)
	{		
		// Datei konnte nicht ermittelt werden
		*piError = IV_ERROR_FILE_OPEN;
		return 0;
	}

  // Reset der StudioFile-Klasse
  SetDocumentDefault(file);

  // Laden der Dateidaten. Keine Beachtung der Studioversion.
  if (!file.Load(sourceFile, &fileError, AESHEADER_IGNORE_MIN_STUDIO_VERSION | AESHEADER_IGNORE_SHOULD_STUDIO_VERSION)) {
    switch (fileError) {
      case -3:
        *piError = IV_ERROR_FILE_READ;
        break;
      case -7: // Mindestversion DLL wird nicht eingehalten.
        *piError = IV_ERROR_SETUPFILE_MIN_LIBRARY_VERSION;
        break;
      case -8: // Sollteversion DLL
        *piError = IV_ERROR_SETUPFILE_SHOULD_LIBRARY_VERSION;
        break;
      case -9: // Format der Setupdatei
        *piError = IV_ERROR_FILE_FORMAT;
        break;
      case -4: // CRC in Dateidaten
      case -10: case -11: case -12: case -13: case -14: case -15: case -16: case -17: case -18:
        *piError = IV_ERROR_FILE_CRC;
        break;
      default:
        *piError = IV_ERROR_FILE_OPEN;
    }
    return 0;
  }

	// StudioDatei modifizieren
	if (file.Modify(changeData, piError))
	{
		// StudioDatei speichern
		if (file.Save(destFile, &fileError))
		{
			// Alles erfolgreich durchgefuehrt
			return 1;
		}
		else
		{
			// Datei konnte nicht erstellt werden.
			*piError = IV_ERROR_FILE_OPEN;
		}
	}

	return 0;
}

int CAEChannel::CheckDevice(short deviceIDRemote, int *piError, int iPollRetry)
{
	int iRes = IV_ERROR_SUCCESS;

	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(GetDeviceIDRemote()))
	{
	case 1:
	case 2:
		for (m_iDevicePollRetry=1; m_iDevicePollRetry <= iPollRetry; m_iDevicePollRetry++)
		{
			int tmp;
			if (m_channel1_2.ReadMem(0ul, 2, (unsigned char*)&tmp, piError))
			{
				return 1;
			}

			Sleep(20);
		}
		break;

	case 4: { // PZEMasterIV mit Protokollversion 4
		char szVersion[32];

		// Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
		if ((iRes = m_channel4.SetTimeboySlotState(GetDeviceIDRemote(), 1)) == IV_ERROR_SUCCESS)
		{
			// Befehlsausfuehrung starten.
			iRes = m_channel4.GetVersion(GetDeviceIDRemote(), szVersion);

			// Versuchen die rote LED abzuschalten.
			m_channel4.SetTimeboySlotState(GetDeviceIDRemote(), 0, iRes);
		}

		return ResultToError(iRes, piError);}

	default:
		*piError = IV_ERROR_UNREACHABLE;
	}

	// Letzter Fehler ist aktueller
	return 0;
}

int CAEChannel::Restart(short deviceIDRemote, int *piError)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	int tmp = 0;

	//int iRes = IV_ERROR_SUCCESS;
	switch (GetProtocolVersion(GetDeviceIDRemote()))
	{
	case 2:
		if (m_channel1_2.WriteMem(MB_RESTART, 1, (unsigned char*)&tmp, piError))
		{
			return 1;
		}
		return 0;

	case 1:
	case 4:
		*piError = IV_ERROR_VERSION_FIRMWARE;
		break;

	default:
		*piError = IV_ERROR_UNREACHABLE;
	}

	// Letzter Fehler ist aktueller
	return 0;
}

int CAEChannel::Reset(short deviceIDRemote, int mode, int *piError)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	int iRes = IV_ERROR_SUCCESS;
	switch (GetProtocolVersion(GetDeviceIDRemote()))
	{
	case 1:
	case 2:
		*piError = IV_ERROR_VERSION_FIRMWARE;
		break;

	case 4:
		// Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
		if ((iRes = m_channel4.SetTimeboySlotState(GetDeviceIDRemote(), 1)) == IV_ERROR_SUCCESS)
		{
			// Befehlsausfuehrung starten.
			iRes =  m_channel4.SystemReset(deviceIDRemote, mode);

			// Versuchen die rote LED abzuschalten.
			m_channel4.SetTimeboySlotState(GetDeviceIDRemote(), 0, iRes);
		}

		iRes = ResultToError(iRes, piError);
		break;

	default:
		*piError = IV_ERROR_UNREACHABLE;
	}

	// Letzter Fehler ist aktueller
	return iRes;
}

int CAEChannel::GetProtocolVersion2(short deviceIDRemote, unsigned char detailed, int *piError)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	unsigned char Buf[4];
	
	// Protokollverwendung pruefen
	unsigned char pv = GetProtocolVersion(GetDeviceIDRemote());
	switch(pv)
	{
	case 1:
	case 2:
		// Detailierte Version auslesen.
		if (detailed)
		{
			// Version fuer Setup lesen
			if (!m_channel1_2.ReadMem(MB_GET_VERSION, 2, Buf, piError))
			{
				return 0;
			}
		}
		else 
		{
			if (pv == 1)
			{
				// Es wird angenommen ein AEIII+ mit neuester Com-Version
				Buf[0] = 0x04;
				Buf[1] = 0x02;
			}
			else
			{
				// Es wird angenommen ein PZE Master IV mit neuester Com-Version
				Buf[0] = 0x05;
				Buf[1] = 0x01;
			}
		}
		break;

	case 3: 
		// Ab dem Protokolltyp 3 wird nur noch ueber die Firmwareversion entschieden.
		Buf[0] = 0x06;
		Buf[1] = 0x00;
		break;

	case 4:
		// Ab dem Protokolltyp 4 wird nur noch ueber die Firmwareversion entschieden.
		Buf[0] = 0x07;
		Buf[1] = 0x00;
		break;

	default:
		*piError = IV_ERROR_UNREACHABLE;
		return 0;
	}
	

	// Version fuer Kommunikation
	
	// Ab Firmware 04.01.00 Protokollanpassung (Version 4) und Listen in Tabellenart abhandeln.
	if ((Buf[0] == 7 && Buf[1] == 0) ||
		// Ab Firmware 04.01.00 Protokollanpassung (Version 3)
		(Buf[0] == 6 && Buf[1] == 0) ||
		// Ab Firmware 04.00.23 Listengroesse auf 256kByte erhoeht.
		(Buf[0] == 5 && Buf[1] == 2) ||
		// Ab Firmware 04.00.21 Listen im Hintergrund laden
		(Buf[0] == 5 && Buf[1] == 1) ||
		// Ab Firmware 04.00.19 des PZEMasterIV (Version 2)
		(Buf[0] == 5 && Buf[1] == 0) ||
		// Ab Firmware 03.01.21 wegen auslesen der Firmwareversion
		(Buf[0] == 4 && Buf[1] == 2) ||
		// Ab Firmware 03.00.12 mit Zutrittskontrolle
		(Buf[0] == 4 && Buf[1] == 1) ||
	    // Ab Firmware 3.0 mit 16kByte Flash fuer Setupdaten und 16kByte Extended Flash
		(Buf[0] == 4 && Buf[1] == 0) )
	{ 
		// Version als Rueckgabe
		return (Buf[0] << 8) | Buf[1];
	}

	*piError = IV_ERROR_VERSION_FIRMWARE;
	return 0;
}

int CAEChannel::GetSerialNum(short deviceIDRemote, int *piSerialNum, int *piError)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	///////////////////////////////////////////////////
	// TESTBESCHREIBUNG: OB TBIII AUF TBIV ADRESSE REAGIERT
	// ERGEBNISS: TBIII ANTWORTET NICHT AUF TBIV ADRESSBEREICH

	//Close();
	//unsigned char buf[100];
	//memset(buf, 0, sizeof(buf));
	//int error, serial = 0;
	//CCom *com = new CCom(m_channel4.GetID());
	//com->ComOpenSerial("COM1", 9600, 1000);
	////com->SetComPort(m_channel4.m_comm, 0, 9600, 2000);
	//com->ChangeBusNum(1);
	//com->GetSerienNum(1, &error, &serial);
	////com->ReadMem(0x70000000, 10, buf, 3);
	//TRACE("WriteMem 1: %i (0x%02X)\n", com->ReadMem(0x49020000, 1, buf, 3), buf[0]);
	//TRACE("WriteMem 2: %i\n", com->ReadMem(0x70000000, 10, buf, 3));
	////com->WriteMem(0x70000000, 10, buf, 3);
	////com->SetComPort(NULL, 0, 0, 0);
	//com->ComClose();
	//*piSerialNum = serial;
	//delete com;
	//return 1;

	switch (GetProtocolVersion(GetDeviceIDRemote()))
	{
	case 1: 
	case 2: {
		unsigned char buf[10];

		// Seriennummer und Optionsbyte lesen
		if (!m_channel1_2.ReadMem(MB_SERIEN_MEMORY, 4, buf, piError))
		{
			return 0;
		}

		*piSerialNum = (int)(buf[2] << 8);
		*piSerialNum |= (int)(buf[3]);
		return 1;}

	case 4: {
		int iRes;
		unsigned int serial;

		// Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
		if ((iRes = m_channel4.SetTimeboySlotState(GetDeviceIDRemote(), 1)) == IV_ERROR_SUCCESS)
		{
			// Befehlsausfuehrung starten.
			iRes = m_channel4.GetSerial(GetDeviceIDRemote(), serial);

			// Versuchen die rote LED abzuschalten.
			m_channel4.SetTimeboySlotState(GetDeviceIDRemote(), 0, iRes);
		}

		iRes = ResultToError(iRes, piError);
		if (iRes)
		{
			// OK
			*piSerialNum = serial;
		}

		return iRes;}

	default:
		*piError = IV_ERROR_UNREACHABLE;
	}

	return 0;
}

// ListenBuffer in AE laden
int CAEChannel::LoadListData(short deviceIDRemote, int *piError)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	int i;
	// Listengesamtgroesse berechnen und ggf. sortieren.
	unsigned int listSize = MAX_DATENSATZ_BESCHREIBUNG * sizeof(unsigned int);
	unsigned int listPointer[MAX_DATENSATZ_BESCHREIBUNG];
	for (i=0; i < MAX_DATENSATZ_BESCHREIBUNG; i++)
	{
		if (m_listSize[i])
		{			
			listPointer[i] = listSize + FLASH_LIST_START;
			ChangeOrderLong((unsigned char*)&listPointer[i]);

			listSize += m_listSize[i];
		}
		else
		{
			listPointer[i] = 0;
		}
	}

	if (listSize == (MAX_DATENSATZ_BESCHREIBUNG * sizeof(unsigned int)))
	{
		*piError = IV_ERROR_LIST_PROCESSING;
		return 0; // keine Listen vorhanden
	}

	switch (GetProtocolVersion(GetDeviceIDRemote()))
	{
	case 1: 
	case 2: {

		unsigned int listStart = FLASH_LIST_START;
		unsigned char temp = 1; // Bedienstopp setzen
		// Der Bedienstopp wird nur noch bei aelteren Versionen gesetzt.
		// Wenn der Befehl zum Bedienstopp aufheben gegeben wird,
		// dann werden die zuvor uebertragenen Daten an die entsprechende Position kopiert.
		switch (GetProtocolVersion2(GetDeviceIDRemote(), 1, piError))
		{
		case 0: // Kommunikationsfehler
			return 0;

		case 0x0501:
			// Listengroesse nur maximal 100kByte
			if (listSize >= FLASH_SIZE_LIST)
			{
				*piError = IV_ERROR_LIST_PROCESSING;
				return 0;
			}
			break;

		case 0x0502:
			listStart = FLASH_LIST_START_NEW;

			// Listengroesse nur maximal 244kByte
			if (listSize >= FLASH_SIZE_LIST_NEW)
			{
				*piError = IV_ERROR_LIST_PROCESSING;
				return 0;
			}
			break;

		default:
			// Listengroesse nur maximal 100kByte
			if (listSize >= FLASH_SIZE_LIST)
			{
				*piError = IV_ERROR_LIST_PROCESSING;
				return 0;
			}
			if (!m_channel1_2.WriteMem(MB_CONTROL_STOP, 1, &temp, piError))
			{
				return 0;
			}
		}

		// Flash schreiben 
		if (!m_channel1_2.WriteMem(listStart, (MAX_DATENSATZ_BESCHREIBUNG * sizeof(unsigned int)), (unsigned char*)listPointer, piError))
		{
			return 0;
		}

		unsigned int tmpSize = 0;
		for (i=0; i < MAX_DATENSATZ_BESCHREIBUNG; i++)
		{
			if (m_listSize[i])
			{
				// Datensatzanzahl drehen
				ChangeOrderLong((unsigned char*)&m_pucListMem[i][0]);

				// Flash schreiben 
				if (!m_channel1_2.WriteMem(listStart + (MAX_DATENSATZ_BESCHREIBUNG * sizeof(unsigned int)) + tmpSize, m_listSize[i], m_pucListMem[i], piError))
				{
					// Datensatzanzahl drehen
					ChangeOrderLong((unsigned char*)&m_pucListMem[i][0]);
					return 0;
				}

				// Datensatzanzahl drehen
				ChangeOrderLong((unsigned char*)&m_pucListMem[i][0]);
				tmpSize += m_listSize[i];
			}
		}

		temp = 0; // Bedienstopp ruecksetzen
		if (!m_channel1_2.WriteMem(MB_CONTROL_STOP, 1, &temp, piError))
		{
			return 0;
		}	

		return 1;}
		
	case 4: { // PZEMasterIV mit Protokollversion 4
		
		int iRes = IV_ERROR_SUCCESS;

		// Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
		if ((iRes = m_channel4.SetTimeboySlotState(GetDeviceIDRemote(), 1)) == IV_ERROR_SUCCESS)
		{
			// Befehlsausfuehrung starten.
			iRes = LoadListData4(GetDeviceIDRemote());

			// Versuchen die rote LED abzuschalten.
			m_channel4.SetTimeboySlotState(GetDeviceIDRemote(), 0, iRes);
		}

		if ((iRes > 0) && (((iRes >> 8) & 0xFF) == 3))
		{
			*piError = IV_ERROR_NOT_ENOUGH_FLASH; // Es steht nicht genuegend freier Flash Speicher zur Verfuegung.
			iRes = 0;
		}
		else
		{
			iRes = ResultToError(iRes, piError);
		}

		return iRes;}

	default:
		*piError = IV_ERROR_UNREACHABLE;
	}

	return 0;
}

// Fuer neue Listentechnik
int CAEChannel::LoadListData4(short deviceIDRemote)
{
	unsigned char tableNum;
	int res = IV_ERROR_SUCCESS;

	STable2 table;
	char name[32];

	//
	unsigned short segmentUpdateVariation = IV_VALUE_SEGMENT_UPDATE_BG;
	unsigned int address, size, crc;
	address = size = crc = 0;

	// Systemstopp Listen uebertragen
	while ((res = m_channel4.BeginSetSegment(deviceIDRemote, segmentUpdateVariation, address, size, crc)) != IV_ERROR_SUCCESS)
	{
		// Wenn das Hintergrunduebertragen nicht unterstuetzt wird, dann die Standardmethode verwenden.
		if (((res & 0xFF) == 1) && (segmentUpdateVariation == IV_VALUE_SEGMENT_UPDATE_BG))
		{
			segmentUpdateVariation = IV_VALUE_SEGMENT_UPDATE;
			continue;
		}

		return res;
	}

	// Listendaten Tabellenweise uebertragen
	for (tableNum=0; tableNum < MAX_DATENSATZ_BESCHREIBUNG; tableNum++)
	{
		// Nicht importierte Listen uebergehen
		if (m_listSize[tableNum] == 0) continue;

		// Maximale Segmentanzahl nicht ueberschreiten
		if (tableNum && !(tableNum % 16))
		{
			crc = 0;
			// Systemstopp Listen uebertragen
			if ((res = m_channel4.EndSetSegment(deviceIDRemote, segmentUpdateVariation, crc)) != IV_ERROR_SUCCESS)
			{
				return res;
			}

			// Systemstopp Listen uebertragen
			if ((res = m_channel4.BeginSetSegment(deviceIDRemote, segmentUpdateVariation, address, size, crc)) != IV_ERROR_SUCCESS)
			{
				return res;
			}
		}

		memset(name, 0, sizeof(name));
		sprintf(name, "LIST.%i", tableNum);
		// Zur Liste Informationen ermitteln
		if ((res = m_channel4.GetTable(deviceIDRemote, name, table)) != IV_ERROR_SUCCESS)
		{
			return res;
		}
		strcpy(name, table.name);
		sprintf(table.name, "LIST.%s", name);

		// Muss die Liste auf Sortierung geprueft und ggf. sortiert werden.
		if (m_listSortCheck[tableNum] == false)
		{
			int fieldOffs = 0;

			if (table.flags & TF_SORT && table.sort < table.count)
			{
				char name[64];
				int fieldNum;
				STable2Field field;
				memset(&field, 0, sizeof(STable2Field));
				m_ucListRecordSize[tableNum] = 0;
				// Alle Felder fuer Offset lesen
				for (fieldNum=0; fieldNum < table.count; fieldNum++)
				{
					if (fieldNum < (table.sort + 1))
					{
						// Datensatzoffset fuer Sortierfeld aufsummieren.
						fieldOffs += field.size;
					}

					sprintf(name, "LIST.%i.%i", tableNum, fieldNum);

					if ((res = m_channel4.GetField(deviceIDRemote, name, field)) != IV_ERROR_SUCCESS)
					{
						if (res < 0)
						{
							// Kommunikationsfehler
							return res;
						}
						else
						{
							// Funktionsfehler
							return res;
						}
					}

					// Datensatzgroesse fuer Swap-Routine berechnen
					m_ucListRecordSize[tableNum] += field.size;
				}
/*
				int x, y;
				for (x=0; x < 10 && x <= *(unsigned int*)&m_pucListMem[tableNum]; x++)
				{
					TRACE("Unsort(%i.%i): ", tableNum, x+1);
					for (y=0; y < m_ucListRecordSize[tableNum] + sizeof(unsigned int); y++)
					{
						if (m_pucListMem[tableNum][x * m_ucListRecordSize[tableNum] + sizeof(unsigned int) + y])
						{
							TRACE("%c", m_pucListMem[tableNum][x * m_ucListRecordSize[tableNum] + sizeof(unsigned int) + y]);
						}
						else
						{
							TRACE(" ");
						}
					}
					TRACE("\n");
				}
*/
				// Liste sortieren.
				//QuicksortListRecord(tableNum, fieldOffs, 0, *(unsigned int*)m_pucListMem[tableNum] - 1);
				//BubbleListSort(tableNum, fieldOffs, 0, *(unsigned int*)m_pucListMem[tableNum] - 1);
				QuicksortList(m_pucListMem[tableNum], m_ucListRecordSize[tableNum], fieldOffs, 0, *(unsigned int*)m_pucListMem[tableNum] - 1);

/*
				for (x=0; x < 10 && x <= *(unsigned int*)&m_pucListMem[tableNum]; x++)
				{
					TRACE("Sort(%i.%i): ", tableNum, x+1);
					for (y=0; y < m_ucListRecordSize[tableNum] + sizeof(unsigned int); y++)
					{
						if (m_pucListMem[tableNum][x * m_ucListRecordSize[tableNum] + sizeof(unsigned int) + y])
						{
							TRACE("%c", m_pucListMem[tableNum][x * m_ucListRecordSize[tableNum] + sizeof(unsigned int) + y]);
						}
						else
						{
							TRACE(" ");
						}
					}
					TRACE("\n");
				}
*/
				// Bei allen Listen ist 0 die sortierte Spalte
				// Flag dafuer dass die Liste sortiert wurde setzen.
				table.flags |= TF_SORTED;
			}

			// Sortiert oder zumindes geprueft.
			m_listSortCheck[tableNum] = true;
		}

		// CRC berechnen
		crc = 0xFFFFFFFF;
		crc = m_channel4.GetCRC32((unsigned char*)&table, sizeof(STable2), crc, FALSE);
		crc = m_channel4.GetCRC32(m_pucListMem[tableNum], m_listSize[tableNum], crc);

		// Groesze berechnen
		size = sizeof(STable2) + m_listSize[tableNum];

		// Systemstopp Liste uebertragen
		if ((res = m_channel4.BeginSetSegment(deviceIDRemote, IV_VALUE_SEGMENT_LIST, address, size, crc)) != IV_ERROR_SUCCESS)
		{
			return res;
		}

		// Daten uebertragen
		if ((res = m_channel4.SetFlash(deviceIDRemote, address, (unsigned char*)&table, sizeof(STable2))) != IV_ERROR_SUCCESS)
		{
			// Speicherplatz reicht nicht aus.
			if (res == 3) { res = -IV_ERROR_NOT_ENOUGH_FLASH; }
			return res;
		}
		// Daten uebertragen
		address += sizeof(STable2);
		if ((res = m_channel4.SetFlash(deviceIDRemote, address, m_pucListMem[tableNum], m_listSize[tableNum])) != IV_ERROR_SUCCESS)
		{
			// Speicherplatz reicht nicht aus.
			if (res == 3) { res = -IV_ERROR_NOT_ENOUGH_FLASH; }
			return res;
		}

		// Systemstopp Liste uebertragen
		if ((res = m_channel4.EndSetSegment(deviceIDRemote, IV_VALUE_SEGMENT_LIST, crc)) != IV_ERROR_SUCCESS)
		{
			return res;
		}
	}

	crc = 0;
	// Systemstopp Liste uebertragen
	if ((res = m_channel4.EndSetSegment(deviceIDRemote, segmentUpdateVariation, crc)) != IV_ERROR_SUCCESS)
	{
		return res;
	}

	return res;
}

// ListenBuffer in AE laden
int CAEChannel::LoadEntranceListData(short deviceIDRemote, int iListNum, int *piError)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	// Alle Listen laden?
	if (iListNum == -1)
	{
		// Mindestens eine Liste geladen?
		int i;
		for (i=0; i < 4; i++)
		{
			if (m_ulEntranceListSize[i] != 0)
			{
				break;
			}
		}
		if (i == 4)
		{
			*piError = IV_ERROR_LIST_PROCESSING;
			return 0;
		}
	}
	else
	{
		if (m_ulEntranceListSize[iListNum] == 0)
		{
			*piError = IV_ERROR_LIST_PROCESSING ;
			return 0; // keine Listen vorhanden
		}
	}

	// Protokollklasse
	switch (GetProtocolVersion(GetDeviceIDRemote()))
	{
	case 4:
		*piError = IV_ERROR_FUNCTION_UNAVAILABLE;
		return 0;

	case 1:
	case 2:
		break;

	default:
		*piError = IV_ERROR_UNREACHABLE;
		return 0;
	}

	unsigned char temp = 1; // Bedienstopp setzen
	// Der Bedienstopp wird nur noch bei aelteren Versionen gesetzt.
	// Wenn der Befehl zum Bedienstopp aufheben gegeben wird,
	// dann werden die zuvor uebertragenen Daten an die entsprechende Position kopiert.
	switch (GetProtocolVersion2(GetDeviceIDRemote(), 1, piError))
	{
	case 0x0501:
	case 0x0502:
		break;

	default:
		if (!m_channel1_2.WriteMem(MB_CONTROL_STOP, 1, &temp, piError))
		{
			return 0;
		}
	}

	switch (iListNum)
	{
	case -1:
	case 0:
		if (m_ulEntranceListSize[0] != 0) 
		{
			// Flash schreiben 
			if (!m_channel1_2.WriteMem(MB_ENTRANCE_TABLE_0, m_ulCompressedListSize[0], m_pucCompressedListMem[0], piError, 32))
			{
				return 0;
			}
		}
		if (iListNum != -1) break;
	case 1:
		if (m_ulEntranceListSize[1] != 0)
		{
			// Flash schreiben 
			if (!m_channel1_2.WriteMem(MB_ENTRANCE_TABLE_1, m_ulCompressedListSize[1], m_pucCompressedListMem[1], piError, 32))
			{
				return 0;
			}
		}
		if (iListNum != -1) break;
	case 2:
		if (m_ulEntranceListSize[2] != 0)
		{
			// Flash schreiben 
			if (!m_channel1_2.WriteMem(MB_ENTRANCE_TABLE_2, m_ulCompressedListSize[2], m_pucCompressedListMem[2], piError, 32))
			{
				return 0;
			}
		}
		if (iListNum != -1) break;
	case 3:
		if (m_ulEntranceListSize[3] != 0)
		{
			// Flash schreiben 
			if (!m_channel1_2.WriteMem(MB_ENTRANCE_TABLE_3, m_ulCompressedListSize[3], m_pucCompressedListMem[3], piError, 32))
			{
				return 0;
			}
		}
	}

	temp = 0; // Bedienstopp ruecksetzen
	if (!m_channel1_2.WriteMem(MB_CONTROL_STOP, 1, &temp, piError))
	{
		return 0;
	}

	return 1;
}

// ListenBuffer in AE laden
int CAEChannel::LoadEntrance2ListData(short deviceIDRemote, int iListNum, int *piError)
{
	int iRes = IV_ERROR_SUCCESS;

	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(GetDeviceIDRemote()))
	{
	case 1:
	case 2:
		*piError = IV_ERROR_VERSION_FIRMWARE;
		return 0;

	case 4: {

		// Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
		if ((iRes = m_channel4.SetTimeboySlotState(GetDeviceIDRemote(), 1)) == IV_ERROR_SUCCESS)
		{
			// Befehlsausfuehrung starten.
			// Listendaten zum AEIV uebertragen
			iRes = LoadEntrance2ListData4(deviceIDRemote, iListNum);

			// Versuchen die rote LED abzuschalten.
			m_channel4.SetTimeboySlotState(GetDeviceIDRemote(), 0, iRes);
		}

		if ((iRes > 0) && (((iRes >> 8) & 0xFF) == 3))
		{
			*piError = IV_ERROR_NOT_ENOUGH_FLASH; // Es steht nicht genuegend freier Flash Speicher zur Verfuegung.
			iRes = 0;
		}
		else
		{
			iRes = ResultToError(iRes, piError);
		}
			
		return iRes;}

	default:
		*piError = IV_ERROR_UNREACHABLE;
	}

	return 0;
}

// ListenBuffer in AE laden
int CAEChannel::LoadEntrance2ListData4(short deviceIDRemote, unsigned char listNum)
{
	// Alle Listen laden?
	if (listNum == 0xFF)
	{
		// Mindestens eine Liste geladen?
		int i;
		for (i=0; i < 7; i++)
		{
			if (m_ulEntrance2ListSize[i] != 0)
			{
				break;
			}
		}
		if (i == 7)
		{
			return -IV_ERROR_LIST_PROCESSING; // keine Listen vorhanden
		}
	}
	else
	{
		if (m_ulEntrance2ListSize[listNum] == 0)
		{
			return -IV_ERROR_LIST_PROCESSING; // keine Listen vorhanden
		}
	}

	unsigned short segmentUpdateVariation = IV_VALUE_SEGMENT_UPDATE_BG;
	int res = IV_ERROR_SUCCESS;
	unsigned int address, size, crc;
	address = size = crc = 0;

	// Systemstopp Setup uebertragen
	while ((res = m_channel4.BeginSetSegment(deviceIDRemote, segmentUpdateVariation, address, size, crc)) != IV_ERROR_SUCCESS)
	{
		// Wenn das Hintergrunduebertragen nicht unterstuetzt wird, dann die Standardmethode verwenden.
		if (((res & 0xFF) == 1) && (segmentUpdateVariation == IV_VALUE_SEGMENT_UPDATE_BG))
		{
			segmentUpdateVariation = IV_VALUE_SEGMENT_UPDATE;
			continue;
		}

		return res;
	}

	switch (listNum)
	{
	case 0xFF:
	case 0x00:
		// nur geladene Liste bei 0xFF uebertragen
		if (m_ulEntrance2ListSize[0] != 0)
		{
			// Liste uebertragen
			if ((res = LoadEntrance2List4(deviceIDRemote, 0)) != IV_ERROR_SUCCESS)
			{
				return res;
			}
		}
		if (listNum != 0xFF) break;
	case 0x01:
		// nur geladene Liste bei 0xFF uebertragen
		if (m_ulEntrance2ListSize[1] != 0)
		{
			// Liste uebertragen
			if ((res = LoadEntrance2List4(deviceIDRemote, 1)) != IV_ERROR_SUCCESS)
			{
				return res;
			}
		}
		if (listNum != 0xFF) break;
	case 0x02:
		// nur geladene Liste bei 0xFF uebertragen
		if (m_ulEntrance2ListSize[2] != 0)
		{
			// Liste uebertragen
			if ((res = LoadEntrance2List4(deviceIDRemote, 2)) != IV_ERROR_SUCCESS)
			{
				return res;
			}
		}
		if (listNum != 0xFF) break;
	case 0x03:
		// nur geladene Liste bei 0xFF uebertragen
		if (m_ulEntrance2ListSize[3] != 0)
		{
			// Liste uebertragen
			if ((res = LoadEntrance2List4(deviceIDRemote, 3)) != IV_ERROR_SUCCESS)
			{
				return res;
			}
		}
		if (listNum != 0xFF) break;
	case 0x04:
		// nur geladene Liste bei 0xFF uebertragen
		if (m_ulEntrance2ListSize[4] != 0)
		{
			// Liste uebertragen
			if ((res = LoadEntrance2List4(deviceIDRemote, 4)) != IV_ERROR_SUCCESS)
			{
				return res;
			}
		}
		if (listNum != 0xFF) break;
	case 0x05:
		// nur geladene Liste bei 0xFF uebertragen
		if (m_ulEntrance2ListSize[5] != 0)
		{
			// Liste uebertragen
			if ((res = LoadEntrance2List4(deviceIDRemote, 5)) != IV_ERROR_SUCCESS)
			{
				return res;
			}
		}
		if (listNum != 0xFF) break;
	case 0x06:
		// nur geladene Liste bei 0xFF uebertragen
		if (m_ulEntrance2ListSize[6] != 0)
		{
			// Liste uebertragen
			if ((res = LoadEntrance2List4(deviceIDRemote, 6)) != IV_ERROR_SUCCESS)
			{
				return res;
			}
		}
	}

	crc = 0;
	// Systemstopp Setup uebertragen
	if ((res = m_channel4.EndSetSegment(deviceIDRemote, segmentUpdateVariation, crc)) != IV_ERROR_SUCCESS)
	{
		return res;
	}

	return res;
}

// ListenBuffer in AE laden
int CAEChannel::LoadEntrance2List4(short deviceIDRemote, unsigned char listNum)
{
	int res = IV_ERROR_SUCCESS;
	STable2 table;
	char name[32];
	memset(name, 0, sizeof(name));
	sprintf(name, "ACCESS.%i", listNum);
	// Zur Liste Informationen ermitteln
	if ((res = m_channel4.GetTable(deviceIDRemote, name, table)) != IV_ERROR_SUCCESS)
	{
		return res;
	}
	strcpy(name, table.name);
	sprintf(table.name, "ACCESS.%s", name);

	// Muss die Liste auf Sortierung geprueft und ggf. sortiert werden.
	if (m_entrance2SortCheck[listNum] == false)
	{		
		int fieldOffs = 0;
		
		if (table.flags & TF_SORT && table.sort < table.count)
		{			
			//char name[64];
			int fieldNum;
			STable2Field field;
			memset(&field, 0, sizeof(STable2Field));
			// Alle Felder fuer Offset lesen
			for (fieldNum=0; fieldNum < table.count; fieldNum++)
			{				
				if (fieldNum < (table.sort + 1))
				{					
					// Datensatzoffset fuer Sortierfeld aufsummieren.
					fieldOffs += field.size;
				}
				else // Da die Datensatzgroesse bekannt ist, brauchen die restlichen Felder nicht ermittelt zu werden. 
				{ 
					break; 
				}
				
				/*
				sprintf(name, "ACCESS.%i.%i", listNum, fieldNum);
				
				if ((res = m_channel4.GetField(deviceIDRemote, name, field)) != IV_ERROR_SUCCESS)
				{					
					if (res < 0)
					{						
						// Kommunikationsfehler
						return res;						
					}
					else
					{						
						// Funktionsfehler
						return res;
					}
				}
				*/
				field = m_entrance2TableField[listNum][fieldNum];
				
				// Datensatzgroesse wird fuer Swap-Routine benoetigt!
			}
			
			// Liste sortieren.
			//QuicksortEntrance2Record(listNum, fieldOffs, 0, *(unsigned int*)m_pucEntrance2ListMem[listNum] - 1);
			QuicksortList(m_pucEntrance2ListMem[listNum], m_ucEntrance2RecordSize[listNum], fieldOffs, 0, *(unsigned int*)m_pucEntrance2ListMem[listNum] - 1);
			
			// Bei allen Listen ist 0 die sortierte Spalte
			// Flag dafuer dass die Liste sortiert wurde setzen.
			table.flags |= TF_SORTED;
		}
		
		// Sortiert oder zumindes geprueft.
		m_entrance2SortCheck[listNum] = true;
	}
	

	//
	unsigned short art = IV_VALUE_SEGMENT_LIST_ACCESS;
	unsigned int address, size, crc;
	address = size = crc = 0;
/*
	// Systemstopp Setup uebertragen
	if ((res = m_channel4.SegmentBegin(deviceIDRemote, SS_UPDATE, address, size, crc)) != IV_ERROR_SUCCESS)
	{
		return res;
	}
*/
	// CRC berechnen
	crc = 0xFFFFFFFF;
	crc = m_channel4.GetCRC32((unsigned char*)&table, sizeof(STable2), crc, FALSE);
	crc = m_channel4.GetCRC32(m_pucEntrance2ListMem[listNum], m_ulEntrance2ListSize[listNum], crc);

	// Groesze berechnen
	size = sizeof(STable2) + m_ulEntrance2ListSize[listNum];

	// Systemstopp List uebertragen
	while ((res = m_channel4.BeginSetSegment(deviceIDRemote, art, address, size, crc)) != IV_ERROR_SUCCESS)
	{
		// Typ unbekannt?, dann als "normale" Liste versuchen..
		if (((res & 0xFF) == 1) && (art == IV_VALUE_SEGMENT_LIST_ACCESS))
		{
			art = IV_VALUE_SEGMENT_LIST;
			continue;
		}

		return res;
	}

	// Daten uebertragen
	if ((res = m_channel4.SetFlash(deviceIDRemote, address, (unsigned char*)&table, sizeof(STable2))) != IV_ERROR_SUCCESS)
	{
		return res;
	}
	// Daten uebertragen
	address += sizeof(STable2);
	if ((res = m_channel4.SetFlash(deviceIDRemote, address, m_pucEntrance2ListMem[listNum], m_ulEntrance2ListSize[listNum])) != IV_ERROR_SUCCESS)
	{
		return res;
	}

	// Systemstopp Setup uebertragen
	if ((res = m_channel4.EndSetSegment(deviceIDRemote, art, crc)) != IV_ERROR_SUCCESS)
	{
		return res;
	}
/*
	crc = 0;
	// Systemstopp Setup uebertragen
	if ((res = m_channel4.SegmentEnd(deviceIDRemote, SS_UPDATE, crc)) != IV_ERROR_SUCCESS) {

		return res;
	}
*/
	return res;
}

// ListenBuffer in AE laden
int CAEChannel::LoadTimeboyListData(short deviceIDRemote, unsigned char id, unsigned char num, int reserve, int *piError)
{
	int iRes = IV_ERROR_SUCCESS;

	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(GetDeviceIDRemote()))
	{
	case 1:
	case 2:
		*piError = IV_ERROR_VERSION_FIRMWARE;
		return 0;

	case 4: {

		// Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
		if ((iRes = m_channel4.SetTimeboySlotState(GetDeviceIDRemote(), 1)) == IV_ERROR_SUCCESS)
		{
			// Befehlsausfuehrung starten.
			// Listendaten zum AEIV uebertragen
			iRes = LoadTimeboyListData4(deviceIDRemote, id, num, reserve);

			// Versuchen die rote LED abzuschalten.
			m_channel4.SetTimeboySlotState(GetDeviceIDRemote(), 0, iRes);
		}

		if ((iRes > 0) && (((iRes >> 8) & 0xFF) == 3))
		{
			*piError = IV_ERROR_NOT_ENOUGH_FLASH; // Es steht nicht genuegend freier Flash Speicher zur Verfuegung.
			iRes = 0;
		}
		else
		{
			iRes = ResultToError(iRes, piError);
		}
			
		return iRes;}

	default:
		*piError = IV_ERROR_UNREACHABLE;
	}

	return 0;
}

// ListenBuffer in AE laden
int CAEChannel::LoadTimeboyListData4(short deviceIDRemote, unsigned char id, unsigned char num, int reserve)
{
	// Alle Listen laden?
	if (id == 0xFF)
	{
		int groupID, listNum, found = 0;
		for (groupID=0; groupID < MAX_DATENSATZ_BESCHREIBUNG; groupID++)
		{			
			if (num == 0xFF)
			{
				// Mindestens eine Liste geladen?
				for (listNum=0; listNum < MAX_DATENSATZ_BESCHREIBUNG; listNum++)
				{
					if (m_ulTimeboyListSize[groupID][listNum] != 0)
					{
						found = 1;
						break;
					}
				}
				if (found)
				{
					break;
				}
			}
			else
			{
				// Mindestens eine Liste geladen?
				if (m_ulTimeboyListSize[groupID][num] == 0)
				{
					found = 1;
					break;
				}
			}
		}
		if (found == 0)
		{
			return -IV_ERROR_LIST_PROCESSING; // keine Listen vorhanden
		}
	}
	else
	{
		if (num == 0xFF)
		{
			int listNum, found = 0;
			// Mindestens eine Liste geladen?
			for (listNum=0; listNum < MAX_DATENSATZ_BESCHREIBUNG; listNum++)
			{
				if (m_ulTimeboyListSize[id][listNum] != 0)
				{
					found = 1;
					break;
				}
			}
			if (found == 0)
			{
				return -IV_ERROR_LIST_PROCESSING; // keine Listen vorhanden
			}
		}
		else
		{
			if (m_ulTimeboyListSize[id][num] == 0)
			{
				return -IV_ERROR_LIST_PROCESSING; // keine Listen vorhanden
			}
		}
	}

	unsigned short segmentUpdateVariation = IV_VALUE_SEGMENT_UPDATE_BG;
	int res = IV_ERROR_SUCCESS;
	unsigned int address, size, crc;
	address = size = crc = 0;

	// Systemstopp Listen uebertragen
	while ((res = m_channel4.BeginSetSegment(deviceIDRemote, segmentUpdateVariation, address, size, crc)) != IV_ERROR_SUCCESS)
	{
		// Wenn das Hintergrunduebertragen nicht unterstuetzt wird, dann die Standardmethode verwenden.
		if (((res & 0xFF) == 1) && (segmentUpdateVariation == IV_VALUE_SEGMENT_UPDATE_BG))
		{
			segmentUpdateVariation = IV_VALUE_SEGMENT_UPDATE;
			continue;
		}

		return res;
	}

	// Alle ids abarbeiten?
	if (id == 0xFF)
	{
		int tableNum, tableCount = 0;
		int groupID;
		for (groupID=0; groupID < MAX_DATENSATZ_BESCHREIBUNG; groupID++)
		{
			if (num == 0xFF)
			{
				for (tableNum=0; tableNum < MAX_DATENSATZ_BESCHREIBUNG; tableNum++)
				{
					// nur geladene Liste bei 0xFF uebertragen
					if (m_ulTimeboyListSize[groupID][tableNum] != 0)
					{
						// Maximale Segmentanzahl nicht ueberschreiten
						if (tableCount && !(tableCount % 16))
						{
							crc = 0;
							// Systemstopp Listen uebertragen
							if ((res = m_channel4.EndSetSegment(deviceIDRemote, segmentUpdateVariation, crc)) != IV_ERROR_SUCCESS)
							{
								return res;
							}

							// Systemstopp Listen uebertragen
							if ((res = m_channel4.BeginSetSegment(deviceIDRemote, segmentUpdateVariation, address, size, crc)) != IV_ERROR_SUCCESS)
							{
								return res;
							}
						}

						// Liste uebertragen
						if ((res = LoadTimeboyList4(deviceIDRemote, groupID, tableNum, reserve)) != IV_ERROR_SUCCESS)
						{
							return res;
						}

						tableCount++;
					}
				}
			}
			else
			{
				// nur geladene Liste bei 0xFF uebertragen
				if (m_ulTimeboyListSize[groupID][num] != 0)
				{
					// Maximale Segmentanzahl nicht ueberschreiten
					if (tableCount && !(tableCount % 16))
					{
						crc = 0;
						// Systemstopp Listen uebertragen
						if ((res = m_channel4.EndSetSegment(deviceIDRemote, segmentUpdateVariation, crc)) != IV_ERROR_SUCCESS)
						{
							return res;
						}

						// Systemstopp Listen uebertragen
						if ((res = m_channel4.BeginSetSegment(deviceIDRemote, segmentUpdateVariation, address, size, crc)) != IV_ERROR_SUCCESS)
						{
							return res;
						}
					}

					// Liste uebertragen
					if ((res = LoadTimeboyList4(deviceIDRemote, groupID, num, reserve)) != IV_ERROR_SUCCESS)
					{
						return res;
					}

					tableCount++;
				}
			}
		}
	}
	else
	{
		if (num == 0xFF)
		{
			int tableNum, tableCount = 0;
			for (tableNum=0; tableNum < MAX_DATENSATZ_BESCHREIBUNG; tableNum++)
			{
				// nur geladene Liste bei 0xFF uebertragen
				if (m_ulTimeboyListSize[id][tableNum] != 0)
				{
					// Maximale Segmentanzahl nicht ueberschreiten
					if (tableCount && !(tableCount % 16))
					{
						crc = 0;
						// Systemstopp Listen uebertragen
						if ((res = m_channel4.EndSetSegment(deviceIDRemote, segmentUpdateVariation, crc)) != IV_ERROR_SUCCESS)
						{
							return res;
						}

						// Systemstopp Listen uebertragen
						if ((res = m_channel4.BeginSetSegment(deviceIDRemote, segmentUpdateVariation, address, size, crc)) != IV_ERROR_SUCCESS)
						{
							return res;
						}
					}

					// Liste uebertragen
					if ((res = LoadTimeboyList4(deviceIDRemote, id, tableNum, reserve)) != IV_ERROR_SUCCESS)
					{
						return res;
					}

					tableCount++;
				}
			}
		}
		else
		{
			// Liste uebertragen
			if ((res = LoadTimeboyList4(deviceIDRemote, id, num, reserve)) != IV_ERROR_SUCCESS)
			{
				return res;
			}
		}
	}

	crc = 0;
	// Systemstopp Listen uebertragen
	if ((res = m_channel4.EndSetSegment(deviceIDRemote, segmentUpdateVariation, crc)) != IV_ERROR_SUCCESS)
	{
		return res;
	}

	return res;
}

// ListenBuffer in AE laden
int CAEChannel::LoadTimeboyList4(short deviceIDRemote, unsigned char id, unsigned char num, int reserve)
{
	int res = IV_ERROR_SUCCESS;
	STable2 table;
	memset(&table, 0, sizeof(STable2));
	table.num = num;
	sprintf(table.name, "TBLIST.%i.%i", id, num);
	table.flags = m_timeboyFlags[id][num];

	//
	unsigned short art = IV_VALUE_SEGMENT_LIST_TIMEBOY;
	unsigned int address, size, crc;
	address = size = crc = 0;
/*
	// Systemstopp Setup uebertragen
	if ((res = m_channel4.SegmentBegin(deviceIDRemote, SS_UPDATE, address, size, crc)) != IV_ERROR_SUCCESS)
	{
		return res;
	}
*/
	// CRC berechnen
	crc = 0xFFFFFFFF;
	crc = m_channel4.GetCRC32((unsigned char*)&table, sizeof(STable2), crc, FALSE);
	crc = m_channel4.GetCRC32(m_pucTimeboyListMem[id][num], m_ulTimeboyListSize[id][num], crc);

	// Groesze berechnen
	size = sizeof(STable2) + m_ulTimeboyListSize[id][num];

	// Systemstopp List uebertragen
	while ((res = m_channel4.BeginSetSegment(deviceIDRemote, art, address, size, crc)) != IV_ERROR_SUCCESS)
	{
		// Typ unbekannt?, dann als "normale" Liste versuchen..
		if (((res & 0xFF) == 1) && (art == IV_VALUE_SEGMENT_LIST_ACCESS))
		{
			art = IV_VALUE_SEGMENT_LIST;
			continue;
		}

		return res;
	}

	// Daten uebertragen
	if ((res = m_channel4.SetFlash(deviceIDRemote, address, (unsigned char*)&table, sizeof(STable2))) != IV_ERROR_SUCCESS)
	{
		return res;
	}
	// Daten uebertragen
	address += sizeof(STable2);
	if ((res = m_channel4.SetFlash(deviceIDRemote, address, m_pucTimeboyListMem[id][num], m_ulTimeboyListSize[id][num])) != IV_ERROR_SUCCESS)
	{
		return res;
	}

	// Systemstopp Setup uebertragen
	if ((res = m_channel4.EndSetSegment(deviceIDRemote, art, crc)) != IV_ERROR_SUCCESS)
	{
		return res;
	}
/*
	crc = 0;
	// Systemstopp Setup uebertragen
	if ((res = m_channel4.SegmentEnd(deviceIDRemote, SS_UPDATE, crc)) != IV_ERROR_SUCCESS)
	{
		return res;
	}
*/
	return res;
}

// Abfrage ob Datensaetze im AEIII+ vorliegen.
// Erst ab Datensatzweisem lesen.
int CAEChannel::CheckData(short deviceIDRemote, int *piError)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(GetDeviceIDRemote()))
	{
	case 1:  // AEIII+
	case 2: { // PZEMasterIV
		unsigned char ucIsData = FALSE;

		// Abfrage ob Daten vorhanden
		if (!m_channel1_2.ReadMem(MB_RECORD_AVAILABLE, 1, &ucIsData, piError))
		{			
			return -1;
		}

		return ucIsData;}

	case 4: {// PZEMasterIV mit Protokollversion 4
		int iRes = IV_ERROR_SUCCESS;
		unsigned char buf[0x255];
		int count = 0;
		unsigned char status = 0;

		// Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
		if ((iRes = m_channel4.SetTimeboySlotState(GetDeviceIDRemote(), 1)) == IV_ERROR_SUCCESS)
		{
			// Befehlsausfuehrung starten.
			iRes = m_channel4.GetRecord(GetDeviceIDRemote(), buf, count, status);

			// Versuchen die rote LED abzuschalten.
			m_channel4.SetTimeboySlotState(GetDeviceIDRemote(), 0, iRes);
		}
		
		iRes = ResultToError(iRes, piError);
		if (iRes)
		{
			// OK
			iRes = status;
		}
		else
		{
			iRes = -1;
		}
		return iRes;}

	default:
		*piError = IV_ERROR_UNREACHABLE;
	}

	return -1;
}

// Schliesst Relais mit Nummer fuer Zeit.
int CAEChannel::CloseRelay(short deviceIDRemote, int iNum, int iTime, int *piError)
{
	int iRes = IV_ERROR_SUCCESS;
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(GetDeviceIDRemote()))
	{
	case 1:
	case 2:
		return m_channel1_2.CloseRelay((unsigned char)GetDeviceIDRemote(), iNum, iTime, piError);

	case 4: { // PZEMasterIV mit Protokollversion 4
		if (iNum > 4)
		{
			*piError = IV_ERROR_FUNCTION_UNAVAILABLE;
			iRes = 0;
		}
		else
		{
			// Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
			if ((iRes = m_channel4.SetTimeboySlotState(GetDeviceIDRemote(), 1)) == IV_ERROR_SUCCESS)
			{
				// Befehlsausfuehrung starten.
				iRes = m_channel4.SetRelay(GetDeviceIDRemote(), "INTERN", 0, iNum, 1, iTime);

				// Versuchen die rote LED abzuschalten.
				m_channel4.SetTimeboySlotState(GetDeviceIDRemote(), 0, iRes);
			}

			iRes = ResultToError(iRes, piError);
		}
		return iRes;}

	default:
		*piError = IV_ERROR_UNREACHABLE;
	}

	return 0;
}

// Ermittelt aktuellen Zustand von Relais mit Nummer fuer Zeit.
int CAEChannel::GetRelayState(short deviceIDRemote, int iNum, int *piState, int *piTime, int *piError)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(GetDeviceIDRemote()))
	{
	case 1:
	case 2:
		return m_channel1_2.GetRelayState((unsigned char)GetDeviceIDRemote(), iNum, piState, piTime, piError);

	case 4: { // PZEMasterIV mit Protokollversion 4
		int iRes = IV_ERROR_SUCCESS;
		if (iNum > 4)
		{
			*piError = IV_ERROR_FUNCTION_UNAVAILABLE;
			iRes = 0;
		}
		else
		{		
			unsigned char state;
			unsigned short time;

			// Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
			if ((iRes = m_channel4.SetTimeboySlotState(GetDeviceIDRemote(), 1)) == IV_ERROR_SUCCESS)
			{
				// Befehlsausfuehrung starten.
				iRes = m_channel4.GetRelay(GetDeviceIDRemote(), "INTERN", 0, iNum, state, time);

				// Versuchen die rote LED abzuschalten.
				m_channel4.SetTimeboySlotState(GetDeviceIDRemote(), 0, iRes);
			}

			iRes = ResultToError(iRes, piError);
			if (iRes)
			{
				// OK
				*piState = state;
				*piTime = time;
			}
		}
		return iRes;}

	default:
		*piError = IV_ERROR_UNREACHABLE;
	}

	return 0;
}

// Oeffnet Relais mit Nummer.
int CAEChannel::OpenRelay(short deviceIDRemote, int iNum, int *piError)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	int iRes = IV_ERROR_SUCCESS;

	switch (GetProtocolVersion(GetDeviceIDRemote()))
	{
	case 1:
	case 2:
		return m_channel1_2.OpenRelay((unsigned char)GetDeviceIDRemote(), iNum, piError);

	case 4: { // PZEMasterIV mit Protokollversion 4
		if (iNum > 4)
		{
			*piError = IV_ERROR_FUNCTION_UNAVAILABLE;
			iRes = 0;
		}
		else
		{		
			unsigned short time = 0;

			// Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
			if ((iRes = m_channel4.SetTimeboySlotState(GetDeviceIDRemote(), 1)) == IV_ERROR_SUCCESS)
			{
				// Befehlsausfuehrung starten.
				iRes = m_channel4.SetRelay(GetDeviceIDRemote(), "INTERN", 0, iNum, 0, time);

				// Versuchen die rote LED abzuschalten.
				m_channel4.SetTimeboySlotState(GetDeviceIDRemote(), 0, iRes);
			}

			iRes = ResultToError(iRes, piError);
		}
		return iRes;}

	default:
		*piError = IV_ERROR_UNREACHABLE;
	}

	return 0;
}

// Daten einer Variablen setzen.
int CAEChannel::SetGlobVar(short deviceIDRemote, unsigned char *pucVar, int iType, char *val, int *piError)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	int iRes = IV_ERROR_SUCCESS;

	switch (GetProtocolVersion(GetDeviceIDRemote()))
	{
	case 1:
	case 2:
		return m_channel1_2.SetGlobVar((unsigned char)GetDeviceIDRemote(), pucVar, iType, val, piError);

	case 4: { // PZEMasterIV mit Protokollversion 4

		// Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
		if ((iRes = m_channel4.SetTimeboySlotState(GetDeviceIDRemote(), 1)) == IV_ERROR_SUCCESS)
		{
			// Befehlsausfuehrung starten.
			switch (iType)
			{
			case 0: {
				char name[64];
				memset(name, 0, sizeof(name));
				sprintf(name, "SETUP.%i", *(unsigned char*)pucVar);
				iRes = m_channel4.SetVariable(GetDeviceIDRemote(), name, val);
				break;}

			case 1: {
				char name[64];
				memset(name, 0, sizeof(name));
				strcpy(name, "SETUP.");
				strcat(name, (char*)pucVar);
				iRes = m_channel4.SetVariable(GetDeviceIDRemote(), name, val);
				break;}

			case 2:
				iRes = m_channel4.SetSystemVariable(GetDeviceIDRemote(), (char*)pucVar, val);
				break;
			}

			// Versuchen die rote LED abzuschalten.
			m_channel4.SetTimeboySlotState(GetDeviceIDRemote(), 0, iRes);
		}

		if (iRes > 0)
		{
			switch(iRes)
			{
			case 1:
			case 2:
				*piError = IV_ERROR_VARIABLE_UNDEFINED;
				iRes = 0;
				break;
			case 3:
				*piError = IV_ERROR_VARIABLE_VALUE;
				iRes = 0;
				break;
			default:
				iRes = ResultToError(iRes, piError);
			}
		}
		else
		{
			iRes = ResultToError(iRes, piError);
		}
		return iRes;}

	default:
		*piError = IV_ERROR_UNREACHABLE;
	}

	return 0;
}

// Daten einer Globalen Variablen lesen.
int CAEChannel::GetGlobVar(short deviceIDRemote, unsigned char *pucVar, int iType, char *val, int iValueLength, int *piError)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);
  char lValue[IV_PACKET_LENGTH_DATA];
  memset(lValue, 0, sizeof(lValue));

	// muss mindestens ein Zeichen anfordern.
	if ((iValueLength - 1) <= 0)
	{
		*piError = IV_ERROR_FUNCTION_PARAMETER;
		return 0;
	} 

	int iRes = IV_ERROR_SUCCESS;
	switch (GetProtocolVersion(GetDeviceIDRemote()))
	{
	case 1:
	case 2:
		return m_channel1_2.GetGlobVar((unsigned char)GetDeviceIDRemote(), pucVar, iType, val, iValueLength, piError);

	case 4: { // PZEMasterIV mit Protokollversion 4

		// Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
		if ((iRes = m_channel4.SetTimeboySlotState(GetDeviceIDRemote(), 1)) == IV_ERROR_SUCCESS)
		{
			// Befehlsausfuehrung starten.
			switch (iType) 
			{
			case 0: {
				char name[64];
				memset(name, 0, sizeof(name));
				sprintf(name, "SETUP.%i", *(unsigned char*)pucVar);
				iRes = m_channel4.GetVariable(deviceIDRemote, name, lValue);
				break;}

			case 1: {
				char name[64];
				memset(name, 0, sizeof(name));
				strcpy(name, "SETUP.");
				strcat(name, (char*)pucVar);
				iRes = m_channel4.GetVariable(deviceIDRemote, name, lValue);
				break;}

			case 2:
				iRes = m_channel4.GetSystemVariable(deviceIDRemote, (char*)pucVar, lValue);
				break;
			}

			// Versuchen die rote LED abzuschalten.
			m_channel4.SetTimeboySlotState(GetDeviceIDRemote(), 0, iRes);
		}

		if (iRes > 0)
		{
			switch ( iRes )
			{
			case 1:
			case 2:
				*piError = IV_ERROR_VARIABLE_UNDEFINED;
				iRes = 0;
				break;
			case 3:
				*piError = IV_ERROR_VARIABLE_VALUE;
				iRes = 0;
				break;
			default:
				iRes = ResultToError(iRes, piError);
			}
		}
		else
		{
			iRes = ResultToError(iRes, piError);
		}
    if (iRes) {
      strncpy(val, lValue, iValueLength - 1);
    }

		return iRes;}

	default:
		*piError = IV_ERROR_UNREACHABLE;
	}

	return 0;
}

int CAEChannel::GetVersionFirmware(short deviceIDRemote, char *buf, int *piError)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	unsigned char lBuf[32];

	unsigned char pv = GetProtocolVersion(GetDeviceIDRemote());
	switch (pv)
	{
	case 1: 
	case 2: {

		if (pv == 1) // Nur Protokollversion 1
		{
			switch (GetProtocolVersion2(GetDeviceIDRemote(), 1, piError))
			{
			case 0:
				return 0;

			case 0x0400: // AEIII+
			case 0x0401:
				*piError = IV_ERROR_VERSION_FIRMWARE;
				return 0;
			}
		}

		// Firmwareversion lesen
		if (!m_channel1_2.ReadMem(MB_VERSION, 32, lBuf, piError))
		{
			return 0;
		}
		memcpy(buf, lBuf, 32);

		return 1;}

	case 4: {
		int iRes = IV_ERROR_SUCCESS;

		// Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
		if ((iRes = m_channel4.SetTimeboySlotState(GetDeviceIDRemote(), 1)) == IV_ERROR_SUCCESS)
		{
			// Befehlsausfuehrung starten.
			iRes = m_channel4.GetVersion(GetDeviceIDRemote(), (char*)lBuf);

			// Versuchen die rote LED abzuschalten.
			m_channel4.SetTimeboySlotState(GetDeviceIDRemote(), 0, iRes);
		}

		iRes = ResultToError(iRes, piError);

		if (iRes)
		{
			// OK
			strcpy(buf, (char*)lBuf);
		}
		return iRes;}

	default:
		*piError = IV_ERROR_UNREACHABLE;
	}

	return 0;
}

int CAEChannel::GetVersionFirmwareFromFile(char *fileName, char *versionFirmware, int *error)
{
	int res = 0;
	// Lokale Membervariable reseten
	m_fileDevice.Reset();

	if (!(res = m_fileDevice.GetVersionFirmware(fileName, versionFirmware)))
	{
		switch (m_fileDevice.GetLastErrorState())
		{
		case ES_FILE_OPEN:
			*error = IV_ERROR_FILE_OPEN;
			break;

		case ES_HEAD_CRC:
			*error = IV_ERROR_CRC_PACKET;
			break;

		case ES_VERSION:
		case ES_DEVICE_UNKNOWN:
		case ES_FW_VERSION_LENGTH:
		case ES_FUNCTION_UNAVAILABLE:
			*error = IV_ERROR_FUNCTION_UNAVAILABLE;
			break;
		}

		return 0;
	}

	return res;
}

// Prueft ob Displaydaten in Ordnung.
int CAEChannel::CheckFileDisplay(CDFFileDisplay *pFileDisplay, int *piError)
{
	unsigned int ulFontPtr, ulGraficPtr;
	int i, iFontCount, iGraficCount;
	CDFFileDisplay fileDisplay;

	// Fehlerobjekt loeschen
	*piError = IV_ERROR_SUCCESS;

	// EEPROM - Zeiger auf Zeichensatz lesen 
 	if (!m_channel1_2.ReadMem(MB_START_EEP + 0x0F78, sizeof(unsigned int), (unsigned char*)&ulFontPtr, piError))
	{
		return 0;
	}
	ChangeOrderLong((unsigned char*)&ulFontPtr);
	if (ulFontPtr == 0xFFFFFFFF)
	{ 
		// Zeiger nicht vorhanden
		*piError = IV_ERROR_SUCCESS;
		return 0;
	}

	// EEPROM - Zeiger auf Grafik lesen 
 	if (!m_channel1_2.ReadMem(MB_START_EEP + 0x0F7C, sizeof(unsigned int), (unsigned char*)&ulGraficPtr, piError))
	{
		return 0;
	}
	ChangeOrderLong((unsigned char*)&ulGraficPtr);
	if (ulGraficPtr == 0xFFFFFFFF)
	{ 
		// Zeiger nicht vorhanden
		*piError = IV_ERROR_SUCCESS;
		return 0;
	}

	// Hauptstruktur aus Flash lesen um Anzahl zu ermitteln 
	if (!m_channel1_2.ReadMem(0x000B0200 + ulFontPtr, sizeof(FONTNODE), fileDisplay.m_memFont, piError))
	{
		return 0;
	}
	// Anzahl Zeichensaetze
	iFontCount = fileDisplay.m_pFontNode->cnt;
	if ((iFontCount > MAX_FONT) || (iFontCount != pFileDisplay->m_pFontNode->cnt))
	{
		*piError = IV_ERROR_SUCCESS;
		return 0;
	}

	// Hauptstruktur aus Flash lesen um Anzahl zu ermitteln 
	if (!m_channel1_2.ReadMem(0x000B0200 + ulGraficPtr, sizeof(GRAFICNODE), fileDisplay.m_memGrafic, piError))
	{
		return 0;
	}
	// Anzahl Grafiken
	iGraficCount = fileDisplay.m_pGraficNode->cnt;
	if ((iGraficCount > MAX_GRAFIC) || (iGraficCount != pFileDisplay->m_pGraficNode->cnt))
	{
		*piError = IV_ERROR_SUCCESS;
		return 0;
	}
	// Klassenstrukturen initialisieren
	fileDisplay.Default(iFontCount, iGraficCount);

	// Strukturen aus Flash lesen 
	if (!m_channel1_2.ReadMem(0x000B0200 + ulFontPtr, sizeof(FONTNODE) + iFontCount * sizeof(unsigned int) + iFontCount * sizeof(FONT), fileDisplay.m_memFont, piError))
	{
		return 0;
	}
	// CRC von Zeichensaetzen ueberpruefen
	for (i=0; i < iFontCount; i++)
	{
		if (fileDisplay.m_pFont[i]->crc != pFileDisplay->m_pFont[i]->crc)
		{
			*piError = IV_ERROR_SUCCESS;
			return 0;
		}
	}

	// Strukturen aus Flash lesen 
	if (!m_channel1_2.ReadMem(0x000B0200 + ulGraficPtr, sizeof(GRAFICNODE) + iGraficCount * sizeof(unsigned int) + iGraficCount * sizeof(GRAFIC), fileDisplay.m_memGrafic, piError))
	{
		return 0;
	}
	// CRC von Grafiken ueberpruefen
	for (i=0; i < iGraficCount; i++)
	{
		if (fileDisplay.m_pGrafic[i]->crc != pFileDisplay->m_pGrafic[i]->crc)
		{
			*piError = IV_ERROR_SUCCESS;
			return 0;
		}
	}

	// Zeichensatz- und Grafikdaten vorhanden und identisch
	return 1;
}

int CAEChannel::UploadTextData(short deviceIDRemote, char *szFileName)
{
	// Zeichensatz- und Grafikdaten
	CDFFileDeviceText *pFileDeviceText = m_fileDevice.GetFileDeviceText();
	int res = IV_ERROR_SUCCESS;
	unsigned int address = 0;
	unsigned char *buf;
	unsigned int size = 0;
	unsigned int crc = 0;

	if (!pFileDeviceText->Import(szFileName))
	{
		switch (pFileDeviceText->GetLastErrorState())
		{
		case ES_FILE_OPEN:
			return -IV_ERROR_FILE_OPEN;

		case ES_FILE_FORMAT:
		default:
			return -IV_ERROR_FILE_FORMAT;
		}
	}

//	memset(version, 0, sizeof(version));
	// Bootversion lesen 
//	if ((res = m_channel4.GetVersionBootloader(deviceIDRemote, version)) != IV_ERROR_SUCCESS)
//	{

//		return res;
//	}

	// Systemstopp setzen
	if ((res = m_channel4.BeginSetSegment(deviceIDRemote, IV_VALUE_SEGMENT_UPDATE, address, size, crc)) != IV_ERROR_SUCCESS)
	{
		return res;
	}

	// Sind Textdaten vorhanden?
	if (pFileDeviceText->GetSizeData())
	{
		// Texte als ersten Block uebertragen
		// TEXTE

		// Type User setzen.
		TEXTDATA *td = (TEXTDATA*)pFileDeviceText->m_memData;
		td->type = (m_loadAsDefault) ? 2 : 3; // Textdata Default/User
		m_loadAsDefault = 0; // ggf. zuruecksetzen

		address = 0;
		buf = pFileDeviceText->m_memData;
		size = (unsigned int)pFileDeviceText->m_sizeData;
		crc = m_channel4.GetCRC32(buf, size);

		// Systemstopp setzen
		res = m_channel4.BeginSetSegment(deviceIDRemote, IV_VALUE_SEGMENT_TEXT, address, size, crc);
		if (res != IV_ERROR_SUCCESS && (res & 0xFF) != 1)
		{
			return res;
		}

		// Funktionsfehler nicht Blocktyp unbekannt?
		if ((res & 0xFF) != 1)
		{
			// Textdaten uebertragen
			if ((res = m_channel4.SetFlash(deviceIDRemote, address, buf, size)) != IV_ERROR_SUCCESS)
			{
				return res;
			}

			// Systemstopp aufheben
			if ((res = m_channel4.EndSetSegment(deviceIDRemote, IV_VALUE_SEGMENT_TEXT, crc)) != IV_ERROR_SUCCESS)
			{
				return res;
			}

		}
		else
		{
			// Funktion wird von Firmware nicht unterstuetzt.
			return -IV_ERROR_FUNCTION_UNAVAILABLE;
		}
	}
	else
	{
		// Es liegen keine Daten fuer die Uebertragung vor.
		return -IV_ERROR_NO_DATA_AVAILABLE;

	}

	crc = 0;
	// Systemstopp aufheben
	if ((res = m_channel4.EndSetSegment(deviceIDRemote, IV_VALUE_SEGMENT_UPDATE, crc)) != IV_ERROR_SUCCESS)
	{
		return res;
	}

	// Textdaten erfolgreich uebertragen
	return res;
}

int CAEChannel::UploadTouchKeyData(short deviceIDRemote, char *szFileName)
{
	// Touchkeydaten
	CDFFileTouchKeys fileTouchKeys;;
	int res = IV_ERROR_SUCCESS;
	unsigned int address = 0;
	unsigned char *buf;
	unsigned int size = 0;
	unsigned int crc = 0;
	unsigned char val;

	// Pruefen ob ein Touch in den Hardwaresettings vorliegt
	if ((res = m_channel4.GetTypeTouch(deviceIDRemote, val)) != IV_ERROR_SUCCESS)
	{
		return res;
	}
	if (val == IV_VALUE_TYPE_TOUCH_NONE)
	{
		return -IV_ERROR_NO_MODULE;
	}

	// Touchdatendatei laden
	if (!fileTouchKeys.Load(szFileName))
	{
		switch (fileTouchKeys.GetLastErrorState())
		{
		case ES_FILE_OPEN:
			return -IV_ERROR_FILE_OPEN;

		case ES_FILE_FORMAT:
		default:
			return -IV_ERROR_FILE_FORMAT;
		}
	}

	//	memset(version, 0, sizeof(version));
	// Bootversion lesen 
	//	if ((res = m_channel4.GetVersionBootloader(deviceIDRemote, version)) != IV_ERROR_SUCCESS)
	//{
	//		return res;
	//	}

	// Systemstopp setzen
	if ((res = m_channel4.BeginSetSegment(deviceIDRemote, IV_VALUE_SEGMENT_UPDATE, address, size, crc)) != IV_ERROR_SUCCESS)
	{
		return res;
	}

	// Sind TouchKeyDaten vorhanden?
	if (fileTouchKeys.GetSizeData())
	{
		address = 0;
		buf = fileTouchKeys.m_memTouchKeys;
		size = fileTouchKeys.m_sizeData;
		crc = m_channel4.GetCRC32(buf, size);

		// Systemstopp setzen
		res = m_channel4.BeginSetSegment(deviceIDRemote, IV_VALUE_SEGMENT_TOUCHKEYS, address, size, crc);
		if (res != IV_ERROR_SUCCESS && (res & 0xFF) != 1)
		{
			return res;
		}

		// Funktionsfehler nicht Blocktyp unbekannt?
		if ((res & 0xFF) != 1)
		{
			// Textdaten uebertragen
			if ((res = m_channel4.SetFlash(deviceIDRemote, address, buf, size)) != IV_ERROR_SUCCESS)
			{
				return res;
			}

			// Systemstopp aufheben
			if ((res = m_channel4.EndSetSegment(deviceIDRemote, IV_VALUE_SEGMENT_TOUCHKEYS, crc)) != IV_ERROR_SUCCESS)
			{
				return res;
			}
		}
		else
		{
			// Funktion wird von Firmware nicht unterstuetzt.
			return -IV_ERROR_FUNCTION_UNAVAILABLE;
		}
	}
	else
	{
		// Es liegen keine Daten fuer die Uebertragung vor.
		return -IV_ERROR_NO_DATA_AVAILABLE;

	}

	crc = 0;
	// Systemstopp aufheben
	if ((res = m_channel4.EndSetSegment(deviceIDRemote, IV_VALUE_SEGMENT_UPDATE, crc)) != IV_ERROR_SUCCESS)
	{
		return res;
	}

	// TouchKeyDaten erfolgreich uebertragen
	return res;
}

int CAEChannel::UploadDisplayColors(short deviceIDRemote, char *szFileName)
{
	// Farbdaten
	CDFFileDisplayColors fileDisplayColors;
	int res = IV_ERROR_SUCCESS;
	unsigned int address = 0;
	unsigned char *buf;
	unsigned int size = 0;
	unsigned int crc = 0;
	unsigned char val;

	// Pruefen ob ein Timeboy in den Hardwaresettings vorliegt
	if ((res = m_channel4.GetTypeDevice(deviceIDRemote, val)) != IV_ERROR_SUCCESS)
	{
		return res;
	}
	if (val != IV_VALUE_TYPE_DEVICE_TIMEBOY)
	{
		return -IV_ERROR_DEVICE_WRONG;
	}

	// Pruefen ob ein Farbdisplay in den Hardwaresettings vorliegt
	if ((res = m_channel4.GetTypeDisplay(deviceIDRemote, val)) != IV_ERROR_SUCCESS)
	{
		return res;
	}
	if (val != IV_VALUE_TYPE_DISPLAY_SMALLER)
	{
		return -IV_ERROR_NO_MODULE;
	}

	// Touchdatendatei laden
	if (!fileDisplayColors.Load(szFileName))
	{
		switch (fileDisplayColors.GetLastErrorState())
		{
		case ES_FILE_OPEN:
			return -IV_ERROR_FILE_OPEN;

		case ES_FILE_FORMAT:
		default:
			return -IV_ERROR_FILE_FORMAT;
		}
	}

	//	memset(version, 0, sizeof(version));
	// Bootversion lesen 
	//	if ((res = m_channel4.GetVersionBootloader(deviceIDRemote, version)) != IV_ERROR_SUCCESS)
	//{
	//		return res;
	//	}

	// Systemstopp setzen
	if ((res = m_channel4.BeginSetSegment(deviceIDRemote, IV_VALUE_SEGMENT_UPDATE, address, size, crc)) != IV_ERROR_SUCCESS)
	{
		return res;
	}

	// Sind TouchKeyDaten vorhanden?
	if (fileDisplayColors.GetSizeData())
	{
		address = 0;
		buf = fileDisplayColors.m_memDisplayColors;
		size = fileDisplayColors.m_sizeData;
		crc = m_channel4.GetCRC32(buf, size);

		// Systemstopp setzen
		res = m_channel4.BeginSetSegment(deviceIDRemote, IV_VALUE_SEGMENT_DISPLAYCOLORS, address, size, crc);
		if (res != IV_ERROR_SUCCESS && (res & 0xFF) != 1)
		{
			return res;
		}

		// Funktionsfehler nicht Blocktyp unbekannt?
		if ((res & 0xFF) != 1)
		{
			// Textdaten uebertragen
			if ((res = m_channel4.SetFlash(deviceIDRemote, address, buf, size)) != IV_ERROR_SUCCESS)
			{
				return res;
			}

			// Systemstopp aufheben
			if ((res = m_channel4.EndSetSegment(deviceIDRemote, IV_VALUE_SEGMENT_DISPLAYCOLORS, crc)) != IV_ERROR_SUCCESS)
			{
				return res;
			}
		}
		else
		{
			// Funktion wird von Firmware nicht unterstuetzt.
			return -IV_ERROR_FUNCTION_UNAVAILABLE;
		}
	}
	else
	{
		// Es liegen keine Daten fuer die Uebertragung vor.
		return -IV_ERROR_NO_DATA_AVAILABLE;
	}

	crc = 0;
	// Systemstopp aufheben
	if ((res = m_channel4.EndSetSegment(deviceIDRemote, IV_VALUE_SEGMENT_UPDATE, crc)) != IV_ERROR_SUCCESS)
	{
		return res;
	}

	// DisplayColorDaten erfolgreich uebertragen
	return res;
}

int CAEChannel::UploadFileDevice(short deviceIDRemote, char *szFileName, int *piError)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	int iRes = IV_ERROR_SUCCESS, 
		protocolVersion = GetProtocolVersion(GetDeviceIDRemote());

	// Puffer fuer die Hardwareinfo in Textform
	memset(m_info, 0, sizeof(m_info));

	//#########################################################################
	// Pruefen ob gewaehlte Datei in das Geraet eingespielt werden kann.
	switch(protocolVersion)
	{
	case 1: // AEIII+ nicht unterstuetzt
		*piError = IV_ERROR_VERSION_FIRMWARE;
		return 0;

	case 2: { // Nur fuer PZE-MasterIV moeglich.

		// Dateityp ermitteln
		struct stat st;
		if (stat(szFileName, &st) == -1) 
		{	
			*piError = IV_ERROR_FILE_OPEN;
			return 0;
		}
		
		// Angabe ist eine Datei
		if (S_ISREG(st.st_mode)) 
		{	
			// Nur ".hex" gueltig.
			size_t pos = strlen(szFileName) - 4;
#ifdef _WIN32
			if (szFileName[pos] != '.' || 
				((!(szFileName[pos + 1] == 'h' || szFileName[pos + 1] == 'H') ||
				  !(szFileName[pos + 2] == 'e' || szFileName[pos + 2] == 'E') ||
				  !(szFileName[pos + 3] == 'x' || szFileName[pos + 3] == 'X')) &&
				  (!(szFileName[pos + 1] == 'd' || szFileName[pos + 1] == 'D') ||
				  !(szFileName[pos + 2] == 'f' || szFileName[pos + 2] == 'F') ||
				  !(szFileName[pos + 3] == 'z' || szFileName[pos + 3] == 'Z')) &&
				  (!(szFileName[pos + 1] == 'z' || szFileName[pos + 1] == 'Z') ||
				  !(szFileName[pos + 2] == 'i' || szFileName[pos + 2] == 'I') ||
				  !(szFileName[pos + 3] == 'p' || szFileName[pos + 3] == 'P'))) ) 
			{	
				*piError = IV_ERROR_FILE_EXTENSION;
				return 0;
			}
#else 
			if (!S_ISREG(st.st_mode) || szFileName[pos] != '.' || 
				!(szFileName[pos + 1] == 'h' || szFileName[pos + 1] == 'H') ||
				!(szFileName[pos + 2] == 'e' || szFileName[pos + 2] == 'E') ||
				!(szFileName[pos + 3] == 'x' || szFileName[pos + 3] == 'X') ) 
			{
				*piError = IV_ERROR_FILE_EXTENSION;
				return 0;
			}
#endif	
			// Handelt es sich um eine Hexdatei
			if (szFileName[pos + 1] == 'h' || szFileName[pos + 1] == 'H') 
			{
				// Dateidaten laden
				m_fileDevice.Reset();
				if (!m_fileDevice.Load(szFileName)) 
				{
					if (m_fileDevice.GetLastErrorState() == ES_VERSION || m_fileDevice.GetLastErrorState() == ES_DEVICE_UNKNOWN) 
					{
						*piError = IV_ERROR_VERSION_DEVICEFILE;
					} 
					else if (m_fileDevice.GetLastErrorState() == ES_FILE_OPEN) 
					{
						*piError = IV_ERROR_FILE_OPEN;
					} 
					else 
					{
						*piError = IV_ERROR_DEVICEFILE_LOAD;
					}

					m_fileDevice.Reset();
					return 0;
				}

				// Ist die Softwareinformation in der Geraetedatei verfuegbar?
				if (m_fileDevice.GetHeaderSoftwareInfo().size) 
				{	
					// Auslesen der Hardwareinfo
					TDFHardwareInfo hwInfo;
					if (!m_channel1_2.GetHardwareInfo(deviceIDRemote, hwInfo, piError)) 
					{
						return 0;
					} // Hardwareinfo
					
					// Pruefen ob die Datei in die vorliegende Hardware eingespielt werden kann.
					if (0 == m_fileDevice.CompareInfos(hwInfo, m_fileDevice.GetHeaderSoftwareInfo())) 
					{	
						// Hardwareinfo in Text aufloesen
						m_channel4.GetInfos(hwInfo, m_info);
						strcpy(m_fileDevice.m_errorExtended, m_fileDevice.GetErrorText());
						*piError = IV_ERROR_DEVICEFILE_MATCH;
						return 0;
					}
				} 
				else 
				{
					char version[64];
					// Die Geraetedatei muss vom Typ "PZE-MasterIV" sein, oder diesen unterstuetzen.
					if (m_fileDevice.GetHeaderVersionFile() != IV_DFFILEDEVICE_VERSION_PZE) 
					{
						// Die gewaehlte Geraetedatei ist NICHT fuer das PZE-MasterIV.
						*piError = IV_ERROR_DEVICEFILE_WRONG;
						return 0;
					}

					// Da Protokollversion 2 muss es sich um eine Boardversion < 3 handeln.
					// Es muss sich um eine 04.00.xx oder 04.01.xx Firmwareversion handeln.
					if (m_fileDevice.GetHeaderVersionFirmware(version)) 
					{
						if (strstr(version, "04.00.") != version && strstr(version, "04.01.") != version)
						{
							*piError = IV_ERROR_DEVICEFILE_MATCH;
							return 0;
						}
					}
				}
			} 
			else 
			{ // Es handelt sich um eine Zipdatei

				// Auslesen der Hardwareinfo
				TDFHardwareInfo hwInfo;
				if (!m_channel1_2.GetHardwareInfo(deviceIDRemote, hwInfo, piError)) 
				{	
					return 0;
				}
				
				// Pruefen ob die Datei in die vorliegende Hardware eingespielt werden kann.
				if (0 == m_fileDevice.SearchInfosZip(szFileName, hwInfo)) 
				{	
					// Hardwareinfo in Text aufloesen
					m_channel4.GetInfos(hwInfo, m_info);
					if (m_fileDevice.GetLastErrorState() == ES_VERSION || m_fileDevice.GetLastErrorState() == ES_DEVICE_UNKNOWN) 
					{
						*piError = IV_ERROR_VERSION_DEVICEFILE;
					} 
					else if (m_fileDevice.GetLastErrorState() == ES_NO_FILE_MATCH) 
					{
						*piError = IV_ERROR_DEVICEFILE_MATCH;
					} 
					else 
					{
						*piError = IV_ERROR_DEVICEFILE_LOAD;
					}

					return 0;
				}
			}	
		} 
		else if (S_ISDIR(st.st_mode)) 
		{
#ifdef _WIN32			
			// Auswahl aus Verzeichnis nur unter Linux gueltig!
			*piError = IV_ERROR_FUNCTION_PARAMETER;
			return 0;
#else
			// Auslesen der Hardwareinfo
			TDFHardwareInfo hwInfo;
			if (!m_channel1_2.GetHardwareInfo(deviceIDRemote, hwInfo, piError)) 
			{
				return FALSE;
			}
			
			// Pruefen ob eine passende Datei in die vorliegende Hardware eingespielt werden kann.
			if (0 == m_fileDevice.SearchInfosDir(szFileName, hwInfo)) 
			{	
				// Hardwareinfo in Text aufloesen
				m_channel4.GetInfos(hwInfo, m_info);
				if (m_fileDevice.GetLastErrorState() == ES_VERSION || m_fileDevice.GetLastErrorState() == ES_DEVICE_UNKNOWN) 
				{
					*piError = IV_ERROR_VERSION_DEVICEFILE;
				} 
				else if (m_fileDevice.GetLastErrorState() == ES_NO_FILE_MATCH) 
				{
					*piError = IV_ERROR_DEVICEFILE_MATCH;
				} 
				else 
				{
					*piError = IV_ERROR_DEVICEFILE_LOAD;
				}

				return 0;
			}
#endif
		} 
		else 
		{	
			*piError = IV_ERROR_FILE_OPEN;
			return 0;
		}	
		break;}

	case 4: {

		// Dateityp ermitteln
		struct stat st;
		if (stat(szFileName, &st) == -1) 
		{	
			*piError = IV_ERROR_FILE_OPEN;
			return 0;
		}
		
		// Angabe ist eine Datei
		if (S_ISREG(st.st_mode)) 
		{	
			// Nur ".hex" gueltig.
			size_t pos = strlen(szFileName) - 4;
#ifdef _WIN32
			if (szFileName[pos] != '.' || 
				((!(szFileName[pos + 1] == 'h' || szFileName[pos + 1] == 'H') ||
				  !(szFileName[pos + 2] == 'e' || szFileName[pos + 2] == 'E') ||
				  !(szFileName[pos + 3] == 'x' || szFileName[pos + 3] == 'X')) &&
				  (!(szFileName[pos + 1] == 'd' || szFileName[pos + 1] == 'D') ||
				  !(szFileName[pos + 2] == 'f' || szFileName[pos + 2] == 'F') ||
				  !(szFileName[pos + 3] == 'z' || szFileName[pos + 3] == 'Z')) &&
				  (!(szFileName[pos + 1] == 'z' || szFileName[pos + 1] == 'Z') ||
				  !(szFileName[pos + 2] == 'i' || szFileName[pos + 2] == 'I') ||
				  !(szFileName[pos + 3] == 'p' || szFileName[pos + 3] == 'P'))) ) 
			{	
				*piError = IV_ERROR_FILE_EXTENSION;
				return 0;
			}
#else 
			if (!S_ISREG(st.st_mode) || szFileName[pos] != '.' || 
				!(szFileName[pos + 1] == 'h' || szFileName[pos + 1] == 'H') ||
				!(szFileName[pos + 2] == 'e' || szFileName[pos + 2] == 'E') ||
				!(szFileName[pos + 3] == 'x' || szFileName[pos + 3] == 'X') ) 
			{
				*piError = IV_ERROR_FILE_EXTENSION;
				return 0;
			}
#endif	
			// Handelt es sich um eine Hexdatei
			if (szFileName[pos + 1] == 'h' || szFileName[pos + 1] == 'H') 
			{	
				// Dateidaten laden
				m_fileDevice.Reset();
				if (!m_fileDevice.Load(szFileName) || m_fileDevice.IsCrypt()) 
				{	
					if (m_fileDevice.GetLastErrorState() == ES_VERSION || m_fileDevice.GetLastErrorState() == ES_DEVICE_UNKNOWN) 
					{
						*piError = IV_ERROR_VERSION_DEVICEFILE;
					} 
					else if (m_fileDevice.GetLastErrorState() == ES_FILE_OPEN) 
					{
						*piError = IV_ERROR_FILE_OPEN;
					} 
					else 
					{
						*piError = IV_ERROR_DEVICEFILE_LOAD;
					}

					m_fileDevice.Reset();
					return 0;
				}
				
				// Ist die Softwareinformation in der Geraetedatei verfuegbar?
				if (m_fileDevice.GetHeaderSoftwareInfo().size) 
				{	
					// Auslesen der Hardwareinfo
					TDFHardwareInfo hwInfo;
					iRes = m_channel4.GetHardwareInfo(deviceIDRemote, hwInfo);
					iRes = ResultToError(iRes, piError);
					if (iRes == 0)
					{
						return iRes;

					} // Hardwareinfo
					
					// Liegt keine Hardwareinfo vor?
					// Verwendung der ZK-Option ueber Vorhandensein der Readerliste ermitteln.
					if (hwInfo.version == 0) 
					{	
						int handle;
						if (m_channel4.TableOpen(deviceIDRemote, "ACCESS.READER", 0, handle, "") == IV_ERROR_SUCCESS && 
							m_channel4.TableClose(deviceIDRemote, handle) == IV_ERROR_SUCCESS) 
						{	
							// ZK-Option in Verwendung.
							hwInfo.administrationFlags |= HIAF_HAS_ZK;
						}
					}

					// Pruefen ob die Datei in die vorliegende Hardware eingespielt werden kann.
					if (0 == m_fileDevice.CompareInfos(hwInfo, m_fileDevice.GetHeaderSoftwareInfo())) 
					{	
						// Hardwareinfo in Text aufloesen
						m_channel4.GetInfos(hwInfo, m_info);
						strcpy(m_fileDevice.m_errorExtended, m_fileDevice.GetErrorText());
						*piError = IV_ERROR_DEVICEFILE_MATCH;
						return 0;
					}
					
				} 
				else 
				{	
					unsigned char deviceType = 255, boardType = 255;
					// Geraetetyp ermitteln
					if ((iRes = m_channel4.GetTypeDevice(deviceIDRemote, deviceType)) != IV_ERROR_SUCCESS || (iRes = m_channel4.GetTypeBoard(deviceIDRemote, boardType)) != IV_ERROR_SUCCESS) 
					{	
						iRes = ResultToError(iRes, piError);
						if (iRes == 0)
						{
							if (*piError == IV_ERROR_FUNCTION_UNAVAILABLE)
							{
								// Der Geraetetyp wird manuell gesetzt.
								deviceType = 0;
							}
							else
							{
								return iRes;
							}
						}						
					}
					
					//#######################################################################
					// Wenn die SoftwareInfo fehlt kann die Datei keine Mulitple sein!!!
					
					switch (deviceType) 
					{
					case IV_VALUE_TYPE_DEVICE_PZE: // PZE-MasterIV
						if (m_fileDevice.GetHeaderVersionFile() != IV_DFFILEDEVICE_VERSION_PZE) 
						{
							// Die gewaehlte Geraetedatei ist NICHT fuer das PZE-MasterIV.
							*piError = IV_ERROR_DEVICEFILE_WRONG;
							return 0;
						}
						break;
						
					case IV_VALUE_TYPE_DEVICE_ZK: // ZK-MasterIV
						if (m_fileDevice.GetHeaderVersionFile() != IV_DFFILEDEVICE_VERSION_ZK) 
						{
							// Die gewaehlte Geraetedatei ist NICHT fuer das ZK-MasterIV.
							*piError = IV_ERROR_DEVICEFILE_WRONG;
							return 0;
						}
						break;
						
					case IV_VALUE_TYPE_DEVICE_BDE: // BDE-MasterIV
						if (m_fileDevice.GetHeaderVersionFile() != IV_DFFILEDEVICE_VERSION_BDE) 
						{
							// Die gewaehlte Geraetedatei ist NICHT fuer das BDE-MasterIV.
							*piError = IV_ERROR_DEVICEFILE_WRONG;
							return 0;
						}
						break;
						
					case IV_VALUE_TYPE_DEVICE_MOBIL: // Mobil-MasterIV
						if (m_fileDevice.GetHeaderVersionFile() != IV_DFFILEDEVICE_VERSION_MOBIL) 
						{
							// Die gewaehlte Geraetedatei ist NICHT fuer das Mobil-MasterIV.
							*piError = IV_ERROR_DEVICEFILE_WRONG;
							return 0;
						}
						break;
						
					case IV_VALUE_TYPE_DEVICE_MDE: // MDE-BoxIV
						if (m_fileDevice.GetHeaderVersionFile() != IV_DFFILEDEVICE_VERSION_MDE) 
						{
							// Die gewaehlte Geraetedatei ist NICHT fuer das MDE-BoxIV
							*piError = IV_ERROR_DEVICEFILE_WRONG;
							return 0;
						}
						break;
						
					case IV_VALUE_TYPE_DEVICE_AE: // AE-MasterIV
						if (m_fileDevice.GetHeaderVersionFile() != IV_DFFILEDEVICE_VERSION_AE) 
						{
							// Die gewaehlte Geraetedatei ist NICHT fuer das AE-MasterIV.
							*piError = IV_ERROR_DEVICEFILE_WRONG;
							return 0;
						}
						break;
						
					case IV_VALUE_TYPE_DEVICE_TIMEBOY: // TimeboyIV
						if (m_fileDevice.GetHeaderVersionFile() != IV_DFFILEDEVICE_VERSION_TIMEBOY) 
						{
							// Die gewaehlte Geraetedatei ist NICHT fuer das TimeboyIV.
							*piError = IV_ERROR_DEVICEFILE_WRONG;
							return 0;
						}
						break;

					case IV_VALUE_TYPE_DEVICE_EXKLUSIV: // Exklusiv-LineIV
						if (m_fileDevice.GetHeaderVersionFile() != IV_DFFILEDEVICE_VERSION_EXKLUSIV) 
						{
							// Die gewaehlte Geraetedatei ist NICHT fuer das Exklusiv-LineIV.
							*piError = IV_ERROR_DEVICEFILE_WRONG;
							return 0;
						}
						break;

					case IV_VALUE_TYPE_DEVICE_FLEX: // Flex-MasterIV
						if (m_fileDevice.GetHeaderVersionFile() != IV_DFFILEDEVICE_VERSION_FLEX) 
						{
							// Die gewaehlte Geraetedatei ist NICHT fuer das Flex-MasterIV.
							*piError = IV_ERROR_DEVICEFILE_WRONG;
							return 0;
						}
						break;

					case IV_VALUE_TYPE_DEVICE_TIMEBOY_MOBIL_PZE: // Timeboy-Mobil-PZE
						if (m_fileDevice.GetHeaderVersionFile() != IV_DFFILEDEVICE_VERSION_TIMEBOY_MOBIL_PZE) 
						{
							// Die gewaehlte Geraetedatei ist NICHT fuer das Timeboy-Mobil-PZE.
							*piError = IV_ERROR_DEVICEFILE_WRONG;
							return 0;
						}
						break;

					default:
						// Die Setupanwendung ist nicht aktuell genug fuer diesen Geraetetyp.
						*piError = IV_ERROR_DEVICE_UNKNOWN;
						return 0;
					}
					char version[64];
					if (boardType < IV_VALUE_TYPE_BOARD_3_0)
					{
						// Handelt es sich um einen Boardtyp < 3 muss die Firmwareversion 04.00.xx oder 04.01.xx sein.
						if (m_fileDevice.GetHeaderVersionFirmware(version)) 
						{
							if (strstr(version, "04.00.") != version && strstr(version, "04.01.") != version)
							{
								*piError = IV_ERROR_DEVICEFILE_MATCH;
								return 0;
							}
						}
					}
					else
					{
						// Handelt es sich um einen Boardtyp >= 3 darf die Firmwareversion nicht 04.00.xx oder 04.01.xx sein.
						if (m_fileDevice.GetHeaderVersionFirmware(version)) 
						{
							if (strstr(version, "04.00.") == version || strstr(version, "04.01.") == version)
							{
								*piError = IV_ERROR_DEVICEFILE_MATCH;
								return 0;
							}
						}
					}

				}
				
			} 
			else 
			{ // Es handelt sich um eine Zipdatei

				// Auslesen der Hardwareinfo
				TDFHardwareInfo hwInfo;
				if ((iRes = m_channel4.GetHardwareInfo(deviceIDRemote, hwInfo)) != IV_ERROR_SUCCESS) 
				{
					return ResultToError(iRes, piError);
				}

				// Liegt keine Hardwareinfo vor?
				// Verwendung der ZK-Option ueber Vorhandensein der Readerliste ermitteln.
				if (hwInfo.version == 0) 
				{
					int handle;
					if (m_channel4.TableOpen(deviceIDRemote, "ACCESS.READER", 0, handle, "") == IV_ERROR_SUCCESS && 
						m_channel4.TableClose(deviceIDRemote, handle) == IV_ERROR_SUCCESS) 
					{
						// ZK-Option in Verwendung.
						hwInfo.administrationFlags |= HIAF_HAS_ZK;
					}
				}
				
				// Pruefen ob die Datei in die vorliegende Hardware eingespielt werden kann.
				if (0 == m_fileDevice.SearchInfosZip(szFileName, hwInfo)) 
				{
					// Hardwareinfo in Text aufloesen
					m_channel4.GetInfos(hwInfo, m_info);
					if (m_fileDevice.GetLastErrorState() == ES_VERSION || m_fileDevice.GetLastErrorState() == ES_DEVICE_UNKNOWN) 
					{
						*piError = IV_ERROR_VERSION_DEVICEFILE;
					} 
					else if (m_fileDevice.GetLastErrorState() == ES_NO_FILE_MATCH) 
					{
						*piError = IV_ERROR_DEVICEFILE_MATCH;
					} 
					else 
					{
						*piError = IV_ERROR_DEVICEFILE_LOAD;
					}
					return 0;
				}
			}	
		} 
		else if (S_ISDIR(st.st_mode)) 
		{		
#ifdef _WIN32
			// Auswahl aus Verzeichnis nur unter Linux gueltig!
			*piError = IV_ERROR_FUNCTION_PARAMETER;
			return 0;
#else
			// Auslesen der Hardwareinfo
			TDFHardwareInfo hwInfo;
			if ((iRes = m_channel4.GetHardwareInfo(deviceIDRemote, hwInfo)) != IV_ERROR_SUCCESS) 
			{
				return ResultToError(iRes, piError);
			}

			// Liegt keine Hardwareinfo vor?
			// Verwendung der ZK-Option ueber Vorhandensein der Readerliste ermitteln.
			if (hwInfo.version == 0) 
			{
				int handle;
				if (m_channel4.TableOpen(deviceIDRemote, "ACCESS.READER", 0, handle, "") == IV_ERROR_SUCCESS && 
					m_channel4.TableClose(deviceIDRemote, handle) == IV_ERROR_SUCCESS) 
				{
					// ZK-Option in Verwendung.
					hwInfo.administrationFlags |= HIAF_HAS_ZK;
				}
			}
			
			// Pruefen ob eine passende Datei in die vorliegende Hardware eingespielt werden kann.
			if (0 == m_fileDevice.SearchInfosDir(szFileName, hwInfo)) 
			{
				// Hardwareinfo in Text aufloesen
				m_channel4.GetInfos(hwInfo, m_info);
				if (m_fileDevice.GetLastErrorState() == ES_VERSION || m_fileDevice.GetLastErrorState() == ES_DEVICE_UNKNOWN) 
				{
					*piError = IV_ERROR_VERSION_DEVICEFILE;
				} 
				else if (m_fileDevice.GetLastErrorState() == ES_NO_FILE_MATCH) 
				{
					*piError = IV_ERROR_DEVICEFILE_MATCH;
				} 
				else 
				{
					*piError = IV_ERROR_DEVICEFILE_LOAD;
				}
				return 0;
			}
#endif
		} 
		else 
		{
			*piError = IV_ERROR_FILE_OPEN;
			return 0;
		}		

		break;}

	default:
		*piError = IV_ERROR_UNREACHABLE;
		return 0;
	}


	switch (protocolVersion) 
	{
	case 2: { // Protokollversion 2 gibt es nur beim PZE-MasterIV

		// Geraetemodus ermitteln
		int modus = m_channel1_2.GetModus(piError);
		if (!modus) 
		{
			return 0;
		}

		// Applikationsmodus? 
		if (modus == 2) 
		{
			// Abfrage ob Daten vorhanden
			if ((iRes = CheckData(deviceIDRemote, piError)) == -1) 
			{
				return 0;
			} 
			else if (iRes) 
			{
				*piError = IV_ERROR_DATA_AVAILABLE;
				return 0;
			}
		}

		// Geraetetexte ggf als Zeichensatz anfuegen
		if (m_fileDevice.GetFileDeviceText()->m_sizeData) 
		{
			// Zeichensatz- und Grafikdaten
			//CDFFileDisplay *pFileDisplay = m_fileDevice.GetFileDisplay();

			CDFFileDeviceText *fileDeviceText = m_fileDevice.GetFileDeviceText();
			unsigned char *buf = fileDeviceText->m_memData;
			unsigned int size = (unsigned int)fileDeviceText->m_sizeData;

			// Texte als Zeichensatz mit uebertragen.
			int err;
			if (!(err = m_fileDevice.GetFileDisplay()->AppendSegmentText(buf, size))) 
			{
				if (ES_ALLOC_MEM == m_fileDevice.GetFileDisplay()->GetLastErrorState()) 
				{
					*piError = IV_ERROR_ALLOC_MEM;
					return 0;
				}

				*piError = IV_ERROR_FILE_PROCESSING;
				return 0;
			}

			// Ruecksetzen des Flags fuer "gueltige Default-Textdaten vorhanden"
			unsigned char v = 0;
			if (!m_channel1_2.WriteMem(MB_START_EEP + 0x0ECB, sizeof(unsigned char), (unsigned char*)&v, piError)) 
			{
				return 0;
			}
		}

		unsigned char loadDisplayData = 1;
		unsigned int sizeData = m_fileDevice.GetSizeData();
		unsigned int tmp = ~0;
		unsigned int address;
		unsigned int firstAddress = m_fileDevice.GetController()->GetProgramFirstAddress();
		unsigned int lastAddress  = m_fileDevice.GetController()->GetProgramLastAddress();
		unsigned int sizeImage   = lastAddress - firstAddress + 1;

		unsigned int crc = 0;

		// EEPROM - Setupzeiger -pKonf- loeschen
		//@@: Uebergangsbug wegen Speicherueberlappung, von Firmware und Setup, beseitigen
 		if (!m_channel1_2.WriteMem(MB_START_EEP + 4, sizeof(unsigned int), (unsigned char*)&tmp, piError)) 
		{
			return 0;
		}

		/* Zeichensatz und Displaydaten immer uebertragen
		// Pruefen ob die Zeichensatz und Grafikdaten vorhanden
		if (!CheckFileDisplay(m_fileDevice.GetFileDisplay(), piError)) 
		{
			if (*piError != IV_ERROR_SUCCESS) 
			{
				// Fehler waehrend Kommunikation, abbruch
				return 0;
			}

			// Keine Displaydaten oder ungleich!

		} 
		else 
		{
			// Anzeigedaten vorhanden und Identisch, also nicht uebertragen
			sizeData -= m_fileDevice.GetFileDisplay()->GetSizeData();
			loadDisplayData = 0;
		}*/

		// EEPROM - Bootload Beginn
 		if (!m_channel1_2.WriteMem(MB_BOOTLOAD, sizeof(unsigned int), (unsigned char*)&sizeData, piError)) 
		{
			return 0;
		}

		tmp = FLASH_IMAGE_BASE + firstAddress; // Basisadresse im Flash
		// EEPROM - schreiben: Startadresse des Image im externen Flash
 		if (!m_channel1_2.WriteMem(MB_START_EEP + EEPROM_IMAGE_SOURCE, sizeof(unsigned int), (unsigned char*)&tmp, piError)) 
		{
			return 0;
		}

		// EEPROM - schreiben: Zieladresse des Image
 		if (!m_channel1_2.WriteMem(MB_START_EEP + EEPROM_IMAGE_DEST, sizeof(unsigned int), (unsigned char*)&firstAddress, piError)) 
		{
			return 0;
		}

		// EEPROM - schreiben: Laenge des Images
 		if (!m_channel1_2.WriteMem(MB_START_EEP + EEPROM_IMAGE_LEN, sizeof(unsigned int), (unsigned char*)&sizeImage, piError)) 
		{
			return 0;
		}

		// CRC berechnen
		crc = GetCRC16(m_fileDevice.GetController()->GetMemProgram(firstAddress), sizeImage);

		// EEPROM - schreiben: CRC des Images
 		if (!m_channel1_2.WriteMem(MB_START_EEP + EEPROM_IMAGE_CRC, sizeof(unsigned int), (unsigned char*)&crc, piError)) 
		{
			return 0;
		}

		// Imagedaten
		for (address=firstAddress; address < (lastAddress + 1); address += MAX_MB_PAKET_DATA) 
		{
			// FLASH - schreiben
 			if (!m_channel1_2.WriteMem(FLASH_IMAGE_BASE + address, 
				(((lastAddress + 1) - address) > MAX_MB_PAKET_DATA) ? MAX_MB_PAKET_DATA : ((lastAddress + 1) - address), 
				m_fileDevice.GetController()->GetMemProgram(address), piError)) 
			{
				return 0;
			}
		} 

		// Laden der Zeichensatz- und Grafikdaten?
		if (loadDisplayData) 
		{
			// Zeichensatz- und Grafikdaten
			CDFFileDisplay *pFileDisplay = m_fileDevice.GetFileDisplay();

			// Zeichensatzdaten schreiben
			for (address=0; address < pFileDisplay->m_sizeFont; address += MAX_MB_PAKET_DATA) 
			{
				// Flash schreiben 
				if (!m_channel1_2.WriteMem(0x000B0200 + address, 
					((address + MAX_MB_PAKET_DATA) < pFileDisplay->m_sizeFont) ? MAX_MB_PAKET_DATA : pFileDisplay->m_sizeFont - address, 
					&pFileDisplay->m_memFont[address], piError)) 
				{
					return 0;
				}
			}

			// Grafikdaten schreiben
			for (address=0; address < pFileDisplay->m_sizeGrafic; address += MAX_MB_PAKET_DATA) 
			{
				if (!m_channel1_2.WriteMem(0x000B0200 + pFileDisplay->m_sizeFont + address, 
					((address + MAX_MB_PAKET_DATA) < pFileDisplay->m_sizeGrafic) ? MAX_MB_PAKET_DATA : pFileDisplay->m_sizeGrafic - address,
					&pFileDisplay->m_memGrafic[address], piError)) 
				{
					return 0;
				}
			}

			tmp = 0;
			ChangeOrderLong((unsigned char*)&tmp);
			// EEPROM - Zeiger auf Font setzen 
 			if (!m_channel1_2.WriteMem(MB_START_EEP + 0x0F78, sizeof(unsigned int), (unsigned char*)&tmp, piError)) 
			{
				return 0;
			}

			tmp = pFileDisplay->m_sizeFont;
			ChangeOrderLong((unsigned char*)&tmp);
			// EEPROM - Zeiger auf Grafik setzen 
 			if (!m_channel1_2.WriteMem(MB_START_EEP + 0x0F7C, sizeof(unsigned int), (unsigned char*)&tmp, piError)) 
			{
				return 0;
			}
		}

		// Schreiben der Geraetetexte
		// Noch einbauen!! 

		sizeData = 0;
		// EEPROM - Bootload Ende
 		if (!m_channel1_2.WriteMem(MB_BOOTLOAD, sizeof(unsigned int), (unsigned char*)&sizeData, piError)) 
		{
			return 0;
		}

		// Protokollversion ist aufgrund eines Kommunikationsfehlers auf 0 zu setzen.
		// Naechste Kommunikationsanfage ermittelt die Protokollversion neu.
		m_protocolVersion[(GetDeviceIDRemote() == 0xFE) ? 0 : GetDeviceIDRemote()] = 0;

		return 1;}

	case 4: { // MasterIV mit Protokollversion 4

		//######################################################################
		// 1. Ermittlung ob auf dem Geraet eine Applicationfirmware laeuft.
		// 2. Wenn Punkt 1 == TRUE, dann wird geprueft ob Datensaetze verfuegbar sind.
		// 2. Wenn Punkt 2 == TRUE, wird die Aktion abgebrochen.
		char version[64];
		int modus = 0;
		// Geraetemodus (Application/Bootloader) ermitteln
		if ((iRes = m_channel4.GetVersion(deviceIDRemote, version)) != IV_ERROR_SUCCESS) 
		{
			iRes = ResultToError(iRes, piError);
			if (*piError == IV_ERROR_FUNCTION_UNAVAILABLE)
			{
				// Bootloadermodus
				modus = 1;
			}
			else
			{
				return iRes;
			}
		}

		iRes = 0;
		// Abfrage nur wenn Geraet im Applicationmodus
		// Abfrage ob Daten vorhanden
		if (!modus && (iRes = CheckData(deviceIDRemote, piError)) == -1) 
		{
			return 0;
		} 
		else if (iRes) 
		{
			*piError = IV_ERROR_DATA_AVAILABLE;
			return 0;
		}

		// Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
		if ((iRes = m_channel4.SetTimeboySlotState(GetDeviceIDRemote(), 1)) == IV_ERROR_SUCCESS)
		{
			// Befehlsausfuehrung starten.
			iRes = UploadDevice4(deviceIDRemote);

			// Versuchen die rote LED abzuschalten.
			m_channel4.SetTimeboySlotState(GetDeviceIDRemote(), 0, iRes);
		}

		iRes = ResultToError(iRes, piError);
		if (iRes)
		{
			// Protokollversion ist aufgrund eines Kommunikationsfehlers auf 0 zu setzen.
			// Naechste Kommunikationsanfage ermittelt die Protokollversion neu.
			m_protocolVersion[(GetDeviceIDRemote() == 0xFE) ? 0 : GetDeviceIDRemote()] = 0;
      // Ggf. unbekannte Dockingstation
      m_channel4.ResetTimeboyStation();
		}

		return iRes;}
	}

	return 0;
}

int CAEChannel::UploadDevice4(short deviceIDRemote)
{
	// Zeichensatz- und Grafikdaten
	CDFFileDisplay *pFileDisplay = m_fileDevice.GetFileDisplay();
	CDFFileDeviceText *pFileDeviceText = m_fileDevice.GetFileDeviceText();
	int res = IV_ERROR_SUCCESS;
	unsigned int address = 0;
	unsigned char *buf;
	unsigned int size = 0;
	unsigned int crc = 0;
	char version[64];

	memset(version, 0, sizeof(version));
	// Bootversion lesen 
	if ((res = m_channel4.GetVersionBootloader(deviceIDRemote, version)) != IV_ERROR_SUCCESS)
	{
		return res;
	}

	// Systemstopp setzen
	if ((res = m_channel4.BeginSetSegment(deviceIDRemote, IV_VALUE_SEGMENT_UPDATE, address, size, crc)) != IV_ERROR_SUCCESS)
	{
		return res;
	}

	// Sind Textdaten vorhanden?
	if (pFileDeviceText->GetSizeData())
	{
		// Texte als ersten Block uebertragen
		// TEXTE

		address = 0;
		buf = pFileDeviceText->m_memData;
		size = (unsigned int)pFileDeviceText->m_sizeData;
		crc = m_channel4.GetCRC32(buf, size);

		// Systemstopp setzen
		res = m_channel4.BeginSetSegment(deviceIDRemote, IV_VALUE_SEGMENT_TEXT, address, size, crc);
		if (res != IV_ERROR_SUCCESS && (res & 0xFF) != 1)
		{
			return res;
		}

		// Funktionsfehler nicht Blocktyp unbekannt?
		if ((res & 0xFF) != 1)
		{
			// Textdaten uebertragen
			if ((res = m_channel4.SetFlash(deviceIDRemote, address, buf, size)) != IV_ERROR_SUCCESS)
			{
				return res;
			}

			// Systemstopp aufheben
			if ((res = m_channel4.EndSetSegment(deviceIDRemote, IV_VALUE_SEGMENT_TEXT, crc)) != IV_ERROR_SUCCESS)
			{
				return res;
			}
			
		}
		else
		{
			// Texte als Zeichensatz mit uebertragen.
			int err;
			if (!(err = pFileDisplay->AppendSegmentText(buf, size)))
			{
				if (ES_ALLOC_MEM == pFileDisplay->GetLastErrorState())
				{
					return -IV_ERROR_ALLOC_MEM;
				}

				return -IV_ERROR_FILE_PROCESSING;
			}

			// Ruecksetzen des Flags fuer "gueltige Default-Textdaten vorhanden"
			unsigned char v = 0;
			if ((res = m_channel4.SetEEPROM(deviceIDRemote, 0x0ECB, &v, 1)) != IV_ERROR_SUCCESS)
			{
				return res;
			}
		}
	}

	// Sind Displaydaten vorhanden?
	if (pFileDisplay->GetSizeData())
	{
		// ZEICHENSAETZE
		if (pFileDisplay->m_sizeFont)
		{
			address = 0;
			buf = pFileDisplay->m_memFont;
			size = pFileDisplay->m_sizeFont;
			crc = m_channel4.GetCRC32(buf, size);

			// Systemstopp setzen
			if ((res = m_channel4.BeginSetSegment(deviceIDRemote, IV_VALUE_SEGMENT_FONTS, address, size, crc)) != IV_ERROR_SUCCESS)
			{
				return res;
			}

			// Zeichensatzdaten uebertragen
			if ((res = m_channel4.SetFlash(deviceIDRemote, address, buf, size)) != IV_ERROR_SUCCESS)
			{
				return res;
			}

			// Systemstopp aufheben
			if ((res = m_channel4.EndSetSegment(deviceIDRemote, IV_VALUE_SEGMENT_FONTS, crc)) != IV_ERROR_SUCCESS)
			{
				return res;
			}
		}

		// GRAFIKEN
		if (pFileDisplay->m_sizeGrafic)
		{
			address = 0;
			buf = pFileDisplay->m_memGrafic;
			size = pFileDisplay->m_sizeGrafic;
			crc = m_channel4.GetCRC32(buf, size);

			// Systemstopp setzen
			if ((res = m_channel4.BeginSetSegment(deviceIDRemote, IV_VALUE_SEGMENT_GRAFICS, address, size, crc)) != IV_ERROR_SUCCESS)
			{
				return res;
			}

			// Grafikdaten uebertragen
			if ((res = m_channel4.SetFlash(deviceIDRemote, address, buf, size)) != IV_ERROR_SUCCESS)
			{
				return res;
			}

			// Systemstopp aufheben
			if ((res = m_channel4.EndSetSegment(deviceIDRemote, IV_VALUE_SEGMENT_GRAFICS, crc)) != IV_ERROR_SUCCESS) {

				return res;
			}
		}
	}

	// FIRMWARE

	address = 0;
	buf = m_fileDevice.GetController()->GetMemProgram(m_fileDevice.GetController()->GetProgramFirstAddress());
	size   = m_fileDevice.GetController()->GetProgramLastAddress() - m_fileDevice.GetController()->GetProgramFirstAddress() + 1;
	crc = m_channel4.GetCRC32(buf, size);

	// Systemstopp setzen
	if ((res = m_channel4.BeginSetSegment(deviceIDRemote, IV_VALUE_SEGMENT_FIRMWARE, address, size, crc)) != IV_ERROR_SUCCESS)
	{
		return res;
	}

	// Firmwaredaten uebertragen
	if ((res = m_channel4.SetFlash(deviceIDRemote, address, buf, size)) != IV_ERROR_SUCCESS)
	{
		return res;
	}

	// Systemstopp aufheben
	if ((res = m_channel4.EndSetSegment(deviceIDRemote, IV_VALUE_SEGMENT_FIRMWARE, crc)) != IV_ERROR_SUCCESS)
	{
		return res;
	}

	// Schreiben der Geraetetexte
	/* Noch einbauen!! */

	crc = 0;
	// Systemstopp aufheben
	if ((res = m_channel4.EndSetSegment(deviceIDRemote, IV_VALUE_SEGMENT_UPDATE, crc)) != IV_ERROR_SUCCESS)
	{
		return res;
	}

	return res;
}

int CAEChannel::UploadModule(short deviceIDRemote, int type, char *szFileName, int *piError)
{
	// Zeichensatz- und Grafikdaten
	int res;
	unsigned int address = 0;
	unsigned char *buf = NULL;
	unsigned int size = 0;
	unsigned int crc = 0;
//	char version[64];
	unsigned char val;

//	memset(version, 0, sizeof(version));
	// Bootversion lesen 
//	if ((res = m_channel4.GetVersionBootloader(deviceIDRemote, version)) != IV_ERROR_SUCCESS) {

//		return res;
//	}

	// Dateigroesse ermitteln
	// Datei oeffnen
	struct stat st;
	FILE *fp = NULL;

	if (stat(szFileName, &st) != 0 || st.st_size == 0 || (fp = fopen(szFileName, "rb")) == NULL) {

		*piError = IV_ERROR_FILE_OPEN;
		res = 0;
		goto error_label;
	}

	// Dateigroesse muss groesser 4kByte sein.
	if (st.st_size < 4096)
	{
		*piError = IV_ERROR_FILE_FORMAT;
		res = 0;
		goto error_label;
	}

	// Speicher reservieren
	buf = new unsigned char[st.st_size + 1];
	if (buf == NULL) {

		*piError = IV_ERROR_FILE_OPEN;
		res = 0;
		goto error_label;
	}

	// Daten laden
	if (fread(buf, sizeof(unsigned char), st.st_size, fp) != st.st_size) {

		*piError = IV_ERROR_FILE_READ;
		res = 0;
		goto error_label;
	}

	// Zugriff ueber Struktur initialisieren
	m_uploadModuleSignatur = (SUploadModuleSignatur*)&buf[st.st_size - sizeof(SUploadModuleSignatur)];
	// Gesamt-CRC pruefen
	crc = m_channel4.GetCRC32(buf, st.st_size - sizeof(unsigned int));
	if (crc != m_uploadModuleSignatur->crc32)
	//if (crc = m_channel4.GetCRC32(buf, st.st_size))
	{
		*piError = IV_ERROR_FILE_CRC;
		res = 0;
		goto error_label;
	}

	// Pruefung auf unterstutzte Version
	if (m_uploadModuleSignatur->version != 1)
	{
		*piError = IV_ERROR_FILE_FORMAT;
		res = 0;
		goto error_label;
	}

	// Type des Moduls lesen
	if ((res = m_channel4.GetTypeFingerprint(deviceIDRemote, val)) != IV_ERROR_SUCCESS) {

		if (res < 0) {

			// Kommunikationsfehler
			*piError = -res;
		} else if (res > 0) {

			// Wenn der Modulfehler -Funktion unbekannt- ist,
			// dann ist die Firmwareversion nicht kompatible zum Befehl.
			if (((res >> 8) & 0xFF) == 2) {

				// Hintergrundladen noch nicht implementiert
				*piError = IV_ERROR_VERSION_FIRMWARE;
			} else if ((res & 0xFF) == 28) {

				// Passwort f�r Kommunikation wird ben�tigt.
				*piError = IV_ERROR_NEED_PASSWORD;
			} else if ((res & 0xFF) == 29) {

				// Nach 3 maliger Passwortfalscheingabe ist Ger�t f�r 60 Sekunden gesperrt.
				*piError = IV_ERROR_PASSWORD_LOCK;
			} else {

				// Funktionsfehler
				*piError = IV_ERROR_PROCESSING;
			}
		}

		// Fehlerhafte durchfuehrung zurueckmelden.
		res = 0;
		goto error_label;
	}

	// Pruefung auf zu aktualisierendes Modul
	if (m_uploadModuleSignatur->type != val)
	{
		// In der Datei liegt ein Update fuer eine andere Modulversion vor
		*piError = IV_ERROR_FILE_DATA_FORMAT;
		res = 0;
		goto error_label;
	}

	// Systemstopp setzen
	if ((res = m_channel4.BeginSetSegment(deviceIDRemote, IV_VALUE_SEGMENT_UPDATE, address, size, crc)) != IV_ERROR_SUCCESS) {

		if (res < 0) {

			// Kommunikationsfehler
			*piError = -res;
		} else if (res > 0) {

			// Wenn der Modulfehler -Funktion unbekannt- ist,
			// dann ist die Firmwareversion nicht kompatible zum Befehl.
			if (((res >> 8) & 0xFF) == 2) {

				// Hintergrundladen noch nicht implementiert
				*piError = IV_ERROR_VERSION_FIRMWARE;
			} else if ((res & 0xFF) == 28) {

				// Passwort f�r Kommunikation wird ben�tigt.
				*piError = IV_ERROR_NEED_PASSWORD;
			} else if ((res & 0xFF) == 29) {

				// Nach 3 maliger Passwortfalscheingabe ist Ger�t f�r 60 Sekunden gesperrt.
				*piError = IV_ERROR_PASSWORD_LOCK;
			} else {

				// Funktionsfehler
				*piError = IV_ERROR_PROCESSING;
			}
		}

		// Fehlerhafte durchfuehrung zurueckmelden.
		res = 0;
		goto error_label;
	}


	// FIRMWARE des Biokey laden

	address = 0;
	size   = st.st_size - sizeof(SUploadModuleSignatur);
	crc = m_channel4.GetCRC32(buf, size);

	// Systemstopp setzen
	if ((res = m_channel4.BeginSetSegment(deviceIDRemote, IV_VALUE_SEGMENT_FINGERPRINT, address, size, crc)) != IV_ERROR_SUCCESS) {

		if (res < 0) {

			// Kommunikationsfehler
			*piError = -res;
		} else if (res > 0) {

			// Wenn der Modulfehler -Funktion unbekannt- ist,
			// dann ist die Firmwareversion nicht kompatible zum Befehl.
			if (((res >> 8) & 0xFF) == 2) {

				// Fingerprintmodul noch nicht implementiert
				*piError = IV_ERROR_VERSION_FIRMWARE;
			} else if ((res & 0xFF) == 28) {

				// Passwort f�r Kommunikation wird ben�tigt.
				*piError = IV_ERROR_NEED_PASSWORD;
			} else if ((res & 0xFF) == 29) {

				// Nach 3 maliger Passwortfalscheingabe ist Ger�t f�r 60 Sekunden gesperrt.
				*piError = IV_ERROR_PASSWORD_LOCK;
			} else {

				// Funktionsfehler
				*piError = IV_ERROR_PROCESSING;
			}
		}

		// Fehlerhafte durchfuehrung zurueckmelden.
		res = 0;
		goto error_label;
	}

	// Firmwaredaten uebertragen
	if ((res = m_channel4.SetFlash(deviceIDRemote, address, buf, size)) != IV_ERROR_SUCCESS) {

		if (res < 0) {

			// Kommunikationsfehler
			*piError = -res;
		} else if (res > 0) {

			// Funktionsfehler
			*piError = IV_ERROR_PROCESSING;
		}

		// Fehlerhafte durchfuehrung zurueckmelden.
		res = 0;
		goto error_label;
	}

	// Systemstopp aufheben
	if ((res = m_channel4.EndSetSegment(deviceIDRemote, IV_VALUE_SEGMENT_FINGERPRINT, crc)) != IV_ERROR_SUCCESS) {

		if (res < 0) {

			// Kommunikationsfehler
			*piError = -res;
		} else if (res > 0) {

			// Wenn der Modulfehler -Funktion unbekannt- ist,
			// dann ist die Firmwareversion nicht kompatible zum Befehl.
			if (((res >> 8) & 0xFF) == 2) {

				// Hintergrundladen noch nicht implementiert
				*piError = IV_ERROR_VERSION_FIRMWARE;
			} else if ((res & 0xFF) == 28) {

				// Passwort f�r Kommunikation wird ben�tigt.
				*piError = IV_ERROR_NEED_PASSWORD;
			} else if ((res & 0xFF) == 29) {

				// Nach 3 maliger Passwortfalscheingabe ist Ger�t f�r 60 Sekunden gesperrt.
				*piError = IV_ERROR_PASSWORD_LOCK;
			} else {

				// Funktionsfehler
				*piError = IV_ERROR_PROCESSING;
			}
		}

		// Fehlerhafte durchfuehrung zurueckmelden.
		res = 0;
		goto error_label;
	}


	crc = 0;
	// Systemstopp aufheben
	if ((res = m_channel4.EndSetSegment(deviceIDRemote, IV_VALUE_SEGMENT_UPDATE, crc)) != IV_ERROR_SUCCESS) {

		if (res < 0) {

			// Kommunikationsfehler
			*piError = -res;
		} else if (res > 0) {

			// Funktionsfehler
			*piError = IV_ERROR_PROCESSING;
		}

		// Fehlerhafte durchfuehrung zurueckmelden.
		res = 0;
		goto error_label;
	}

	// Ausfuehrung erfolgreich abgeschlossen
	res = 1;

error_label:

	if (buf) delete[] buf;
	if (fp) fclose(fp);

	return res;
}

//#########################################################
//######## Schnittstelle zu Channel1/Channel2 #############
//#########################################################

int CAEChannel::GetModus(short deviceIDRemote, int *error)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 4: return -IV_ERROR_VERSION_FIRMWARE;
	case 2: return m_channel1_2.GetModus(error);
	}

	return -IV_ERROR_UNREACHABLE;
}


//#########################################################
//############# Schnittstelle zu Channel4 #################
//#########################################################

void CAEChannel::SetCBFunctionWithParams(int (*pFunction)(unsigned int, unsigned int))
{
	m_channel4.SetCBFunctionWithParams(pFunction);
}

unsigned int CAEChannel::GetCRC32(unsigned char *data, unsigned int size, unsigned int crc, unsigned char closure)
{
	return m_channel4.GetCRC32(data, size, crc, closure);
}

// MODULE_HARDWARE

// Einfache Anfrage ob das Geraet antwortet.
int CAEChannel::Ping(short deviceIDRemote)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.Ping(deviceIDRemote, 0, 0);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetTypeDevice(short deviceIDRemote, unsigned char &type)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.GetTypeDevice(deviceIDRemote, type);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetTypeBoard(short deviceIDRemote, unsigned char &type)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.GetTypeBoard(deviceIDRemote, type);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetComponentRS485_0(short deviceIDRemote, unsigned char &val)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.GetComponentRS485_0(deviceIDRemote, val);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetTypeTCPIP(short deviceIDRemote, unsigned char &val)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.GetTypeTCPIP(deviceIDRemote, val);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetTypeModemMC(short deviceIDRemote, unsigned char &val)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.GetTypeModemMC(deviceIDRemote, val);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetComponentRS485_1(short deviceIDRemote, unsigned char &val)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.GetComponentRS485_1(deviceIDRemote, val);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetTypeDisplay(short deviceIDRemote, unsigned char &type)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.GetTypeDisplay(deviceIDRemote, type);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetTypeKeypad(short deviceIDRemote, unsigned char &type)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.GetTypeKeypad(deviceIDRemote, type);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetTypeTransponder(short deviceIDRemote, unsigned char &type)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.GetTypeTransponder(deviceIDRemote, type);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetTypeMMC(short deviceIDRemote, unsigned char &type)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.GetTypeMMC(deviceIDRemote, type);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetTypeRadio(short deviceIDRemote, unsigned char &type)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.GetTypeRadio(deviceIDRemote, type);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetTypeGPS(short deviceIDRemote, unsigned char &type)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.GetTypeGPS(deviceIDRemote, type);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetComponentUSB(short deviceIDRemote, unsigned char &val)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.GetComponentUSB(deviceIDRemote, val);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetTypeDCF77(short deviceIDRemote, unsigned char &type)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.GetTypeDCF77(deviceIDRemote, type);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetComponentInAnalog(short deviceIDRemote, unsigned int &val)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.GetComponentInAnalog(deviceIDRemote, val);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetComponentInDigital(short deviceIDRemote, unsigned int &val)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.GetComponentInDigital(deviceIDRemote, val);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetComponentOutDigital(short deviceIDRemote, unsigned int &val)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.GetComponentOutDigital(deviceIDRemote, val);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetTypeFingerprint(short deviceIDRemote, unsigned char &type)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.GetTypeFingerprint(deviceIDRemote, type);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetTypeExternalReader(short deviceIDRemote, unsigned char &type)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.GetTypeExternalReader(deviceIDRemote, type);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetTypeTouch(short deviceIDRemote, unsigned char &type)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.GetTypeTouch(deviceIDRemote, type);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetTypeSensor(short deviceIDRemote, unsigned char &type)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.GetTypeSensor(deviceIDRemote, type);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetComponentBattery(short deviceIDRemote, unsigned char &val)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.GetComponentBattery(deviceIDRemote, val);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetComponentCharge(short deviceIDRemote, unsigned char &val)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.GetComponentCharge(deviceIDRemote, val);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetOptionIOModule(short deviceIDRemote, unsigned char &val)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.GetOptionIOModule(deviceIDRemote, val);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetOptionTBAdaption(short deviceIDRemote, unsigned char &val)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.GetOptionTBAdaption(deviceIDRemote, val);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetOptionComServer(short deviceIDRemote, unsigned char &val)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.GetOptionComServer(deviceIDRemote, val);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetOptionTPWrite(short deviceIDRemote, unsigned char &val)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.GetOptionTPWrite(deviceIDRemote, val);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetOptionHTTP(short deviceIDRemote, unsigned char &val)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.GetOptionHTTP(deviceIDRemote, val);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetOptionLanguage(short deviceIDRemote, unsigned char &val)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.GetOptionLanguage(deviceIDRemote, val);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetOptionBiokeyUpdate(short deviceIDRemote, unsigned char &val)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.GetOptionFingerprintUpdate(deviceIDRemote, val);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::SetSetupOptions(short deviceID, unsigned short mask, unsigned short options)
{
	switch (GetProtocolVersion(deviceID))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.SetSetupOptions(deviceID, mask, options);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetSetupOptions(short deviceIDRemote, unsigned short &mask, unsigned short &options)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.GetSetupOptions(deviceIDRemote, mask, options);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetHWConfig(short deviceIDRemote, unsigned char &val)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.GetHWConfig(deviceIDRemote, val);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetHardwareInfo(short deviceIDRemote, TDFHardwareInfo &hwInfo)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.GetHardwareInfo(deviceIDRemote, hwInfo);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetInfos(TDFHardwareInfo hwInfo, char *info)
{
	return m_channel4.GetInfos(hwInfo, info);
}

// MODULE_SYSTEM

// Systemreset; Speicherbereiche loeschen
//int SystemReset(short deviceIDRemote, unsigned short modus);

// Firmwareversion lesen
int CAEChannel::GetVersion(short deviceIDRemote, char *version)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.GetVersion(deviceIDRemote, version);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetVersionBootloader(short deviceIDRemote, char *version)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1: return -IV_ERROR_VERSION_FIRMWARE;
	case 2: {
		unsigned int ver;

		// Fehlerobjekt loeschen
		int error = IV_ERROR_SUCCESS;

		// EEPROM - Bootversion lesen 
 		if (!m_channel1_2.ReadMem(MB_START_EEP + EEPROM_BOOT_VERSION, sizeof(unsigned int), (unsigned char*)&ver, &error)) {

			return -IV_ERROR_UNREACHABLE;
		}

		sprintf(version, "%02X.%02X.%02X.%02X", ((unsigned char*)&ver)[3], ((unsigned char*)&ver)[2], ((unsigned char*)&ver)[1], ((unsigned char*)&ver)[0]);

		return IV_ERROR_SUCCESS;}
	case 4: return m_channel4.GetVersionBootloader(deviceIDRemote, version);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetSerial(short deviceIDRemote, unsigned int &serial)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.GetSerial(deviceIDRemote, serial);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::OpenComServerMode(short deviceIDRemote, char *comm, unsigned int baudRate, unsigned char byteSize, unsigned char parity, unsigned char stopBits, unsigned short flags, unsigned short timeoutToClose)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2:return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.OpenComServerMode(deviceIDRemote, comm, baudRate, byteSize, parity, stopBits, flags, timeoutToClose);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::CloseComServerMode(short deviceIDRemote)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	return m_channel4.CloseComServerMode(deviceIDRemote);
}


// Datum/Uhrzeit setzen/lesen
//int SetRTC(short deviceIDRemote, char *time);
//int GetRTC(short deviceIDRemote, char *time);

// Systemvariable setzen/lesen
//int SetSystemVariable(short deviceIDRemote, char *name, char *val);
//int GetSystemVariable(short deviceIDRemote, char *name, char *val);

// Relais setzen/lesen
//int SetRelay(short deviceIDRemote, char *name, unsigned char busNum, unsigned char relayNum, unsigned char status, unsigned short time);
//int GetRelay(short deviceIDRemote, char *name, unsigned char busNum, unsigned char relayNum, unsigned char &status, unsigned short &time);

// Datensegmentuebertragung beginnen/enden
int CAEChannel::BeginSegment(short deviceIDRemote, unsigned short art, unsigned int &address, unsigned int size, unsigned int crc)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.BeginSetSegment(deviceIDRemote, art, address, size, crc);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::EndSegment(short deviceIDRemote, unsigned short art, unsigned int crc)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.EndSetSegment(deviceIDRemote, art, crc);
	}

	return -IV_ERROR_UNREACHABLE;
}


// Uebertragen von Speicherdaten in Datei.
int CAEChannel::TransferMemToFile(short deviceIDRemote, int type, int address, int size, char *fileName)
{
	StructMemFileHeader header;
	int res = IV_ERROR_SUCCESS, chunk;
	char buf[0x400];
	int baseAddress = 0, offsAddress = 0;

	FILE *fp = fopen(fileName, "wb");
	if (fp == NULL)
	{
		return -IV_ERROR_FILE_OPEN;
	}

	// Remoteadresse einstellen
	SetDeviceIDRemote(deviceIDRemote);

	// Dateiheader aufsetzen
	memset(&header, 0, sizeof(header));
	header.structSize = sizeof(header);
	header.type = type;
	header.baseAddress = baseAddress;
	header.memAddress = address;
	header.memSize = size;

	// Aufgrund Protokollversion entscheiden
	// Protokollverwendung pruefen
	switch (GetProtocolVersion(m_channel4.GetDeviceIDRemote()))
	{
	case 1:
	case 2:
		switch (type)
		{
		case 0:
			baseAddress = MB_ACCESS_SRAM;
			break;

		case 1:
			baseAddress = MB_ACCESS_EEPROM;
			break;

		case 2:
			baseAddress = MB_ACCESS_FLASH;
			break;

		case 3:
			baseAddress = MB_ACCESS_PROGRAM;
			break;

		case 4:
			baseAddress = MB_ACCESS_FLASH_PAGE;
			break;

		case 5:
			baseAddress = MB_ACCESS_MMC;
			break;
		}

		// Basisaddress nachtragen
		header.baseAddress = baseAddress;
		// Dateiheader schreiben
		if (fwrite(&header, sizeof(header), 1, fp) != 1)
		{	
			res = -IV_ERROR_FAIL;
			goto label_out;
		}

		for ( ;; )
		{
			// Naechsten Block berechnen
			chunk = ((size - offsAddress) < (int)sizeof(buf)) ? chunk = size - offsAddress : chunk = sizeof(buf);
			if (chunk == 0) break; // Keine weiteren Daten

			if (!m_channel1_2.ReadMem(baseAddress + address + offsAddress, chunk, (unsigned char*)buf, &res))
			{
				res = -IV_ERROR_FAIL;
				goto label_out;
			}

			if (fwrite(buf, chunk, 1, fp) != 1)
			{
				res = -IV_ERROR_FAIL;
				goto label_out;
			}

			offsAddress += chunk;
		}
		break;

	case 4:
		// Dateiheader schreiben
		if (fwrite(&header, sizeof(header), 1, fp) != sizeof(header))
		{
			return -IV_ERROR_FAIL;
		}

		switch (type)
		{
		case 0:
			break;

		case 1:
			if ((res = m_channel4.GetEEPROM(deviceIDRemote, address, (unsigned char*)buf, size)) != IV_ERROR_SUCCESS)
			{
				goto label_out;
			}
			break;
		case 2:
			if ((res = m_channel4.GetFlash(deviceIDRemote, address, (unsigned char*)buf, size)) != IV_ERROR_SUCCESS)
			{
				goto label_out;
			}
			break;

		case 3:
			break;
		}
		break;

	default:
		res = -IV_ERROR_UNREACHABLE;
		goto label_out;
	}

label_out:
	if (fp) fclose(fp);
	return res;
}

// Uebertragen von Speicherdaten aus einer Datei.
int CAEChannel::TransferFileToMem(short deviceIDRemote, char *fileName)
{
	StructMemFileHeader header;
	int res = IV_ERROR_SUCCESS, chunk = 0;
	int offsAddress = 0;
	char buf[0x400];
	
	FILE *fp = fopen(fileName, "rb");
	if (fp == NULL)
	{
		return -IV_ERROR_FILE_OPEN;
	}

	// Remoteadresse einstellen
	SetDeviceIDRemote(deviceIDRemote);

	// Dateiheader lesen
	if (fread(&header, sizeof(header), 1, fp) != sizeof(header))
	{	
		res = -IV_ERROR_FAIL;
		goto label_out;
	}

	// Aufgrund Protokollversion entscheiden
	// Protokollverwendung pruefen
	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: {
		offsAddress = header.baseAddress;
		for ( ;; )
		{
			// Naechsten Block berechnen
			chunk = ((header.memSize - offsAddress) < (int)sizeof(buf)) ? chunk = header.memSize - offsAddress : chunk = sizeof(buf);

			if (fread(buf, chunk, 1, fp) != (unsigned)chunk)
			{
				res = -IV_ERROR_FAIL;
				goto label_out;
			}

			if (!m_channel1_2.WriteMem(header.baseAddress + header.memAddress + offsAddress, chunk, (unsigned char*)buf, &res))
			{
				goto label_out;
			}

			offsAddress += chunk;
		}
		break;}

	case 4:
		switch (header.type)
		{
		case 0://MB_ACCESS_SRAM:
			break;

		case 1://MB_ACCESS_EEPROM:
			if ((res = m_channel4.SetEEPROM(deviceIDRemote, header.memAddress, (unsigned char*)buf, chunk)) != IV_ERROR_SUCCESS)
			{
				goto label_out;
			}
			break;

		case 2://MB_ACCESS_FLASH:
			if ((res = m_channel4.SetFlash(deviceIDRemote, header.memAddress, (unsigned char*)buf, chunk)) != IV_ERROR_SUCCESS)
			{
				goto label_out;
			}
			break;

		case 3://MB_ACCESS_PROGRAM:
			break;

		case 5://MB_ACCESS_FLASH_PAGE:
			break;

		case 6://MB_ACCESS_MMC:
			break;
		}
		break;

	default:
		res = -IV_ERROR_UNREACHABLE;
		goto label_out;
	}

label_out:
	if (fp) fclose(fp);
	return res;
}

// Flash setzen/lesen (die ersten 16kByte Bootloader koennen nur gelesen werden)
int CAEChannel::SetFlash(short deviceIDRemote, unsigned int address, unsigned char *buf, int count)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	int err;
	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1: return m_channel1_2.WriteMem(address, count, buf, &err, 32);
	case 2: return m_channel1_2.WriteMem(address, count, buf, &err);
	case 4: return m_channel4.SetFlash(deviceIDRemote, address, buf, count);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetFlash(short deviceIDRemote, unsigned int address, unsigned char *buf, int count)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	int err;
	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1: return m_channel1_2.ReadMem(address, count, buf, &err);
	case 2: return m_channel1_2.ReadMem(address, count, buf, &err);
	case 4: return m_channel4.GetFlash(deviceIDRemote, address, buf, count);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetFlashPage(short deviceIDRemote, unsigned short pageNum, unsigned short offs, unsigned char *buf, int &count)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1: 
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.GetFlashPage(deviceIDRemote, pageNum, offs, buf, count);
	}

	return -IV_ERROR_UNREACHABLE;
}

// EEPROM setzen/lesen 
int CAEChannel::SetEEPROM(short deviceIDRemote, unsigned int address, unsigned char *buf, int count)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.SetEEPROM(deviceIDRemote, address, buf, count);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetEEPROM(short deviceIDRemote, unsigned int address, unsigned char *buf, int count)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.GetEEPROM(deviceIDRemote, address, buf, count);
	}

	return -IV_ERROR_UNREACHABLE;
}

// RAM setzen/lesen 
int CAEChannel::SetRAM(short deviceIDRemote, unsigned int address, unsigned char *buf, int count)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.SetRAM(deviceIDRemote, address, buf, count);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetRAM(short deviceIDRemote, unsigned int address, unsigned char *buf, int count)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE; 
	case 4: return m_channel4.GetRAM(deviceIDRemote, address, buf, count);
	}

	return -IV_ERROR_UNREACHABLE;
}

// PROGRAM setzen/(lesen unterbunden)
//int SetProgram(short deviceIDRemote, unsigned int address, unsigned char *buf, unsigned short count);
//int GetProgram(short deviceIDRemote, unsigned int address, unsigned char *buf, unsigned short count);

// MMC setzen/lesen 
int CAEChannel::SetMMC(short deviceIDRemote, unsigned int address, unsigned char *buf, int count)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1: 
	case 2: return -IV_ERROR_VERSION_FIRMWARE; 
	case 4: return m_channel4.SetMMC(deviceIDRemote, address, buf, count);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetMMC(short deviceIDRemote, unsigned int address, unsigned char *buf, int count)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.GetMMC(deviceIDRemote, address, buf, count);
	}

	return -IV_ERROR_UNREACHABLE;
}

// Systembios aufrufen/verlassen
int CAEChannel::EnterBios(short deviceIDRemote, unsigned char protocolVersion)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (protocolVersion)
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.EnterBios(deviceIDRemote);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::ExitBios(short deviceIDRemote)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.ExitBios(deviceIDRemote);
	}

	return -IV_ERROR_UNREACHABLE;
}

// Interface setzen/lesen (RS232, TCPIP...)
int CAEChannel::SetInterface(short deviceIDRemote, unsigned char type)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.SetInterface(deviceIDRemote, type);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetInterface(short deviceIDRemote, unsigned char &type)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.GetInterface(deviceIDRemote, type);
	}

	return -IV_ERROR_UNREACHABLE;
}

// Baud-Rate setzen/lesen
int CAEChannel::SetBaud(short deviceIDRemote, unsigned int br)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.SetBaud(deviceIDRemote, br);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetBaud(short deviceIDRemote, unsigned int &br)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.GetBaud(deviceIDRemote, br);
	}

	return -IV_ERROR_UNREACHABLE;
}

// DeviceID setzen/lesen
int CAEChannel::SetDeviceIDBios(short deviceIDRemote, unsigned short id)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.SetDeviceIDBios(deviceIDRemote, id);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetDeviceIDBios(short deviceIDRemote, unsigned short &id)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.GetDeviceIDBios(deviceIDRemote, id);
	}

	return -IV_ERROR_UNREACHABLE;
}

// Bootloader aktivieren
int CAEChannel::EnterBootloader(short deviceIDRemote, unsigned char protocolVersion)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (protocolVersion)
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.EnterBootloader(deviceIDRemote);
	}

	return -IV_ERROR_UNREACHABLE;
}

// Monitor Mode vom XPort
int CAEChannel::EnterXPortMonitor(short deviceIDRemote, char *data)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.EnterXPortMonitor(deviceIDRemote, data);
	}

	//return -IV_ERROR_UNREACHABLE;
	return -IV_ERROR_ENTER_MONITOR;
}

// Monitor Mode vom XPort
int CAEChannel::EnterXPortMonitorExt(short deviceIDRemote, char *data)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.EnterXPortMonitorExt(deviceIDRemote, data);
	}

	//return -IV_ERROR_UNREACHABLE;
	return -IV_ERROR_ENTER_MONITOR;
}

int CAEChannel::CommandXPort(short deviceIDRemote, char *buf, int &count)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.CommandXPort(deviceIDRemote, buf, count);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::ExitXPortMonitor(short deviceIDRemote)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.ExitXPortMonitor(deviceIDRemote);
	}

	//return -IV_ERROR_UNREACHABLE;
	return -IV_ERROR_EXIT_MONITOR;
}

int CAEChannel::ExitXPortMonitorExt(short deviceIDRemote)
{
	// Busnummer uebernehmen
	////SetDeviceIDRemote(deviceIDRemote);

	////switch (GetProtocolVersion(deviceIDRemote))
	////{
	////case 1:
	////case 2: return -IV_ERROR_VERSION_FIRMWARE;
	////case 4: return m_channel4.ExitXPortMonitor(deviceIDRemote);
	////}

	//return -IV_ERROR_UNREACHABLE;
	return -IV_ERROR_EXIT_MONITOR;
}

int CAEChannel::MatchPortSetFactoryDefault(short deviceIDRemote, char *buf, int &count)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.MatchPortSetFactoryDefault(deviceIDRemote, buf, count);
	}

	//return -IV_ERROR_UNREACHABLE;
	return -IV_ERROR_SET_FACTORY_DEFAULT;
}
// Softwareinformation der Anwendungsfirmware abfragen
int CAEChannel::GetSoftwareInfo(short deviceIDRemote, unsigned char *swi, int &infoLength)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.GetSoftwareInfo(deviceIDRemote, swi, infoLength);
	}

	return -IV_ERROR_UNREACHABLE;
}

// Oeffnen einer Tabelle
int CAEChannel::TableOpen(short deviceIDRemote, char *name, int options, int &handle, char *reserved)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.TableOpen(deviceIDRemote, name, options, handle, reserved);
	}

	return -IV_ERROR_UNREACHABLE;
}
// Schliessen der Tabelle
int CAEChannel::TableClose(short deviceIDRemote, int handle)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.TableClose(deviceIDRemote, handle);
	}

	return -IV_ERROR_UNREACHABLE;
}

// Setzen eines Filters auf die Tabelle
int CAEChannel::TableSetFilter(short deviceIDRemote, int handle, char *filter)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.TableSetFilter(deviceIDRemote, handle, filter);
	}

	return -IV_ERROR_UNREACHABLE;
}
// Lesen des Filters auf die Tabelle
int CAEChannel::TableGetFilter(short deviceIDRemote, int handle, char *filter, short &len)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.TableGetFilter(deviceIDRemote, handle, filter, len);
	}

	return -IV_ERROR_UNREACHABLE;
}
// Loeschen des Filters der Tabelle
int CAEChannel::TableClearFilter(short deviceIDRemote, int handle)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.TableClearFilter(deviceIDRemote, handle);
	}

	return -IV_ERROR_UNREACHABLE;
}

// Ueberspringen von n Datensaetzen (offset kann negativ sein), ab top/aktuell/bottom
int CAEChannel::TableSkip(short deviceIDRemote, int handle, int offset, short origin)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.TableSkip(deviceIDRemote, handle, offset, origin);
	}

	return -IV_ERROR_UNREACHABLE;
}

// Aendern eines Spaltenwerts, des aktuellen Datensatzes
int CAEChannel::TableSetField(short deviceIDRemote, int handle, char *name, char *val)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.TableSetField(deviceIDRemote, handle, name, val);
	}

	return -IV_ERROR_UNREACHABLE;
}
// Abfragen eines Spaltenwerts, des aktuellen Datensatzes
int CAEChannel::TableGetField(short deviceIDRemote, int handle, char *name, char *val, short &len)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.TableGetField(deviceIDRemote, handle, name, val, len);
	}

	return -IV_ERROR_UNREACHABLE;
}

// Abragen eines Modulinfos
int CAEChannel::GetInfo(short deviceIDRemote, char *name, char *params, char *info, int &len)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.GetInfo(deviceIDRemote, name, params, info, len);
	}

	return -IV_ERROR_UNREACHABLE;
}

// Zutrittsbuchung durchfuehren
int CAEChannel::Entrance2Identification(short deviceIDRemote, char *TM, char *Id, char *Pin, char *status, int &statusLength)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.Entrance2Identification(deviceIDRemote, TM, Id, Pin, status, statusLength);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::Entrance2OnlineAction(short deviceIDRemote, char *TM, unsigned int mask, unsigned char type, unsigned char duration)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.Entrance2OnlineAction(deviceIDRemote, TM, mask, type, duration);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::SetNotifyOnDevice(short deviceIDRemote, int type, int state)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.SetNotify(deviceIDRemote, type, state);
	}

	return -IV_ERROR_UNREACHABLE;
}

// Datensatz an Fingerprint-Tabelle anfuegen.
int CAEChannel::FingerprintAppendRecord(short deviceIDRemote, unsigned char type, unsigned char *fingertemplate, int size)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.FingerprintTableAppendRecord(deviceIDRemote, type, fingertemplate, size);
	}

	return -IV_ERROR_UNREACHABLE;
}

// Datensatz aus Fingerprint-Tabelle lesen.
int CAEChannel::FingerprintGetRecord(short deviceIDRemote, unsigned char &type, int pid, int fid, unsigned char *fingertemplate, int &size)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.FingerprintTableGetRecord(deviceIDRemote, type, pid, fid, fingertemplate, size);
	}

	return -IV_ERROR_UNREACHABLE;
}

// Datensatz aus Fingerprint-Tabelle loeschen.
int CAEChannel::FingerprintDeleteRecord(short deviceIDRemote, int pid, int fid)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.FingerprintTableDeleteRecord(deviceIDRemote, pid, fid);
	}

	return -IV_ERROR_UNREACHABLE;
}

// Liste mit Fingerprintmoduldaten erzeugen.
int CAEChannel::FingerprintList(short deviceIDRemote, int flags, int &recordCount)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;

	case 4: {
		int res;
		unsigned char val = 0;
		// Pruefen ob das Modul verfuegbar ist.
		if ((res = m_channel4.GetTypeFingerprint(deviceIDRemote, val)) != IV_ERROR_SUCCESS)
		{
			return res;
		}
		if (!val)
		{
			return -IV_ERROR_NO_MODULE;
		}
		if ((res = m_channel4.TableGenerate(deviceIDRemote, "FINGERPRINT.IDLIST", "")) != IV_ERROR_SUCCESS)
		{
			// Liegen im Modul keine Daten vor?
			if (res == 133)
			{
				return -IV_ERROR_NO_DATA_AVAILABLE;
			}
		}
		// Flag fuer anschliessende Ermittlung der Datensatzanzahl gesetzt?
		if (flags & 0x01)
		{
			int handle;
			recordCount = -1;
			// Erstellte Tabelle oeffnen.
			if ((res = m_channel4.TableOpen(deviceIDRemote, "FINGERPRINT.IDLIST", 0, handle, "")) != IV_ERROR_SUCCESS)
			{
				return res;
			}

			// Anzahl enthaltener Datensaetze abrufen.
			if ((res = m_channel4.TableGetRecordCount(deviceIDRemote, handle, recordCount)) != IV_ERROR_SUCCESS)
			{
				return res;
			}

			// Tabelle schliessen.
			if ((res = m_channel4.TableClose(deviceIDRemote, handle)) != IV_ERROR_SUCCESS)
			{
				return res;
			}
		}

		return res;}
	}

	return -IV_ERROR_UNREACHABLE;
}

// Backup der Fingerprints erstellen
int CAEChannel::FingerprintBackup(short deviceIDRemote, char *backupFile, int flags)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;

	case 4: {
		int res = IV_ERROR_SUCCESS;
		unsigned char val = 0;
		// Pruefen ob das Modul verfuegbar ist.
		if ((res = m_channel4.GetTypeFingerprint(deviceIDRemote, val)) != IV_ERROR_SUCCESS)
		{
			return res;
		}
		if (!val)
		{
			return -IV_ERROR_NO_MODULE;
		}

		// Tabelle mit Fingerprintdaten erzeugen
		if ((res = m_channel4.TableGenerate(deviceIDRemote, "FINGERPRINT.IDLIST", "")) != IV_ERROR_SUCCESS)
		{
			// Liegen im Modul keine Daten vor?
			if (res == 133)
			{
				res = -IV_ERROR_NO_DATA_AVAILABLE;
			}

			return res;
		}
		int hTable;
		// Tabelle mit Fingerprintdaten oeffnen
		if ((res = m_channel4.TableOpen(deviceIDRemote, "FINGERPRINT.IDLIST", 0, hTable, "")) != IV_ERROR_SUCCESS)
		{
			return res;
		}
		int recordCount;
		// Pruefen ob Datensaetze vorliegen
		if ((res = m_channel4.TableGetRecordCount(deviceIDRemote, hTable, recordCount)) != IV_ERROR_SUCCESS)
		{
			// Modul oder Funktionsfehler?
			if (res > 0) m_channel4.TableClose(deviceIDRemote, hTable);
			return res;
		}
		if (recordCount == 0)
		{
			// Es liegen keine Daten vor
			return -IV_ERROR_NO_DATA_AVAILABLE;
		}
		FILE *fp;
		structId Id;
		unsigned char type = 1; // Idencom-Compactformat zuerst verwenden.
		unsigned char data[512];
		int size;
		unsigned short crc;

		// Datei oeffnen
		if ((fp = fopen(backupFile, "wb")) == NULL)
		{
			m_channel4.TableClose(deviceIDRemote, hTable);
			return -IV_ERROR_FILE_OPEN;
		}

		// Kopfstruktur vorbereiten
		StructFTFileHeader structHeader;
		memset(&structHeader, 0, sizeof(StructFTFileHeader));
		// Kopfdaten reservieren
		if (fwrite(&structHeader, 1, sizeof(StructFTFileHeader), fp) < sizeof(StructFTFileHeader))
		{
			// Daten wurden nicht geschrieben
			m_channel4.TableClose(deviceIDRemote, hTable);
			fclose(fp);
			return -IV_ERROR_FILE_WRITE;
		}
		// Kopfdaten aufsetzen. Die Datei ist erst gueltig, wenn die Kopfdaten geschrieben wurden.
		structHeader.Id = 0xDFDF;
		structHeader.offs = sizeof(StructFTFileHeader);
		structHeader.count = recordCount;

		// Fingertemplates abrufen.
		do
		{
			// Datenspeicher loeschen
			size = sizeof(data);
			memset(data, 0, sizeof(data));

			// Datensatz aus Liste abrufen
			if ((res = m_channel4.TableGetRecord(deviceIDRemote, hTable, data, (short&)size)) != IV_ERROR_SUCCESS)
			{
				// Modul oder Funktionsfehler? Dann ist die Kommunikation noch moeglich.
				if (res > 0) m_channel4.TableClose(deviceIDRemote, hTable);
				fclose(fp);
				return res;
			}
			// bekannte Daten in struktur uebernehmen
			memcpy((unsigned char*)&Id, data, sizeof(structId));

			// Datenspeicher loeschen
			size = sizeof(data);
			memset(data, 0, sizeof(data));

			// Das dazugehoerige Fingertemplate abfragen
			while ((res = m_channel4.FingerprintTableGetRecord(deviceIDRemote, type, (unsigned int)atof(Id.pid), atoi(Id.fid), data, size)) != IV_ERROR_SUCCESS)
			{
				// Wenn das Templateformat falsch war, dann ggf. noch anderes versuchen.
				// Funktionsfehler 3
				if ((res == 3) && type != 0)
				{
					type = 0;
					continue;
				}

				// Modul oder Funktionsfehler? Dann ist die Kommunikation noch moeglich.
				if (res > 0) m_channel4.TableClose(deviceIDRemote, hTable);
				fclose(fp);
				return res;
			}

			// Groesse ist abhaengig von verwendetem Templatetyp.
			size = (type == 0) ? sizeof(structFingertemplateDIN) : sizeof(structFingertemplateIdencom);
			// CRC16 ueber das Fingertemplate erstellen
			crc = GetCRC16(data, size);
			// Datensatz in die Datei wegschreiben
			if (fwrite(data, 1, (size_t)size, fp) < (size_t)size)
			{
				// Daten wurden nicht geschrieben
				m_channel4.TableClose(deviceIDRemote, hTable);
				fclose(fp);
				return -IV_ERROR_FILE_WRITE;
			}
			// CRC schreiben
			if (fwrite(&crc, 1, sizeof(unsigned short), fp) < sizeof(unsigned short))
			{
				// Daten wurden nicht geschrieben
				m_channel4.TableClose(deviceIDRemote, hTable);
				fclose(fp);
				return -IV_ERROR_FILE_WRITE;
			}

			// Naechsten Datensatz abrufen
		} while ((res = m_channel4.TableSkip(deviceIDRemote, hTable, 1, 1)) == IV_ERROR_SUCCESS);

		// Nicht: Sprung fuehrte zu EOF?
		// Funktionsfehler: 4
		if (res != 4)
		{
			// Modul oder Funktionsfehler?
			if (res > 0) m_channel4.TableClose(deviceIDRemote, hTable);
			fclose(fp);
			return res;
		}

		// Headerstruktur abschliessen und schreiben
		structHeader.type = type;
		structHeader.crc = GetCRC16((unsigned char*)&structHeader, sizeof(structHeader) - 2);

		// Auf Dateianfang und Headerstruktur schreiben
		if (fseek(fp, 0, 0) != 0 || fwrite(&structHeader, 1, sizeof(structHeader), fp) < sizeof(structHeader))
		{
			// Daten wurden nicht geschrieben
			m_channel4.TableClose(deviceIDRemote, hTable);
			fclose(fp);
			return -IV_ERROR_FILE_WRITE;
		}

		// Tabelle schliessen
		m_channel4.TableClose(deviceIDRemote, hTable);

		// Datei schliessen
		fclose(fp);

		return IV_ERROR_SUCCESS;}
	}

	return -IV_ERROR_UNREACHABLE;
}

// Flags:
// Bit 1: gesetzt bedeutet: Enthaltene Fingertemplates beibehalten.
// Wiederherstellen der Fingerprints
int CAEChannel::FingerprintRestore(short deviceIDRemote, char *backupFile, int flags)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;

	case 4: {

		int res = IV_ERROR_SUCCESS;
		FILE *fp;
		structId Id;
		unsigned char data[512];
		size_t size, sizeCheck;
		unsigned short crc, crcSegment;
		int recordSkiped = 0, recordCountAfter;
		unsigned char val = 0;
		// Pruefen ob das Modul verfuegbar ist.
		if ((res = m_channel4.GetTypeFingerprint(deviceIDRemote, val)) != IV_ERROR_SUCCESS)
		{
			return res;
		}
		if (!val)
		{
			return -IV_ERROR_NO_MODULE;
		}
		// Datei oeffnen
		if ((fp = fopen(backupFile, "rb")) == NULL)
		{
			return -IV_ERROR_FILE_OPEN;
		}

		// Kopfstruktur vorbereiten
		StructFTFileHeader structHeader;
		memset(&structHeader, 0, sizeof(StructFTFileHeader));
		// Kopfdaten lesen
		if (fread(&structHeader, 1, sizeof(StructFTFileHeader), fp) < sizeof(StructFTFileHeader))
		{
			// Daten wurden nicht gelesen
			fclose(fp);
			return -IV_ERROR_FILE_READ;
		}
		// Kopfdaten pruefen.
		crc = GetCRC16((unsigned char*)&structHeader, sizeof(StructFTFileHeader), structHeader.crc);
		if (structHeader.Id != 0xDFDF || 
			crc != 0 ||
			structHeader.type > 1)
		{
			fclose(fp);
			return -IV_ERROR_FILE_FORMAT;
		}
		// Auf Datenanfang wechseln
		fseek(fp, structHeader.offs, 0);

		// Groesse ist abhaengig von verwendetem Templatetyp.
		size = (structHeader.type == 0) ? sizeof(structFingertemplateDIN) : sizeof(structFingertemplateIdencom);

		// Datensaetze loeschen?
		if (!(flags & 1))
		{
			// Alle Datensaetze vom Modul loeschen
			if ((res = m_channel4.FingerprintTableDeleteRecord(deviceIDRemote, 0, 0)) != IV_ERROR_SUCCESS)
			{
				fclose(fp);
				return res;
			}
		}

		// Fingertemplates einspielen.
		while (!feof(fp) && !ferror(fp))
		{
			// Datenspeicher loeschen
			memset(data, 0, sizeof(data));
			memset(&Id, 0, sizeof(structId));

			// Template aus Datei laden
			if ((sizeCheck = fread(data, 1, size, fp)) < size)
			{
				if (sizeCheck == 0 && feof(fp))
				{
					continue;
				}
				else if (sizeCheck && !ferror(fp))
				{
					fclose(fp);
					return -IV_ERROR_FILE_CRC; // Daten sind inkonsistent
				}
				// Daten wurden nicht gelesen
				fclose(fp);
				return -IV_ERROR_FILE_READ;
			}
			// CRC16 des Fingertemplates erstellen
			crc = GetCRC16(data, (unsigned int)size);
			// CRC lesen
			if ((sizeCheck = fread(&crcSegment, 1, sizeof(unsigned short), fp)) < sizeof(unsigned short)) {
				if ((sizeCheck == 0 && feof(fp)) || (sizeCheck && !ferror(fp)))
				{
					fclose(fp);
					return -IV_ERROR_FILE_CRC; // Daten sind inkonsistent
				}
				// Daten wurden nicht gelesen
				fclose(fp);
				return -IV_ERROR_FILE_READ;
			}
			// CRC pruefen
			if (crc != crcSegment)
			{
				fclose(fp);
				return -IV_ERROR_FILE_CRC;
			}

			// Das Fingertemplate anfuegen
			if ((res = m_channel4.FingerprintTableAppendRecord(deviceIDRemote, structHeader.type, data, (unsigned int)size)) != IV_ERROR_SUCCESS)
			{
				// Wenn das Template bereits vorhanden war, dann ggf. uebergehen.
				// Funktionsfehler 134
				if (res == 134)
				{
					recordSkiped++;
					continue;
				}
				else if (res == 132)
				{
					// Speicher des Moduls ist voll
					res = -IV_ERROR_MEMORY_FULL;
				}
				else if (res == 3)
				{
					// Falsches Templateformat
					res = -IV_ERROR_FILE_DATA_FORMAT;

				}
				else if (res == 2)
				{
					int failLoop = 0;
					// Bei aelterer Hardware des Biokey3000 kann eine Verzoegerung nach dem loeschen der Datensaetze
					// vorkommen. Hier wird versucht das Template nochmals zu uebertragen.
					for (; failLoop < 2; failLoop++)
					{
						if ((res = m_channel4.FingerprintTableAppendRecord(deviceIDRemote, structHeader.type, data, (int)size)) == 2)
						{
							// Kleine Verzoegerung
							//m_channel4.Trace("ERROR: GetInfoVerzoegerung: %i\n", failLoop);
							Sleep(2000);
							continue;
						}
						break;
					}

					// Konnte der Befehl ausgefuehrt werden?
					// Wenn das Template bereits vorhanden war, dann ggf. uebergehen.
					// Funktionsfehler 134
					if (res == 134)
					{
						recordSkiped++;
						continue;
					}
					else if (res == 132)
					{
						// Speicher des Moduls ist voll
						res = -IV_ERROR_MEMORY_FULL;
					}
					else if (res == 3)
					{
						// Falsches Templateformat
						res = -IV_ERROR_FILE_DATA_FORMAT;
					}
					else if (res == IV_ERROR_SUCCESS)
					{
						//m_channel4.Trace("OK: Verzoegerung: %i\n", failLoop);
						continue;
					}
				}

				fclose(fp);
				return res;
			}
		}

		// Nicht: Sprung fuehrte zu EOF?
		if (ferror(fp))
		{
			fclose(fp);
			return -IV_ERROR_FILE_READ;
		}

		// Datei schliessen
		fclose(fp);

		// Tabelle mit Fingerprintdaten erzeugen
		if ((res = m_channel4.TableGenerate(deviceIDRemote, "FINGERPRINT.IDLIST", "")) != IV_ERROR_SUCCESS)
		{
			// Liegen im Modul keine Daten vor?
			if (res == 133)
			{
				res = -IV_ERROR_NO_DATA_AVAILABLE;
			}
			return res;
		}
		int hTable;
		// Tabelle mit Fingerprintdaten oeffnen
		if ((res = m_channel4.TableOpen(deviceIDRemote, "FINGERPRINT.IDLIST", 0, hTable, "")) != IV_ERROR_SUCCESS)
		{
			return res;
		}
		// Pruefen ob Datensaetze vorliegen
		if ((res = m_channel4.TableGetRecordCount(deviceIDRemote, hTable, recordCountAfter)) != IV_ERROR_SUCCESS)
		{
			// Modul oder Funktionsfehler?
			if (res > 0) m_channel4.TableClose(deviceIDRemote, hTable);
			return res;
		}
		// Tabelle schliessen
		if ((res = m_channel4.TableClose(deviceIDRemote, hTable)) != IV_ERROR_SUCCESS)
		{
			return res;
		}

		// Pruefen ob die Datensatzanzahl passt.
		if (structHeader.count > recordCountAfter)
		{
			// Es muessen mindestens die Anzahl geforderter Datensaetze vorhanden sein.		
		}

		return IV_ERROR_SUCCESS;}
	}

	return -IV_ERROR_UNREACHABLE;
}

// Baud-Rate setzen/lesen
int CAEChannel::SetSystemFlags(short deviceIDRemote, unsigned int flags)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);
	
	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE; 
	case 4: return m_channel4.SetSystemFlags(deviceIDRemote, flags);
	}
	
	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::GetSystemFlags(short deviceIDRemote, unsigned int &flags)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);
	
	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE; 
	case 4: return m_channel4.GetSystemFlags(deviceIDRemote, flags);
	}
	
	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::SetOptionFirmware(short deviceIDRemote, int mask, int options)
{
	int res = IV_ERROR_SUCCESS;

	// Externe Leserserie
	if (mask & OFF_EXTERNAL_READER)
	{
		if ((res = m_channel4.SetTypeExternalReader(deviceIDRemote, (options & OFF_EXTERNAL_READER) ? 2 : 1)) != IV_ERROR_SUCCESS)
		{
			return res;
		}
	}

	// IOModule
	if (mask & OFF_IOMODULE)
	{
		if ((res = m_channel4.SetOptionIOModule(deviceIDRemote, (options & OFF_IOMODULE) ? 1 : 0)) != IV_ERROR_SUCCESS)
		{
			// Erst ab Version 2 verfuegbar
			if (((res >> 8) & 0xFF) == 2)
			{
				// Die weitere Verarbeitung abbrechen.
				return IV_ERROR_SUCCESS;
			}

			return res;
		}
	}

	// Timeboy anbindung
	if (mask & OFF_TBADAPTION)
	{
		if ((res = m_channel4.SetOptionTBAdaption(deviceIDRemote, (options & OFF_TBADAPTION) ? 1 : 0)) != IV_ERROR_SUCCESS)
		{
			// Erst ab Version 2 verfuegbar
			if (((res >> 8) & 0xFF) == 2)
			{
				// Die weitere Verarbeitung abbrechen.
				return IV_ERROR_SUCCESS;
			}

			return res;
		}
	}

	// ComServer
	if (mask & OFF_COMSERVER)
	{
		if ((res = m_channel4.SetOptionComServer(deviceIDRemote, (options & OFF_COMSERVER) ? 1 : 0)) != IV_ERROR_SUCCESS)
		{
			// Erst ab Version 2 verfuegbar
			if (((res >> 8) & 0xFF) == 2)
			{
				// Die weitere Verarbeitung abbrechen.
				return IV_ERROR_SUCCESS;
			}

			return res;
		}
	}

	// Transponder schreiben
	if (mask & OFF_TPWRITE)
	{
		if ((res = m_channel4.SetOptionTPWrite(deviceIDRemote, (options & OFF_TPWRITE) ? 1 : 0)) != IV_ERROR_SUCCESS)
		{
			// Erst ab Version 2 verfuegbar
			if (((res >> 8) & 0xFF) == 2)
			{
				// Die weitere Verarbeitung abbrechen.
				return IV_ERROR_SUCCESS;
			}

			return res;
		}
	}

	// HTTP-Protokoll
	if (mask & OFF_HTTP)
	{
		if ((res = m_channel4.SetOptionHTTP(deviceIDRemote, (options & OFF_HTTP) ? 1 : 0)) != IV_ERROR_SUCCESS)
		{
			// Erst ab Version 2 verfuegbar
			if (((res >> 8) & 0xFF) == 2)
			{
				// Die weitere Verarbeitung abbrechen.
				return IV_ERROR_SUCCESS;
			}

			return res;
		}
	}

	// Biokey update
	if (mask & OFF_BIOKEY_UPDATE)
	{
		if ((res = m_channel4.SetOptionBiokeyUpdate(deviceIDRemote, (options & OFF_BIOKEY_UPDATE) ? 1 : 0)) != IV_ERROR_SUCCESS)
		{
			// Erst ab Version 2 verfuegbar
			if (((res >> 8) & 0xFF) == 2)
			{
				// Die weitere Verarbeitung abbrechen.
				return IV_ERROR_SUCCESS;
			}

			return res;
		}
	}

	// SetupOption: Signalverarebeitung
	if (mask & OFF_SO_IOCONTROL)
	{
		if ((res = m_channel4.SetSetupOptions(deviceIDRemote, 0x0001, (options & OFF_SO_IOCONTROL) ? 0x0001 : 0x0000)) != IV_ERROR_SUCCESS)
		{
			// Erst ab Version 3 verfuegbar
			if (((res >> 8) & 0xFF) == 2)
			{
				// Die weitere Verarbeitung abbrechen.
				return IV_ERROR_SUCCESS;
			}

			return res;
		}
	}

	return res;
}

int CAEChannel::GetOptionFirmware(short deviceIDRemote, int &mask, int &options)
{
	int res = IV_ERROR_SUCCESS;
	unsigned char val;

	// Zuruecksetzen der Werte
	mask = options = 0;

	// Externe Leserserie
	if ((res = m_channel4.GetTypeExternalReader(deviceIDRemote, val)) != IV_ERROR_SUCCESS)
	{
		return res;
	}
	mask |= OFF_EXTERNAL_READER;
	options |= (val == 2) ? OFF_EXTERNAL_READER : 0;

	// IOModule
	if ((res = m_channel4.GetOptionIOModule(deviceIDRemote, val)) != IV_ERROR_SUCCESS)
	{
		return res;
	}
	mask |= OFF_IOMODULE;
	options |= (val == 1) ? OFF_IOMODULE : 0;

	// Timeboy anbindung
	if ((res = m_channel4.GetOptionTBAdaption(deviceIDRemote, val)) != IV_ERROR_SUCCESS)
	{
		return res;
	}
	mask |= OFF_TBADAPTION;
	options |= (val == 1) ? OFF_TBADAPTION : 0;

	// ComServer
	if ((res = m_channel4.GetOptionComServer(deviceIDRemote, val)) != IV_ERROR_SUCCESS)
	{	
		// Erst ab Version 2 verfuegbar
		if (((res >> 8) & 0xFF) == 2)
		{
			// Die weitere Verarbeitung abbrechen.
			return IV_ERROR_SUCCESS;
		}

		return res;
	}
	mask |= OFF_COMSERVER;
	options |= (val == 1) ? OFF_COMSERVER : 0;

	// Transponder schreiben
	if ((res = m_channel4.GetOptionTPWrite(deviceIDRemote, val)) != IV_ERROR_SUCCESS)
	{
		// Erst ab Version 2 verfuegbar
		if (((res >> 8) & 0xFF) == 2)
		{
			// Die weitere Verarbeitung abbrechen.
			return IV_ERROR_SUCCESS;
		}

		return res;
	}
	mask |= OFF_TPWRITE;
	options |= (val == 1) ? OFF_TPWRITE: 0;

	// Http-Protokoll
	if ((res = m_channel4.GetOptionHTTP(deviceIDRemote, val)) != IV_ERROR_SUCCESS)
	{
		// Erst ab Version 2 verfuegbar
		if (((res >> 8) & 0xFF) == 2)
		{
			// Die weitere Verarbeitung abbrechen.
			return IV_ERROR_SUCCESS;
		}
		
		return res;
	}
	mask |= OFF_HTTP;
	options |= (val == 1) ? OFF_HTTP : 0;

	// Biokey Update
	if ((res = m_channel4.GetOptionFingerprintUpdate(deviceIDRemote, val)) != IV_ERROR_SUCCESS)
	{
		// Erst ab Version 2 verfuegbar
		if (((res >> 8) & 0xFF) == 2)
		{
			// Die weitere Verarbeitung abbrechen.
			return IV_ERROR_SUCCESS;
		}

		return res;
	}
	mask |= OFF_BIOKEY_UPDATE;
	options |= (val == 1) ? OFF_BIOKEY_UPDATE : 0;

	unsigned short soMask, soValue;
	// Setup Options: Signalverarbeitung
	if ((res = m_channel4.GetSetupOptions(deviceIDRemote, soMask, soValue)) != IV_ERROR_SUCCESS)
	{
		// Erst ab Version 2 verfuegbar
		if (((res >> 8) & 0xFF) == 2)
		{
			// Die weitere Verarbeitung abbrechen.
			return IV_ERROR_SUCCESS;
		}

		return res;
	}
	mask |= OFF_SO_IOCONTROL;
	options |= (soValue & 0x0001) ? OFF_SO_IOCONTROL : 0;

	return res;
}

// Setzt den zu verwendenden Zeichensatztyp fuer SendMessage oder Sendinfotext
int CAEChannel::SetTypeFont(short deviceIDRemote, int iType, int *piError)
{
	int iRes = IV_ERROR_SUCCESS;
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(GetDeviceIDRemote()))
	{
	case 1:
	case 2:
		*piError = IV_ERROR_VERSION_FIRMWARE;
		break;

	case 4: { // PZEMasterIV mit Protokollversion 4

		// Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
		if ((iRes = m_channel4.SetTimeboySlotState(GetDeviceIDRemote(), 1)) == IV_ERROR_SUCCESS)
		{
			// Befehlsausfuehrung starten.
			iRes = m_channel4.SetTypeFont(GetDeviceIDRemote(), iType);

			// Versuchen die rote LED abzuschalten.
			m_channel4.SetTimeboySlotState(GetDeviceIDRemote(), 0, iRes);
		}

		return ResultToError(iRes, piError);}

	default:
		*piError = IV_ERROR_UNREACHABLE;
	}

	return 0;
}

// Passwort pruefen
int CAEChannel::PasswordCheck(short deviceIDRemote, char *passwort)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.PasswordCheck(deviceIDRemote, passwort);
	}

	return -IV_ERROR_UNREACHABLE;
}

// Abragen eines Modulinfos
int CAEChannel::GetPasswordKey(short deviceIDRemote, char *key)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE;
	case 4: return m_channel4.PasswordKey(deviceIDRemote, key);
	}

	return -IV_ERROR_UNREACHABLE;
}

// Tastendruck ueber Kommunikationsbefehl simulieren
int CAEChannel::PressVirtualKey(short deviceIDRemote, unsigned char vk, unsigned char flags, int *piError)
{
	int iRes = IV_ERROR_SUCCESS;
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(GetDeviceIDRemote()))
	{
	case 1:
	case 2:
		*piError = IV_ERROR_VERSION_FIRMWARE;
		break;

	case 4: { // PZEMasterIV mit Protokollversion 4

		// Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
		if ((iRes = m_channel4.SetTimeboySlotState(GetDeviceIDRemote(), 1)) == IV_ERROR_SUCCESS)
		{
			// Befehlsausfuehrung starten.
			iRes = m_channel4.PressVirtualKey(GetDeviceIDRemote(), vk, flags);

			// Versuchen die rote LED abzuschalten.
			m_channel4.SetTimeboySlotState(GetDeviceIDRemote(), 0, iRes);
		}
		
		return ResultToError(iRes, piError);}

	default:
		*piError = IV_ERROR_UNREACHABLE;
	}

	return 0;
}

int CAEChannel::GetFlashStatus(short deviceIDRemote, int *piStatus, int *piError)
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	unsigned char val = 0;

	int iRes = 0;
	switch (GetProtocolVersion(GetDeviceIDRemote()))
	{
	case 1:
	case 2:
		*piError = IV_ERROR_VERSION_FIRMWARE;
		break;

	case 4:

		// Ggf. Beginn der Kommunikation ueber Mehrfachdocking starten.
		if ((iRes = m_channel4.SetTimeboySlotState(GetDeviceIDRemote(), 1)) == IV_ERROR_SUCCESS)
		{
			// Befehlsausfuehrung starten.
			iRes =  m_channel4.GetTypeDevice(deviceIDRemote, val);

			// Versuchen die rote LED abzuschalten.
			m_channel4.SetTimeboySlotState(GetDeviceIDRemote(), 0, iRes);
		}

		iRes = ResultToError(iRes, piError);
		if (iRes)
		{
			// Flags in Statusvariable uebernehmen
			*piStatus = (m_channel4.m_lastFlags & 0x70) >> 4;
		}
		break;
	default:
		*piError = IV_ERROR_UNREACHABLE;
	}

	// Letzter Fehler ist aktueller
	return iRes;
}

#ifndef DATAFOX_STUDIO_IV

// Funktions-Stumps

// MODULE_HARDWARE

int CAEChannel::SetTypeDevice(unsigned char deviceIDRemote, unsigned char type)           { return -IV_ERROR_FUNCTION_UNAVAILABLE; }
int CAEChannel::SetTypeBoard(unsigned char deviceIDRemote, unsigned char type)            { return -IV_ERROR_FUNCTION_UNAVAILABLE; }
int CAEChannel::SetTypeTCPIP(unsigned char deviceIDRemote, unsigned char val)           { return -IV_ERROR_FUNCTION_UNAVAILABLE; }
int CAEChannel::SetTypeModemMC(unsigned char deviceIDRemote, unsigned char val)         { return -IV_ERROR_FUNCTION_UNAVAILABLE; }
int CAEChannel::SetTypeDisplay(unsigned char deviceIDRemote, unsigned char type)          { return -IV_ERROR_FUNCTION_UNAVAILABLE; }
int CAEChannel::SetTypeKeypad(unsigned char deviceIDRemote, unsigned char type)           { return -IV_ERROR_FUNCTION_UNAVAILABLE; }
int CAEChannel::SetTypeTransponder(unsigned char deviceIDRemote, unsigned char type)      { return -IV_ERROR_FUNCTION_UNAVAILABLE; }
int CAEChannel::SetTypeMMC(unsigned char deviceIDRemote, unsigned char type)              { return -IV_ERROR_FUNCTION_UNAVAILABLE; }
int CAEChannel::SetTypeRadio(unsigned char deviceIDRemote, unsigned char type)            { return -IV_ERROR_FUNCTION_UNAVAILABLE; }
int CAEChannel::SetTypeGPS(unsigned char deviceIDRemote, unsigned char type)              { return -IV_ERROR_FUNCTION_UNAVAILABLE; }
int CAEChannel::SetTypeDCF77(unsigned char deviceIDRemote, unsigned char type)            { return -IV_ERROR_FUNCTION_UNAVAILABLE; }
int CAEChannel::SetTypeFingerprint(unsigned char deviceIDRemote, unsigned char type)      { return -IV_ERROR_FUNCTION_UNAVAILABLE; }
int CAEChannel::SetTypeExternalReader(unsigned char deviceIDRemote, unsigned char val)  
{ 
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE; 
	case 4: return m_channel4.SetTypeExternalReader(deviceIDRemote, val);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::SetComponentRS485_0(unsigned char deviceIDRemote, unsigned char val)    { return -IV_ERROR_FUNCTION_UNAVAILABLE; }
int CAEChannel::SetComponentRS485_1(unsigned char deviceIDRemote, unsigned char val)    { return -IV_ERROR_FUNCTION_UNAVAILABLE; }
int CAEChannel::SetComponentUSB(unsigned char deviceIDRemote, unsigned char val)        { return -IV_ERROR_FUNCTION_UNAVAILABLE; }
int CAEChannel::SetComponentInAnalog(unsigned char deviceIDRemote, unsigned int val)   { return -IV_ERROR_FUNCTION_UNAVAILABLE; }
int CAEChannel::SetComponentInDigital(unsigned char deviceIDRemote, unsigned int val)  { return -IV_ERROR_FUNCTION_UNAVAILABLE; }
int CAEChannel::SetComponentOutDigital(unsigned char deviceIDRemote, unsigned int val) { return -IV_ERROR_FUNCTION_UNAVAILABLE; }

int CAEChannel::SetOptionIOModule(unsigned char deviceIDRemote, unsigned char val)      
{ 
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE; 
	case 4: return m_channel4.SetOptionIOModule(deviceIDRemote, val);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::SetOptionTBAdaption(unsigned char deviceIDRemote, unsigned char val)    
{
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE; 
	case 4: return m_channel4.SetOptionTBAdaption(deviceIDRemote, val);
	}

	return -IV_ERROR_UNREACHABLE;

}

int CAEChannel::SetOptionComServer(unsigned char deviceIDRemote, unsigned char val)     
{ 
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE; 
	case 4: return m_channel4.SetOptionComServer(deviceIDRemote, val);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::SetOptionTPWrite(unsigned char deviceIDRemote, unsigned char val)       
{ 
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE; 
	case 4: return m_channel4.SetOptionTPWrite(deviceIDRemote, val);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::SetOptionHTTP(unsigned char deviceIDRemote, unsigned char val)          
{ 
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE; 
	case 4: return m_channel4.SetOptionHTTP(deviceIDRemote, val);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::SetOptionLanguage(unsigned char deviceIDRemote, unsigned char val)      { return -IV_ERROR_FUNCTION_UNAVAILABLE; }

int CAEChannel::SetOptionBiokeyUpdate(unsigned char deviceIDRemote, unsigned char val)  
{ 
	// Busnummer uebernehmen
	SetDeviceIDRemote(deviceIDRemote);

	switch (GetProtocolVersion(deviceIDRemote))
	{
	case 1:
	case 2: return -IV_ERROR_VERSION_FIRMWARE; 
	case 4: return m_channel4.SetOptionBiokeyUpdate(deviceIDRemote, val);
	}

	return -IV_ERROR_UNREACHABLE;
}

int CAEChannel::SetHWConfig(unsigned char deviceIDRemote, unsigned char val)            { return -IV_ERROR_FUNCTION_UNAVAILABLE; }


// MODULE_SYSTEM

int CAEChannel::SetSerial(unsigned char deviceIDRemote, unsigned int serial)             { return -IV_ERROR_FUNCTION_UNAVAILABLE; }

#endif
